@@ -1,5 +1,5 @@
-/**
- * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+/*
+ * Copyright © 2024-2025 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -21,12 +21,9 @@ import org.pkl.cli.CliRepl
 import org.pkl.commons.cli.commands.BaseCommand
 import org.pkl.commons.cli.commands.ProjectOptions
 
-class ReplCommand(helpLink: String) :
-  BaseCommand(
-    name = "repl",
-    help = "Start a REPL session",
-    helpLink = helpLink,
-  ) {
+class ReplCommand : BaseCommand(name = "repl", helpLink = helpLink) {
+  override val helpString = "Start a REPL session"
+
   private val projectOptions by ProjectOptions()
 
   override fun run() {
