@@ -1,5 +1,5 @@
-/**
- * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+/*
+ * Copyright © 2024-2025 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -16,15 +16,15 @@
 package org.pkl.cli.commands
 
 import com.github.ajalt.clikt.core.CliktCommand
+import com.github.ajalt.clikt.core.Context
 import org.pkl.cli.CliServer
 import org.pkl.commons.cli.CliBaseOptions
 
-class ServerCommand(helpLink: String) :
-  CliktCommand(
-    name = "server",
-    help = "Run as a server that communicates over standard input/output",
-    epilog = "For more information, visit $helpLink"
-  ) {
+class ServerCommand : CliktCommand(name = "server") {
+  override fun help(context: Context) =
+    "Run as a server that communicates over standard input/output"
+
+  override fun helpEpilog(context: Context) = "For more information, visit $helpLink"
 
   override fun run() {
     CliServer(CliBaseOptions()).run()
