@@ -1,5 +1,5 @@
-/**
- * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+/*
+ * Copyright © 2024-2025 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -15,6 +15,7 @@
  */
 package org.pkl.cli
 
+import java.io.File
 import java.io.StringWriter
 import java.net.URI
 import java.nio.file.FileSystems
@@ -24,8 +25,11 @@ import java.util.stream.Collectors
 import kotlin.io.path.createDirectories
 import org.assertj.core.api.Assertions.assertThat
 import org.assertj.core.api.Assertions.assertThatCode
+import org.junit.jupiter.api.AfterAll
 import org.junit.jupiter.api.Test
 import org.junit.jupiter.api.assertThrows
+import org.junit.jupiter.api.condition.DisabledOnOs
+import org.junit.jupiter.api.condition.OS
 import org.junit.jupiter.api.io.TempDir
 import org.pkl.commons.cli.CliBaseOptions
 import org.pkl.commons.cli.CliException
@@ -34,9 +38,19 @@ import org.pkl.commons.readString
 import org.pkl.commons.test.FileTestUtils
 import org.pkl.commons.test.PackageServer
 import org.pkl.commons.writeString
-import org.pkl.core.runtime.CertificateUtils
+import org.pkl.core.util.IoUtils
 
 class CliProjectPackagerTest {
+  companion object {
+    private val packageServer = PackageServer()
+
+    @AfterAll
+    @JvmStatic
+    fun afterAll() {
+      packageServer.close()
+    }
+  }
+
   @Test
   fun `missing PklProject when inferring a project dir`(@TempDir tempDir: Path) {
     val packager =
@@ -52,7 +66,7 @@ class CliProjectPackagerTest {
   }
 
   @Test
-  fun `missing PklProject when explict dir is provided`(@TempDir tempDir: Path) {
+  fun `missing PklProject when explicit dir is provided`(@TempDir tempDir: Path) {
     val packager =
       CliProjectPackager(
         CliBaseOptions(workingDir = tempDir),
@@ -101,7 +115,7 @@ class CliProjectPackagerTest {
         }
       }
     """
-        .trimIndent()
+        .trimIndent(),
     )
     tempDir.writeFile(
       "PklProject",
@@ -116,7 +130,7 @@ class CliProjectPackagerTest {
         apiTests { "myTest.pkl" }
       }
     """
-        .trimIndent()
+        .trimIndent(),
     )
     val buffer = StringWriter()
     val packager =
@@ -126,7 +140,7 @@ class CliProjectPackagerTest {
         CliTestOptions(),
         ".out/%{name}@%{version}",
         skipPublishCheck = true,
-        consoleWriter = buffer
+        consoleWriter = buffer,
       )
     val err = assertThrows<CliException> { packager.run() }
     assertThat(err).hasMessageContaining("because its API tests are failing")
@@ -173,7 +187,7 @@ class CliProjectPackagerTest {
         CliTestOptions(),
         ".out/%{name}@%{version}",
         skipPublishCheck = true,
-        consoleWriter = buffer
+        consoleWriter = buffer,
       )
     packager.run()
   }
@@ -198,6 +212,24 @@ class CliProjectPackagerTest {
       """
           .trimIndent()
       )
+    projectDir
+      .resolve("myTest.pkl-expected.pcf")
+      .writeString(
+        """
+        examples {
+          ["Bird"] {
+            new {
+              name = "Finch"
+              favoriteFruit {
+                name = "Tangerine"
+              }
+            }
+          }
+        }
+        
+      """
+          .trimIndent()
+      )
     projectDir
       .resolve("PklProject")
       .writeString(
@@ -214,7 +246,7 @@ class CliProjectPackagerTest {
 
         dependencies {
           ["birds"] {
-            uri = "package://localhost:12110/birds@0.5.0"
+            uri = "package://localhost:0/birds@0.5.0"
           }
         }
       """
@@ -227,18 +259,18 @@ class CliProjectPackagerTest {
         {
           "schemaVersion": 1,
           "resolvedDependencies": {
-            "package://localhost:12110/birds@0": {
+            "package://localhost:0/birds@0": {
               "type": "remote",
-              "uri": "projectpackage://localhost:12110/birds@0.5.0",
+              "uri": "projectpackage://localhost:0/birds@0.5.0",
               "checksums": {
-                "sha256": "3f19ab9fcee2f44f93a75a09e531db278c6d2cd25206836c8c2c4071cd7d3118"
+                "sha256": "04eec465b217fb9779489525d26e9b587e5e47ff4d584c7673a450109715bc31"
               }
             },
-            "package://localhost:12110/fruit@1": {
+            "package://localhost:0/fruit@1": {
               "type": "remote",
-              "uri": "projectpackage://localhost:12110/fruit@1.0.5",
+              "uri": "projectpackage://localhost:0/fruit@1.0.5",
               "checksums": {
-                "sha256": "b4ea243de781feeab7921227591e6584db5d0673340f30fab2ffe8ad5c9f75f5"
+                "sha256": "${PackageServer.FRUIT_SHA}"
               }
             }
           }
@@ -254,7 +286,7 @@ class CliProjectPackagerTest {
         CliTestOptions(),
         ".out",
         skipPublishCheck = true,
-        consoleWriter = buffer
+        consoleWriter = buffer,
       )
     packager.run()
   }
@@ -269,7 +301,7 @@ class CliProjectPackagerTest {
         
         name: String
       """
-          .trimIndent()
+          .trimIndent(),
       )
 
     val fooTxt =
@@ -280,7 +312,7 @@ class CliProjectPackagerTest {
         bar
         baz
       """
-          .trimIndent()
+          .trimIndent(),
       )
 
     tempDir
@@ -305,7 +337,7 @@ class CliProjectPackagerTest {
         CliTestOptions(),
         ".out/%{name}@%{version}",
         skipPublishCheck = true,
-        consoleWriter = StringWriter()
+        consoleWriter = StringWriter(),
       )
     packager.run()
     val expectedMetadata = tempDir.resolve(".out/mypackage@1.0.0/mypackage@1.0.0")
@@ -322,7 +354,7 @@ class CliProjectPackagerTest {
         "version": "1.0.0",
         "packageZipUrl": "https://foo.com",
         "packageZipChecksums": {
-          "sha256": "7f515fbc4b229ba171fac78c7c3f2c2e68e2afebae8cfba042b12733943a2813"
+          "sha256": "e83b67722ea17ba41717ce6e99ae8ee02d66df6294bd319ce403075b1071c3e0"
         },
         "dependencies": {},
         "authors": []
@@ -334,9 +366,9 @@ class CliProjectPackagerTest {
     assertThat(expectedArchive.zipFilePaths())
       .hasSameElementsAs(listOf("/", "/c", "/c/d", "/c/d/foo.txt", "/a", "/a/b", "/a/b/foo.pkl"))
     assertThat(expectedMetadataChecksum)
-      .hasContent("203ef387f217a3caee7f19819ef2b50926929269241cb7b3a29d95237b2c7f4b")
+      .hasContent("72ab32b27393bde5f316b00f184faae919378e4d7643872c605f681b14b647bf")
     assertThat(expectedArchiveChecksum)
-      .hasContent("7f515fbc4b229ba171fac78c7c3f2c2e68e2afebae8cfba042b12733943a2813")
+      .hasContent("e83b67722ea17ba41717ce6e99ae8ee02d66df6294bd319ce403075b1071c3e0")
     FileSystems.newFileSystem(URI("jar:" + expectedArchive.toUri()), mutableMapOf<String, String>())
       .use { fs ->
         assertThat(fs.getPath("a/b/foo.pkl")).hasSameTextualContentAs(fooPkl)
@@ -354,6 +386,10 @@ class CliProjectPackagerTest {
       writeEmptyFile("main.test.pkl")
       writeEmptyFile("child/main.pkl")
       writeEmptyFile("child/main.test.pkl")
+      writeEmptyFile("examples/Workflow.pkl")
+      writeEmptyFile("examples/Ex1.pkl")
+      writeEmptyFile("tests/Test1.pkl")
+      writeEmptyFile("tests/integration/TestIng1.pkl")
     }
 
     tempDir.writeFile(
@@ -370,10 +406,12 @@ class CliProjectPackagerTest {
             "*.bin"
             "child/main.pkl"
             "*.test.pkl"
+            "examples/Ex1.pkl"
+            "tests/**"
           }
         }
       """
-        .trimIndent()
+        .trimIndent(),
     )
     CliProjectPackager(
         CliBaseOptions(workingDir = tempDir),
@@ -381,7 +419,7 @@ class CliProjectPackagerTest {
         CliTestOptions(),
         ".out/%{name}@%{version}",
         skipPublishCheck = true,
-        consoleWriter = StringWriter()
+        consoleWriter = StringWriter(),
       )
       .run()
     val expectedArchive = tempDir.resolve(".out/mypackage@1.0.0/mypackage@1.0.0.zip")
@@ -389,10 +427,8 @@ class CliProjectPackagerTest {
       .hasSameElementsAs(
         listOf(
           "/",
-          "/a",
-          "/a/b",
-          "/a/b/c",
-          "/child",
+          "/examples",
+          "/examples/Workflow.pkl",
           "/input",
           "/input/foo",
           "/input/foo/bar.txt",
@@ -419,12 +455,12 @@ class CliProjectPackagerTest {
         
         dependencies {
           ["birds"] {
-            uri = "package://localhost:12110/birds@0.5.0"
+            uri = "package://localhost:0/birds@0.5.0"
           }
           ["project2"] = import("../project2/PklProject")
         }
       """
-        .trimIndent()
+        .trimIndent(),
     )
     projectDir.writeFile(
       "PklProject.deps.json",
@@ -432,22 +468,22 @@ class CliProjectPackagerTest {
         {
           "schemaVersion": 1,
           "resolvedDependencies": {
-            "package://localhost:12110/birds@0": {
+            "package://localhost:0/birds@0": {
               "type": "remote",
-              "uri": "projectpackage://localhost:12110/birds@0.5.0",
+              "uri": "projectpackage://localhost:0/birds@0.5.0",
               "checksums": {
-                "sha256": "3f19ab9fcee2f44f93a75a09e531db278c6d2cd25206836c8c2c4071cd7d3118"
+                "sha256": "${PackageServer.BIRDS_SHA}"
               }
             },
-            "package://localhost:12110/project2@5": {
+            "package://localhost:0/project2@5": {
               "type": "local",
-              "uri": "projectpackage://localhost:12110/project2@5.0.0",
+              "uri": "projectpackage://localhost:0/project2@5.0.0",
               "path": "../project2"
             }
           }
         }
       """
-        .trimIndent()
+        .trimIndent(),
     )
 
     project2Dir.writeFile(
@@ -457,12 +493,12 @@ class CliProjectPackagerTest {
         
         package {
           name = "project2"
-          baseUri = "package://localhost:12110/project2"
+          baseUri = "package://localhost:0/project2"
           version = "5.0.0"
           packageZipUrl = "https://foo.com/project2.zip"
         }
       """
-        .trimIndent()
+        .trimIndent(),
     )
     project2Dir.writeFile(
       "PklProject.deps.json",
@@ -472,7 +508,7 @@ class CliProjectPackagerTest {
           "resolvedDependencies": {}
         }
       """
-        .trimIndent()
+        .trimIndent(),
     )
 
     CliProjectPackager(
@@ -481,7 +517,7 @@ class CliProjectPackagerTest {
         CliTestOptions(),
         ".out/%{name}@%{version}",
         skipPublishCheck = true,
-        consoleWriter = StringWriter()
+        consoleWriter = StringWriter(),
       )
       .run()
     val expectedMetadata = tempDir.resolve(".out/mypackage@1.0.0/mypackage@1.0.0")
@@ -495,19 +531,19 @@ class CliProjectPackagerTest {
         "version": "1.0.0",
         "packageZipUrl": "https://foo.com",
         "packageZipChecksums": {
-          "sha256": "7d08a65078e0bfc382c16fe1bb94546ab9a11e6f551087f362a4515ca98102fc"
+          "sha256": "8739c76e681f900923b900c9df0ef75cf421d39cabb54650c4b9ad19b6a76d85"
         },
         "dependencies": {
           "birds": {
-            "uri": "package://localhost:12110/birds@0.5.0",
+            "uri": "package://localhost:0/birds@0.5.0",
             "checksums": {
-              "sha256": "3f19ab9fcee2f44f93a75a09e531db278c6d2cd25206836c8c2c4071cd7d3118"
+              "sha256": "${PackageServer.BIRDS_SHA}"
             }
           },
           "project2": {
-            "uri": "package://localhost:12110/project2@5.0.0",
+            "uri": "package://localhost:0/project2@5.0.0",
             "checksums": {
-              "sha256": "ddebb806e5b218ebb1a2baa14ad206b46e7a0c1585fa9863a486c75592bc8b16"
+              "sha256": "981787869571330b2f609a94a5912466990ce00e3fa94e7f290c2f99a6d5e5ed"
             }
           }
         },
@@ -523,11 +559,11 @@ class CliProjectPackagerTest {
         """
     {
       "name": "project2",
-      "packageUri": "package://localhost:12110/project2@5.0.0",
+      "packageUri": "package://localhost:0/project2@5.0.0",
       "version": "5.0.0",
       "packageZipUrl": "https://foo.com/project2.zip",
       "packageZipChecksums": {
-        "sha256": "7d08a65078e0bfc382c16fe1bb94546ab9a11e6f551087f362a4515ca98102fc"
+        "sha256": "8739c76e681f900923b900c9df0ef75cf421d39cabb54650c4b9ad19b6a76d85"
       },
       "dependencies": {},
       "authors": []
@@ -557,12 +593,12 @@ class CliProjectPackagerTest {
         
         dependencies {
           ["birds"] {
-            uri = "package://localhost:12110/birds@0.5.0"
+            uri = "package://localhost:0/birds@0.5.0"
           }
           ["project2"] = import("../project2/PklProject")
         }
       """
-        .trimIndent()
+        .trimIndent(),
     )
     projectDir.writeFile(
       "PklProject.deps.json",
@@ -570,22 +606,22 @@ class CliProjectPackagerTest {
         {
           "schemaVersion": 1,
           "resolvedDependencies": {
-            "package://localhost:12110/birds@0": {
+            "package://localhost:0/birds@0": {
               "type": "remote",
-              "uri": "projectpackage://localhost:12110/birds@0.5.0",
+              "uri": "projectpackage://localhost:0/birds@0.5.0",
               "checksums": {
-                "sha256": "3f19ab9fcee2f44f93a75a09e531db278c6d2cd25206836c8c2c4071cd7d3118"
+                "sha256": "0a5ad2dc13f06f73f96ba94e8d01d48252bc934e2de71a837620ca0fef8a7453"
               }
             },
-            "package://localhost:12110/project2@5": {
+            "package://localhost:0/project2@5": {
               "type": "local",
-              "uri": "projectpackage://localhost:12110/project2@5.0.0",
+              "uri": "projectpackage://localhost:0/project2@5.0.0",
               "path": "../project2"
             }
           }
         }
       """
-        .trimIndent()
+        .trimIndent(),
     )
 
     project2Dir.writeFile(
@@ -595,12 +631,12 @@ class CliProjectPackagerTest {
         
         package {
           name = "project2"
-          baseUri = "package://localhost:12110/project2"
+          baseUri = "package://localhost:0/project2"
           version = "5.0.0"
           packageZipUrl = "https://foo.com/project2.zip"
         }
       """
-        .trimIndent()
+        .trimIndent(),
     )
     project2Dir.writeFile(
       "PklProject.deps.json",
@@ -610,7 +646,7 @@ class CliProjectPackagerTest {
           "resolvedDependencies": {}
         }
       """
-        .trimIndent()
+        .trimIndent(),
     )
     assertThatCode {
         CliProjectPackager(
@@ -619,7 +655,7 @@ class CliProjectPackagerTest {
             CliTestOptions(),
             ".out/%{name}@%{version}",
             skipPublishCheck = true,
-            consoleWriter = StringWriter()
+            consoleWriter = StringWriter(),
           )
           .run()
       }
@@ -635,7 +671,7 @@ class CliProjectPackagerTest {
   
         res = foo
       """
-        .trimIndent()
+        .trimIndent(),
     )
     tempDir.writeFile(
       "PklProject",
@@ -649,7 +685,7 @@ class CliProjectPackagerTest {
           packageZipUrl = "https://foo.com"
         }
       """
-        .trimIndent()
+        .trimIndent(),
     )
     val e =
       assertThrows<CliException> {
@@ -659,7 +695,7 @@ class CliProjectPackagerTest {
             CliTestOptions(),
             ".out/%{name}@%{version}",
             skipPublishCheck = true,
-            consoleWriter = StringWriter()
+            consoleWriter = StringWriter(),
           )
           .run()
       }
@@ -676,7 +712,11 @@ class CliProjectPackagerTest {
       )
   }
 
+  // Absolute path imports on Windows must use an absolute URI (e.g. file:///C:/Foo/Bar), because
+  // they must contain drive letters, which conflict with schemes.
+  // We skip validation for absolute URIs, so effectively we skip this check on Windows.
   @Test
+  @DisabledOnOs(OS.WINDOWS)
   fun `import path verification -- absolute import from root dir`(@TempDir tempDir: Path) {
     tempDir.writeFile(
       "main.pkl",
@@ -685,7 +725,7 @@ class CliProjectPackagerTest {
   
         res = foo
       """
-        .trimIndent()
+        .trimIndent(),
     )
     tempDir.writeFile(
       "PklProject",
@@ -699,7 +739,7 @@ class CliProjectPackagerTest {
           packageZipUrl = "https://foo.com"
         }
       """
-        .trimIndent()
+        .trimIndent(),
     )
     val e =
       assertThrows<CliException> {
@@ -709,7 +749,7 @@ class CliProjectPackagerTest {
             CliTestOptions(),
             ".out/%{name}@%{version}",
             skipPublishCheck = true,
-            consoleWriter = StringWriter()
+            consoleWriter = StringWriter(),
           )
           .run()
       }
@@ -724,13 +764,14 @@ class CliProjectPackagerTest {
   }
 
   @Test
+  @DisabledOnOs(OS.WINDOWS)
   fun `import path verification -- absolute read from root dir`(@TempDir tempDir: Path) {
     tempDir.writeFile(
       "main.pkl",
       """
         res = read("$tempDir/foo.pkl")
       """
-        .trimIndent()
+        .trimIndent(),
     )
     tempDir.writeFile(
       "PklProject",
@@ -744,7 +785,7 @@ class CliProjectPackagerTest {
           packageZipUrl = "https://foo.com"
         }
       """
-        .trimIndent()
+        .trimIndent(),
     )
     val e =
       assertThrows<CliException> {
@@ -754,7 +795,7 @@ class CliProjectPackagerTest {
             CliTestOptions(),
             ".out/%{name}@%{version}",
             skipPublishCheck = true,
-            consoleWriter = StringWriter()
+            consoleWriter = StringWriter(),
           )
           .run()
       }
@@ -775,7 +816,7 @@ class CliProjectPackagerTest {
       """
         import "baz.pkl"
       """
-        .trimIndent()
+        .trimIndent(),
     )
     tempDir.writeFile(
       "PklProject",
@@ -789,7 +830,7 @@ class CliProjectPackagerTest {
           packageZipUrl = "https://foo.com"
         }
       """
-        .trimIndent()
+        .trimIndent(),
     )
     CliProjectPackager(
         CliBaseOptions(workingDir = tempDir),
@@ -797,7 +838,7 @@ class CliProjectPackagerTest {
         CliTestOptions(),
         ".out/%{name}@%{version}",
         skipPublishCheck = true,
-        consoleWriter = StringWriter()
+        consoleWriter = StringWriter(),
       )
       .run()
   }
@@ -817,7 +858,7 @@ class CliProjectPackagerTest {
           packageZipUrl = "https://foo.com"
         }
       """
-        .trimIndent()
+        .trimIndent(),
     )
     tempDir.writeFile("project2/main2.pkl", "res = 2")
     tempDir.writeFile(
@@ -832,7 +873,7 @@ class CliProjectPackagerTest {
           packageZipUrl = "https://foo.com"
         }
       """
-        .trimIndent()
+        .trimIndent(),
     )
     val out = StringWriter()
     CliProjectPackager(
@@ -841,20 +882,21 @@ class CliProjectPackagerTest {
         CliTestOptions(),
         ".out/%{name}@%{version}",
         skipPublishCheck = true,
-        consoleWriter = out
+        consoleWriter = out,
       )
       .run()
+    val sep = File.separatorChar
     assertThat(out.toString())
-      .isEqualTo(
+      .isEqualToNormalizingNewlines(
         """
-      .out/project1@1.0.0/project1@1.0.0.zip
-      .out/project1@1.0.0/project1@1.0.0.zip.sha256
-      .out/project1@1.0.0/project1@1.0.0
-      .out/project1@1.0.0/project1@1.0.0.sha256
-      .out/project2@2.0.0/project2@2.0.0.zip
-      .out/project2@2.0.0/project2@2.0.0.zip.sha256
-      .out/project2@2.0.0/project2@2.0.0
-      .out/project2@2.0.0/project2@2.0.0.sha256
+      .out${sep}project1@1.0.0${sep}project1@1.0.0.zip
+      .out${sep}project1@1.0.0${sep}project1@1.0.0.zip.sha256
+      .out${sep}project1@1.0.0${sep}project1@1.0.0
+      .out${sep}project1@1.0.0${sep}project1@1.0.0.sha256
+      .out${sep}project2@2.0.0${sep}project2@2.0.0.zip
+      .out${sep}project2@2.0.0${sep}project2@2.0.0.zip.sha256
+      .out${sep}project2@2.0.0${sep}project2@2.0.0
+      .out${sep}project2@2.0.0${sep}project2@2.0.0.sha256
 
     """
           .trimIndent()
@@ -867,43 +909,45 @@ class CliProjectPackagerTest {
 
   @Test
   fun `publish checks`(@TempDir tempDir: Path) {
-    PackageServer.ensureStarted()
-    CertificateUtils.setupAllX509CertificatesGlobally(listOf(FileTestUtils.selfSignedCertificate))
     tempDir.writeFile("project/main.pkl", "res = 1")
     tempDir.writeFile(
       "project/PklProject",
-      // intentionally conflict with localhost:12110/birds@0.5.0 from our test fixtures
+      // intentionally conflict with localhost:0/birds@0.5.0 from our test fixtures
       """
         amends "pkl:Project"
         
         package {
           name = "birds"
           version = "0.5.0"
-          baseUri = "package://localhost:12110/birds"
+          baseUri = "package://localhost:0/birds"
           packageZipUrl = "https://foo.com"
         }
       """
-        .trimIndent()
+        .trimIndent(),
     )
     val e =
       assertThrows<CliException> {
         CliProjectPackager(
-            CliBaseOptions(workingDir = tempDir),
+            CliBaseOptions(
+              workingDir = tempDir,
+              caCertificates = listOf(FileTestUtils.selfSignedCertificate),
+              testPort = packageServer.port,
+            ),
             listOf(tempDir.resolve("project")),
             CliTestOptions(),
             ".out/%{name}@%{version}",
             skipPublishCheck = false,
-            consoleWriter = StringWriter()
+            consoleWriter = StringWriter(),
           )
           .run()
       }
     assertThat(e)
       .hasMessageStartingWith(
         """
-      Package `package://localhost:12110/birds@0.5.0` was already published with different contents.
+      Package `package://localhost:0/birds@0.5.0` was already published with different contents.
       
-      Computed checksum: 7324e17214b6dcda63ebfb57d5a29b077af785c13bed0dc22b5138628a3f8d8f
-      Published checksum: 3f19ab9fcee2f44f93a75a09e531db278c6d2cd25206836c8c2c4071cd7d3118
+      Computed checksum: aa8c883841db22e92794f4708b01dc905b5da77645b7dfb5b22a73da8c347db1
+      Published checksum: ${PackageServer.BIRDS_SHA}
     """
           .trimIndent()
       )
@@ -911,8 +955,6 @@ class CliProjectPackagerTest {
 
   @Test
   fun `publish check when package is not yet published`(@TempDir tempDir: Path) {
-    PackageServer.ensureStarted()
-    CertificateUtils.setupAllX509CertificatesGlobally(listOf(FileTestUtils.selfSignedCertificate))
     tempDir.writeFile("project/main.pkl", "res = 1")
     tempDir.writeFile(
       "project/PklProject",
@@ -922,38 +964,129 @@ class CliProjectPackagerTest {
         package {
           name = "mangos"
           version = "1.0.0"
-          baseUri = "package://localhost:12110/mangos"
+          baseUri = "package://localhost:0/mangos"
           packageZipUrl = "https://foo.com"
         }
       """
-        .trimIndent()
+        .trimIndent(),
     )
     val out = StringWriter()
     CliProjectPackager(
-        CliBaseOptions(workingDir = tempDir),
+        CliBaseOptions(
+          workingDir = tempDir,
+          caCertificates = listOf(FileTestUtils.selfSignedCertificate),
+          testPort = packageServer.port,
+        ),
         listOf(tempDir.resolve("project")),
         CliTestOptions(),
         ".out/%{name}@%{version}",
         skipPublishCheck = false,
-        consoleWriter = out
+        consoleWriter = out,
       )
       .run()
+    val sep = File.separatorChar
     assertThat(out.toString())
-      .isEqualTo(
+      .isEqualToNormalizingNewlines(
         """
-      .out/mangos@1.0.0/mangos@1.0.0.zip
-      .out/mangos@1.0.0/mangos@1.0.0.zip.sha256
-      .out/mangos@1.0.0/mangos@1.0.0
-      .out/mangos@1.0.0/mangos@1.0.0.sha256
+      .out${sep}mangos@1.0.0${sep}mangos@1.0.0.zip
+      .out${sep}mangos@1.0.0${sep}mangos@1.0.0.zip.sha256
+      .out${sep}mangos@1.0.0${sep}mangos@1.0.0
+      .out${sep}mangos@1.0.0${sep}mangos@1.0.0.sha256
 
     """
           .trimIndent()
       )
   }
 
+  @Test
+  fun `generate annotations`(@TempDir tempDir: Path) {
+    tempDir
+      .resolve("PklProject")
+      .writeString(
+        """
+        @Unlisted
+        @Deprecated { since = "0.26.1"; message = "do not use" }
+        @ModuleInfo { minPklVersion = "0.26.0" }
+        amends "pkl:Project"
+        
+        package {
+          name = "mypackage"
+          version = "1.0.0"
+          baseUri = "package://example.com/mypackage"
+          packageZipUrl = "https://foo.com"
+        }
+      """
+          .trimIndent()
+      )
+    val packager =
+      CliProjectPackager(
+        CliBaseOptions(workingDir = tempDir),
+        listOf(tempDir),
+        CliTestOptions(),
+        ".out/%{name}@%{version}",
+        skipPublishCheck = true,
+        consoleWriter = StringWriter(),
+      )
+    packager.run()
+    val expectedMetadata = tempDir.resolve(".out/mypackage@1.0.0/mypackage@1.0.0")
+    assertThat(expectedMetadata).exists()
+    assertThat(expectedMetadata)
+      .hasContent(
+        """
+      {
+        "name": "mypackage",
+        "packageUri": "package://example.com/mypackage@1.0.0",
+        "version": "1.0.0",
+        "packageZipUrl": "https://foo.com",
+        "packageZipChecksums": {
+          "sha256": "8739c76e681f900923b900c9df0ef75cf421d39cabb54650c4b9ad19b6a76d85"
+        },
+        "dependencies": {},
+        "authors": [],
+        "annotations": [
+          {
+            "type": "PObject",
+            "classInfo": {
+              "moduleName": "pkl.base",
+              "class": "Unlisted",
+              "moduleUri": "pkl:base"
+            },
+            "properties": {}
+          },
+          {
+            "type": "PObject",
+            "classInfo": {
+              "moduleName": "pkl.base",
+              "class": "Deprecated",
+              "moduleUri": "pkl:base"
+            },
+            "properties": {
+              "since": "0.26.1",
+              "message": "do not use",
+              "replaceWith": null
+            }
+          },
+          {
+            "type": "PObject",
+            "classInfo": {
+              "moduleName": "pkl.base",
+              "class": "ModuleInfo",
+              "moduleUri": "pkl:base"
+            },
+            "properties": {
+              "minPklVersion": "0.26.0"
+            }
+          }
+        ]
+      }
+    """
+          .trimIndent()
+      )
+  }
+
   private fun Path.zipFilePaths(): List<String> {
     return FileSystems.newFileSystem(URI("jar:${toUri()}"), emptyMap<String, String>()).use { fs ->
-      Files.walk(fs.getPath("/")).map { it.toString() }.collect(Collectors.toList())
+      Files.walk(fs.getPath("/")).map(IoUtils::toNormalizedPathString).collect(Collectors.toList())
     }
   }
 }
