@@ -1,5 +1,5 @@
-/**
- * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+/*
+ * Copyright © 2024-2025 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -17,7 +17,7 @@ package org.pkl.cli
 
 import java.nio.charset.StandardCharsets
 import java.nio.file.Path
-import org.pkl.commons.createParentDirectories
+import kotlin.io.path.createParentDirectories
 import org.pkl.commons.writeString
 
 fun Path.writeFile(fileName: String, contents: String): Path {
