@@ -1,5 +1,5 @@
-/**
- * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+/*
+ * Copyright © 2024-2025 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -15,10 +15,9 @@
  */
 package org.pkl.codegen.java
 
-import com.squareup.javapoet.*
+import com.palantir.javapoet.*
 import java.io.StringWriter
 import java.lang.Deprecated
-import java.net.URI
 import java.util.*
 import java.util.regex.Pattern
 import javax.lang.model.element.Modifier
@@ -34,50 +33,77 @@ import kotlin.apply
 import kotlin.let
 import kotlin.takeIf
 import kotlin.to
+import org.pkl.commons.NameMapper
 import org.pkl.core.*
 import org.pkl.core.util.CodeGeneratorUtils
+import org.pkl.core.util.IoUtils
 
 class JavaCodeGeneratorException(message: String) : RuntimeException(message)
 
-data class JavaCodegenOptions(
+@kotlin.Deprecated("renamed to JavaCodeGeneratorOptions", ReplaceWith("JavaCodeGeneratorOptions"))
+typealias JavaCodegenOptions = JavaCodeGeneratorOptions
+
+data class JavaCodeGeneratorOptions(
   /** The characters to use for indenting generated Java code. */
   val indent: String = "  ",
 
+  /** Adds the `org.pkl.config.java.Generated` annotation to the classes to be generated. */
+  val addGeneratedAnnotation: Boolean = false,
+
   /**
    * Whether to generate public getter methods and protected final fields instead of public final
    * fields.
    */
   val generateGetters: Boolean = false,
 
-  /** Whether to generate Javadoc based on doc comments for Pkl modules, classes, and properties. */
+  /** Whether to preserve Pkl doc comments by generating corresponding Javadoc comments. */
   val generateJavadoc: Boolean = false,
 
   /** Whether to generate config classes for use with Spring Boot. */
   val generateSpringBootConfig: Boolean = false,
 
   /**
-   * Fully qualified name of the annotation to use on constructor parameters. If this options is not
-   * set, [org.pkl.config.java.mapper.Named] will be used.
+   * Fully qualified name of the annotation type to use for annotating constructor parameters with
+   * their name.
+   *
+   * The specified annotation type must have a `value` parameter of type [java.lang.String] or the
+   * generated code may not compile.
+   *
+   * If set to `null`, constructor parameters are not annotated. The default value is `null` if
+   * [generateSpringBootConfig] is `true` and `"org.pkl.config.java.mapper.Named"` otherwise.
    */
-  val paramsAnnotation: String? = null,
+  val paramsAnnotation: String? =
+    if (generateSpringBootConfig) null else "org.pkl.config.java.mapper.Named",
 
   /**
-   * Fully qualified name of the annotation to use on non-null properties. If this option is not
-   * set, [org.pkl.config.java.mapper.NonNull] will be used.
+   * Fully qualified name of the annotation type to use for annotating non-null types.
+   *
+   * The specified annotation type must have a [java.lang.annotation.Target] of
+   * [java.lang.annotation.ElementType.TYPE_USE] or the generated code may not compile. If set to
+   * `null`, [org.pkl.config.java.mapper.NonNull] will be used.
    */
   val nonNullAnnotation: String? = null,
 
-  /** Whether to make generated classes implement [java.io.Serializable] */
-  val implementSerializable: Boolean = false
+  /** Whether to generate classes that implement [java.io.Serializable]. */
+  val implementSerializable: Boolean = false,
+
+  /**
+   * A mapping from Pkl module name prefixes to their replacements.
+   *
+   * Can be used when the class or package name in the generated source code should be different
+   * from the corresponding name derived from the Pkl module declaration .
+   */
+  val renames: Map<String, String> = emptyMap(),
 )
 
 /** Entrypoint for the Java code generator API. */
 class JavaCodeGenerator(
   private val schema: ModuleSchema,
-  private val codegenOptions: JavaCodegenOptions
+  private val codegenOptions: JavaCodeGeneratorOptions,
 ) {
 
   companion object {
+    private val OBJECT = ClassName.get(Object::class.java)
     private val STRING = ClassName.get(String::class.java)
     private val DURATION = ClassName.get(Duration::class.java)
     private val DURATION_UNIT = ClassName.get(DurationUnit::class.java)
@@ -91,7 +117,7 @@ class JavaCodeGenerator(
     private val PMODULE = ClassName.get(PModule::class.java)
     private val PCLASS = ClassName.get(PClass::class.java)
     private val PATTERN = ClassName.get(Pattern::class.java)
-    private val URI = ClassName.get(URI::class.java)
+    private val URI = ClassName.get(java.net.URI::class.java)
     private val VERSION = ClassName.get(Version::class.java)
 
     private const val PROPERTY_PREFIX: String = "org.pkl.config.java.mapper."
@@ -123,7 +149,8 @@ class JavaCodeGenerator(
     }
 
   private val propertyFileName: String
-    get() = "resources/META-INF/org/pkl/config/java/mapper/classes/${schema.moduleName}.properties"
+    get() =
+      "resources/META-INF/org/pkl/config/java/mapper/classes/${IoUtils.encodePath(schema.moduleName)}.properties"
 
   private val propertiesFile: String
     get() {
@@ -150,8 +177,16 @@ class JavaCodeGenerator(
       return AnnotationSpec.builder(className).build()
     }
 
-  val javaFileName: String
-    get() = relativeOutputPathFor(schema.moduleName)
+  private val javaFileName: String
+    get() {
+      val (packageName, className) = nameMapper.map(schema.moduleName)
+      val dirPath = packageName.replace('.', '/')
+      return if (dirPath.isEmpty()) {
+        "java/$className.java"
+      } else {
+        "java/$dirPath/$className.java"
+      }
+    }
 
   val javaFile: String
     get() {
@@ -183,9 +218,7 @@ class JavaCodeGenerator(
         moduleClass.addMethod(appendPropertyMethod().addModifiers(modifier).build())
       }
 
-      val moduleName = schema.moduleName
-      val index = moduleName.lastIndexOf(".")
-      val packageName = if (index == -1) "" else moduleName.substring(0, index)
+      val (packageName, _) = nameMapper.map(schema.moduleName)
 
       return JavaFile.builder(packageName, moduleClass.build())
         .indent(codegenOptions.indent)
@@ -193,20 +226,7 @@ class JavaCodeGenerator(
         .toString()
     }
 
-  private fun relativeOutputPathFor(moduleName: String): String {
-    val moduleNameParts = moduleName.split(".")
-    val dirPath = moduleNameParts.dropLast(1).joinToString("/")
-    val fileName = moduleNameParts.last().replaceFirstChar { it.titlecaseChar() }
-    return if (dirPath.isEmpty()) {
-      "java/$fileName.java"
-    } else {
-      "java/$dirPath/$fileName.java"
-    }
-  }
-
-  @Suppress("NAME_SHADOWING")
   private fun generateTypeSpec(pClass: PClass, schema: ModuleSchema): TypeSpec.Builder {
-
     val isModuleClass = pClass == schema.moduleClass
     val javaPoetClassName = pClass.toJavaPoetName()
     val superclass =
@@ -217,31 +237,33 @@ class JavaCodeGenerator(
     val properties = renameIfReservedWord(pClass.properties).filterValues { !it.isHidden }
     val allProperties = superProperties + properties
 
+    fun PClass.Property.isRegex(): Boolean =
+      (this.type as? PType.Class)?.pClass?.info == PClassInfo.Regex
+
     fun addCtorParameter(
       builder: MethodSpec.Builder,
       propJavaName: String,
-      property: PClass.Property
+      property: PClass.Property,
     ) {
-      builder.addParameter(
-        ParameterSpec.builder(property.type.toJavaPoetName(), propJavaName)
-          .addAnnotation(
-            AnnotationSpec.builder(namedAnnotationName)
-              .addMember("value", "\$S", property.simpleName)
-              .build()
-          )
-          .build()
-      )
+      val paramBuilder = ParameterSpec.builder(property.type.toJavaPoetName(), propJavaName)
+      if (paramsAnnotationName != null) {
+        paramBuilder.addAnnotation(
+          AnnotationSpec.builder(paramsAnnotationName)
+            .addMember("value", "\$S", property.simpleName)
+            .build()
+        )
+      }
+      builder.addParameter(paramBuilder.build())
     }
 
-    fun generateConstructor(): MethodSpec {
+    fun generateConstructor(isInstantiable: Boolean): MethodSpec {
       val builder =
         MethodSpec.constructorBuilder()
           // choose most restrictive access modifier possible
           .addModifiers(
             when {
-              pClass.isAbstract -> Modifier.PROTECTED
-              allProperties.isNotEmpty() -> Modifier.PUBLIC // if `false`, has no state
-              pClass.isOpen -> Modifier.PROTECTED
+              isInstantiable -> Modifier.PUBLIC
+              pClass.isAbstract || pClass.isOpen -> Modifier.PROTECTED
               else -> Modifier.PRIVATE
             }
           )
@@ -280,14 +302,12 @@ class JavaCodeGenerator(
           .addStatement("\$T other = (\$T) obj", javaPoetClassName, javaPoetClassName)
 
       for ((propertyName, property) in allProperties) {
-        val accessor =
-          if ((property.type as? PType.Class)?.pClass?.info == PClassInfo.Regex) "\$N.pattern()"
-          else "\$N"
+        val accessor = if (property.isRegex()) "\$N.pattern()" else "\$N"
         builder.addStatement(
           "if (!\$T.equals(this.$accessor, other.$accessor)) return false",
           Objects::class.java,
           propertyName,
-          propertyName
+          propertyName,
         )
       }
 
@@ -303,11 +323,12 @@ class JavaCodeGenerator(
           .returns(Int::class.java)
           .addStatement("int result = 1")
 
-      for (propertyName in allProperties.keys) {
+      for ((propertyName, property) in allProperties) {
+        val accessor = if (property.isRegex()) "this.\$N.pattern()" else "this.\$N"
         builder.addStatement(
-          "result = 31 * result + \$T.hashCode(this.\$N)",
+          "result = 31 * result + \$T.hashCode($accessor)",
           Objects::class.java,
-          propertyName
+          propertyName,
         )
       }
 
@@ -324,13 +345,22 @@ class JavaCodeGenerator(
 
       var builderSize = 50
       val appendBuilder = CodeBlock.builder()
-      for (propertyName in allProperties.keys) {
+      for ((propertyName, property) in allProperties) {
         builderSize += 50
-        appendBuilder.addStatement(
-          "appendProperty(builder, \$S, this.\$N)",
-          propertyName,
-          propertyName
-        )
+        if (property.type.isBytesClass) {
+          appendBuilder.addStatement(
+            "appendProperty(builder, \$S, \$T.toString(this.\$N))",
+            propertyName,
+            Arrays::class.java,
+            propertyName,
+          )
+        } else {
+          appendBuilder.addStatement(
+            "appendProperty(builder, \$S, this.\$N)",
+            propertyName,
+            propertyName,
+          )
+        }
       }
 
       builder
@@ -338,7 +368,7 @@ class JavaCodeGenerator(
           "\$T builder = new \$T(\$L)",
           StringBuilder::class.java,
           StringBuilder::class.java,
-          builderSize
+          builderSize,
         )
         .addStatement("builder.append(\$T.class.getSimpleName()).append(\" {\")", javaPoetClassName)
         .addCode(appendBuilder.build())
@@ -362,7 +392,7 @@ class JavaCodeGenerator(
       annotations: Collection<PObject>,
       hasJavadoc: Boolean,
       addAnnotation: (Class<*>) -> Unit,
-      addJavadoc: (String) -> Unit
+      addJavadoc: (String) -> Unit,
     ) {
       annotations
         .firstOrNull { it.classInfo == PClassInfo.Deprecated }
@@ -398,7 +428,7 @@ class JavaCodeGenerator(
           property.annotations,
           hasJavadoc,
           { builder.addAnnotation(it) },
-          { builder.addJavadoc(it) }
+          { builder.addJavadoc(it) },
         )
         builder.addModifiers(Modifier.PUBLIC)
       }
@@ -411,7 +441,7 @@ class JavaCodeGenerator(
     fun generateGetter(
       propertyName: String,
       property: PClass.Property,
-      isOverridden: Boolean
+      isOverridden: Boolean,
     ): MethodSpec {
       val propertyType = property.type
       val isBooleanProperty =
@@ -441,7 +471,7 @@ class JavaCodeGenerator(
         property.annotations,
         hasJavadoc,
         { builder.addAnnotation(it) },
-        { builder.addJavadoc(it) }
+        { builder.addJavadoc(it) },
       )
 
       return builder.build()
@@ -460,7 +490,7 @@ class JavaCodeGenerator(
         property.annotations,
         false,
         { methodBuilder.addAnnotation(it) },
-        { methodBuilder.addJavadoc(it) }
+        { methodBuilder.addJavadoc(it) },
       )
 
       val codeBuilder = CodeBlock.builder()
@@ -490,10 +520,6 @@ class JavaCodeGenerator(
     }
 
     fun generateSpringBootAnnotations(builder: TypeSpec.Builder) {
-      builder.addAnnotation(
-        ClassName.get("org.springframework.boot.context.properties", "ConstructorBinding")
-      )
-
       if (isModuleClass) {
         builder.addAnnotation(
           ClassName.get("org.springframework.boot.context.properties", "ConfigurationProperties")
@@ -520,7 +546,7 @@ class JavaCodeGenerator(
             AnnotationSpec.builder(
                 ClassName.get(
                   "org.springframework.boot.context.properties",
-                  "ConfigurationProperties"
+                  "ConfigurationProperties",
                 )
               )
               // use "value" instead of "prefix" to entice JavaPoet to generate a single-line
@@ -537,8 +563,17 @@ class JavaCodeGenerator(
     fun generateClass(): TypeSpec.Builder {
       val builder =
         TypeSpec.classBuilder(javaPoetClassName.simpleName()).addModifiers(Modifier.PUBLIC)
+      if (codegenOptions.addGeneratedAnnotation) {
+        val name = ClassName.get("org.pkl.config.java", "Generated")
+        val generated = AnnotationSpec.builder(name).build()
+        builder.addAnnotation(generated)
+      }
 
-      if (codegenOptions.implementSerializable && !isModuleClass) {
+      // stateless final module classes are non-instantiable by choice
+      val isInstantiable =
+        !(pClass.isAbstract || (isModuleClass && !pClass.isOpen && allProperties.isEmpty()))
+
+      if (codegenOptions.implementSerializable && isInstantiable) {
         builder.addSuperinterface(java.io.Serializable::class.java)
         builder.addField(generateSerialVersionUIDField())
       }
@@ -553,7 +588,7 @@ class JavaCodeGenerator(
         pClass.annotations,
         hasJavadoc,
         { builder.addAnnotation(it) },
-        { builder.addJavadoc(it) }
+        { builder.addJavadoc(it) },
       )
 
       if (!isModuleClass) {
@@ -570,7 +605,7 @@ class JavaCodeGenerator(
         generateSpringBootAnnotations(builder)
       }
 
-      builder.addMethod(generateConstructor())
+      builder.addMethod(generateConstructor(isInstantiable))
 
       superclass?.let { builder.superclass(it.toJavaPoetName()) }
 
@@ -586,12 +621,12 @@ class JavaCodeGenerator(
             builder.addMethod(generateGetter(name, property, isOverridden))
           }
         }
-        if (!pClass.isAbstract) {
+        if (isInstantiable) {
           builder.addMethod(generateWithMethod(name, property))
         }
       }
 
-      if (properties.isNotEmpty()) {
+      if (isInstantiable) {
         builder
           .addMethod(generateEqualsMethod())
           .addMethod(generateHashCodeMethod())
@@ -613,7 +648,7 @@ class JavaCodeGenerator(
 
   private fun generateEnumTypeSpec(
     typeAlias: TypeAlias,
-    stringLiterals: Set<String>
+    stringLiterals: Set<String>,
   ): TypeSpec.Builder {
     val enumConstantToPklNames =
       stringLiterals
@@ -655,19 +690,15 @@ class JavaCodeGenerator(
     for ((enumConstantName, pklName) in enumConstantToPklNames) {
       builder.addEnumConstant(
         enumConstantName,
-        TypeSpec.anonymousClassBuilder("\$S", pklName).build()
+        TypeSpec.anonymousClassBuilder("\$S", pklName).build(),
       )
     }
 
     return builder
   }
 
-  private val namedAnnotationName =
-    if (codegenOptions.paramsAnnotation != null) {
-      toClassName(codegenOptions.paramsAnnotation)
-    } else {
-      ClassName.get("org.pkl.config.java.mapper", "Named")
-    }
+  private val paramsAnnotationName: ClassName? =
+    codegenOptions.paramsAnnotation?.let { toClassName(it) }
 
   private fun appendPropertyMethod() =
     MethodSpec.methodBuilder("appendProperty")
@@ -679,7 +710,7 @@ class JavaCodeGenerator(
       .addStatement(
         "\$T lines = \$T.toString(value).split(\"\\n\")",
         ArrayTypeName.of(String::class.java),
-        Objects::class.java
+        Objects::class.java,
       )
       .addStatement("builder.append(lines[0])")
       .beginControlFlow("for (int i = 1; i < lines.length; i++)")
@@ -687,9 +718,7 @@ class JavaCodeGenerator(
       .endControlFlow()
 
   private fun PClass.toJavaPoetName(): ClassName {
-    val index = moduleName.lastIndexOf(".")
-    val packageName = if (index == -1) "" else moduleName.substring(0, index)
-    val moduleClassName = moduleName.substring(index + 1).replaceFirstChar { it.titlecaseChar() }
+    val (packageName, moduleClassName) = nameMapper.map(moduleName)
     return if (isModuleClass) {
       ClassName.get(packageName, moduleClassName)
     } else {
@@ -699,15 +728,13 @@ class JavaCodeGenerator(
 
   // generated type is a nested enum class
   private fun TypeAlias.toJavaPoetName(): ClassName {
-    val index = moduleName.lastIndexOf(".")
-    val packageName = if (index == -1) "" else moduleName.substring(0, index)
-    val moduleClassName = moduleName.substring(index + 1).replaceFirstChar { it.titlecaseChar() }
+    val (packageName, moduleClassName) = nameMapper.map(moduleName)
     return ClassName.get(packageName, moduleClassName, simpleName)
   }
 
   /** Generate `List<? extends Foo>` if `Foo` is `abstract` or `open`, to allow subclassing. */
   private fun PType.toJavaPoetTypeArgumentName(): TypeName {
-    val baseName = toJavaPoetName(nullable = false, boxed = true)
+    val baseName = toJavaPoetName(boxed = true)
     return if (this is PType.Class && (pClass.isAbstract || pClass.isOpen)) {
       WildcardTypeName.subtypeOf(baseName)
     } else {
@@ -715,17 +742,20 @@ class JavaCodeGenerator(
     }
   }
 
+  private val PType.isBytesClass: Boolean
+    get() = this is PType.Class && this.pClass.info == PClassInfo.Bytes
+
   private fun PType.toJavaPoetName(nullable: Boolean = false, boxed: Boolean = false): TypeName =
     when (this) {
-      PType.UNKNOWN -> TypeName.OBJECT.nullableIf(nullable)
+      PType.UNKNOWN -> OBJECT.nullableIf(nullable)
       PType.NOTHING -> TypeName.VOID
       is PType.StringLiteral -> STRING.nullableIf(nullable)
       is PType.Class -> {
         // if in doubt, spell it out
         when (val classInfo = pClass.info) {
-          PClassInfo.Any -> TypeName.OBJECT
+          PClassInfo.Any -> OBJECT
           PClassInfo.Typed,
-          PClassInfo.Dynamic -> TypeName.OBJECT.nullableIf(nullable)
+          PClassInfo.Dynamic -> OBJECT.nullableIf(nullable)
           PClassInfo.Boolean -> TypeName.BOOLEAN.boxIf(boxed).nullableIf(nullable)
           PClassInfo.String -> STRING.nullableIf(nullable)
           // seems more useful to generate `double` than `java.lang.Number`
@@ -734,29 +764,30 @@ class JavaCodeGenerator(
           PClassInfo.Float -> TypeName.DOUBLE.boxIf(boxed).nullableIf(nullable)
           PClassInfo.Duration -> DURATION.nullableIf(nullable)
           PClassInfo.DataSize -> DATA_SIZE.nullableIf(nullable)
+          PClassInfo.Bytes -> ArrayTypeName.of(TypeName.BYTE)
           PClassInfo.Pair ->
             ParameterizedTypeName.get(
                 PAIR,
                 if (typeArguments.isEmpty()) {
-                  TypeName.OBJECT
+                  OBJECT
                 } else {
                   typeArguments[0].toJavaPoetTypeArgumentName()
                 },
                 if (typeArguments.isEmpty()) {
-                  TypeName.OBJECT
+                  OBJECT
                 } else {
                   typeArguments[1].toJavaPoetTypeArgumentName()
-                }
+                },
               )
               .nullableIf(nullable)
           PClassInfo.Collection ->
             ParameterizedTypeName.get(
                 COLLECTION,
                 if (typeArguments.isEmpty()) {
-                  TypeName.OBJECT
+                  OBJECT
                 } else {
                   typeArguments[0].toJavaPoetTypeArgumentName()
-                }
+                },
               )
               .nullableIf(nullable)
           PClassInfo.List,
@@ -764,10 +795,10 @@ class JavaCodeGenerator(
             ParameterizedTypeName.get(
                 LIST,
                 if (typeArguments.isEmpty()) {
-                  TypeName.OBJECT
+                  OBJECT
                 } else {
                   typeArguments[0].toJavaPoetTypeArgumentName()
-                }
+                },
               )
               .nullableIf(nullable)
           }
@@ -775,10 +806,10 @@ class JavaCodeGenerator(
             ParameterizedTypeName.get(
                 SET,
                 if (typeArguments.isEmpty()) {
-                  TypeName.OBJECT
+                  OBJECT
                 } else {
                   typeArguments[0].toJavaPoetTypeArgumentName()
-                }
+                },
               )
               .nullableIf(nullable)
           PClassInfo.Map,
@@ -786,15 +817,15 @@ class JavaCodeGenerator(
             ParameterizedTypeName.get(
                 MAP,
                 if (typeArguments.isEmpty()) {
-                  TypeName.OBJECT
+                  OBJECT
                 } else {
                   typeArguments[0].toJavaPoetTypeArgumentName()
                 },
                 if (typeArguments.isEmpty()) {
-                  TypeName.OBJECT
+                  OBJECT
                 } else {
                   typeArguments[1].toJavaPoetTypeArgumentName()
-                }
+                },
               )
               .nullableIf(nullable)
           PClassInfo.Module -> PMODULE.nullableIf(nullable)
@@ -816,7 +847,7 @@ class JavaCodeGenerator(
       is PType.Constrained -> baseType.toJavaPoetName(nullable = nullable, boxed = boxed)
       is PType.Alias ->
         when (typeAlias.qualifiedName) {
-          "pkl.base#NonNull" -> TypeName.OBJECT.nullableIf(nullable)
+          "pkl.base#NonNull" -> OBJECT.nullableIf(nullable)
           "pkl.base#Int8" -> TypeName.BYTE.boxIf(boxed).nullableIf(nullable)
           "pkl.base#Int16",
           "pkl.base#UInt8" -> TypeName.SHORT.boxIf(boxed).nullableIf(nullable)
@@ -872,6 +903,8 @@ class JavaCodeGenerator(
       } else key
     }
   }
+
+  private val nameMapper = NameMapper(codegenOptions.renames)
 }
 
 internal val javaReservedWords =
@@ -929,5 +962,5 @@ internal val javaReservedWords =
     "try",
     "void",
     "volatile",
-    "while"
+    "while",
   )
