@@ -1,5 +1,5 @@
-/**
- * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+/*
+ * Copyright © 2024-2025 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -16,7 +16,7 @@
 package org.pkl.codegen.kotlin
 
 import java.nio.file.Path
-import org.pkl.commons.createParentDirectories
+import kotlin.io.path.createParentDirectories
 import org.pkl.commons.writeString
 
 data class PklModule(val name: String, val content: String) {
