@@ -1,5 +1,5 @@
-/**
- * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+/*
+ * Copyright © 2024-2025 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -15,22 +15,23 @@
  */
 package org.pkl.commons.cli.commands
 
+import com.github.ajalt.clikt.completion.CompletionCandidates
 import com.github.ajalt.clikt.parameters.arguments.argument
 import com.github.ajalt.clikt.parameters.arguments.convert
 import com.github.ajalt.clikt.parameters.arguments.multiple
 import com.github.ajalt.clikt.parameters.groups.provideDelegate
 import java.net.URI
 
-abstract class ModulesCommand(name: String, helpLink: String, help: String = "") :
-  BaseCommand(
-    name = name,
-    help = help,
-    helpLink = helpLink,
-  ) {
+abstract class ModulesCommand(name: String, helpLink: String) :
+  BaseCommand(name = name, helpLink = helpLink) {
   open val modules: List<URI> by
-    argument(name = "<modules>", help = "Module paths or URIs to evaluate.")
-      .convert { parseModuleName(it) }
+    argument(
+        name = "modules",
+        help = "Module paths or URIs to evaluate.",
+        completionCandidates = CompletionCandidates.Path,
+      )
+      .convert { BaseOptions.parseModuleName(it) }
       .multiple(required = true)
 
-  protected val projectOptions by ProjectOptions()
+  protected val projectOptions: ProjectOptions by ProjectOptions()
 }
