@@ -1,5 +1,5 @@
-/**
- * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+/*
+ * Copyright © 2024-2025 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -16,6 +16,7 @@
 package org.pkl.commons.cli.commands
 
 import com.github.ajalt.clikt.parameters.groups.OptionGroup
+import com.github.ajalt.clikt.parameters.options.default
 import com.github.ajalt.clikt.parameters.options.flag
 import com.github.ajalt.clikt.parameters.options.option
 import com.github.ajalt.clikt.parameters.types.path
@@ -26,13 +27,31 @@ class TestOptions : OptionGroup() {
   private val junitReportDir: Path? by
     option(
         names = arrayOf("--junit-reports"),
-        metavar = "<dir>",
-        help = "Directory where to store JUnit reports."
+        metavar = "dir",
+        help = "Directory where to store JUnit reports.",
       )
       .path()
 
+  private val junitAggregateReports: Boolean by
+    option(
+        names = arrayOf("--junit-aggregate-reports"),
+        help = "Aggregate JUnit reports into a single file.",
+      )
+      .flag()
+
+  private val junitAggregateSuiteName: String by
+    option(
+        names = arrayOf("--junit-aggregate-suite-name"),
+        metavar = "name",
+        help = "The name of the root JUnit test suite.",
+      )
+      .single()
+      .default("pkl-tests")
+
   private val overwrite: Boolean by
     option(names = arrayOf("--overwrite"), help = "Force generation of expected examples.").flag()
 
-  val cliTestOptions: CliTestOptions by lazy { CliTestOptions(junitReportDir, overwrite) }
+  val cliTestOptions: CliTestOptions by lazy {
+    CliTestOptions(junitReportDir, overwrite, junitAggregateReports, junitAggregateSuiteName)
+  }
 }
