@@ -1,4 +1,4 @@
-/**
+/*
  * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
@@ -23,7 +23,7 @@ final class PNullToAny implements ConverterFactory {
   @Override
   public Optional<Converter<?, ?>> create(PClassInfo<?> sourceType, Type targetType) {
     if (sourceType != PClassInfo.Null
-        || (targetType instanceof Class && ((Class<?>) targetType).isPrimitive())) {
+        || (targetType instanceof Class<?> clazz && clazz.isPrimitive())) {
       return Optional.empty();
     }
 
