@@ -1,4 +1,4 @@
-/**
+/*
  * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
@@ -29,9 +29,7 @@ public class Person {
   @Override
   public boolean equals(@Nullable Object obj) {
     if (this == obj) return true;
-    if (!(obj instanceof Person)) return false;
-
-    var other = (Person) obj;
+    if (!(obj instanceof Person other)) return false;
     return name.equals(other.name) && age == other.age;
   }
 
