@@ -1,5 +1,5 @@
-/**
- * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+/*
+ * Copyright © 2024-2025 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -20,6 +20,10 @@ import java.nio.file.Path;
 import java.util.*;
 import java.util.regex.Pattern;
 import org.pkl.core.SecurityManagers.StandardBuilder;
+import org.pkl.core.evaluatorSettings.PklEvaluatorSettings.ExternalReader;
+import org.pkl.core.evaluatorSettings.TraceMode;
+import org.pkl.core.externalreader.ExternalReaderProcess;
+import org.pkl.core.http.HttpClient;
 import org.pkl.core.module.ModuleKeyFactories;
 import org.pkl.core.module.ModuleKeyFactory;
 import org.pkl.core.module.ModulePathResolver;
@@ -38,6 +42,10 @@ public final class EvaluatorBuilder {
 
   private @Nullable SecurityManager securityManager;
 
+  // Default to a client with a fixed set of built-in certificates.
+  // Make it lazy to avoid creating a client unnecessarily.
+  private HttpClient httpClient = HttpClient.builder().buildLazily();
+
   private Logger logger = Loggers.noop();
 
   private final List<ModuleKeyFactory> moduleKeyFactories = new ArrayList<>();
@@ -48,16 +56,20 @@ public final class EvaluatorBuilder {
 
   private final Map<String, String> externalProperties = new HashMap<>();
 
-  private java.time.@Nullable Duration timeout;
+  private @Nullable java.time.Duration timeout;
 
   private @Nullable Path moduleCacheDir = IoUtils.getDefaultModuleCacheDir();
 
   private @Nullable String outputFormat;
 
+  private boolean color = false;
+
   private @Nullable StackFrameTransformer stackFrameTransformer;
 
   private @Nullable DeclaredDependencies dependencies;
 
+  private TraceMode traceMode = TraceMode.COMPACT;
+
   private EvaluatorBuilder() {}
 
   /**
@@ -82,6 +94,7 @@ public final class EvaluatorBuilder {
    *   <li>{@link ResourceReaders#https}
    *   <li>{@link ResourceReaders#pkg}
    *   <li>{@link ResourceReaders#projectpackage}
+   *   <li>{@link ResourceReaders#fromServiceProviders}
    *   <li>{@link System#getProperties}
    * </ul>
    */
@@ -99,6 +112,7 @@ public final class EvaluatorBuilder {
         .addResourceReader(ResourceReaders.https())
         .addResourceReader(ResourceReaders.pkg())
         .addResourceReader(ResourceReaders.projectpackage())
+        .addResourceReaders(ResourceReaders.fromServiceProviders())
         .addModuleKeyFactory(ModuleKeyFactories.standardLibrary);
 
     if (!TruffleOptions.AOT) {
@@ -115,6 +129,7 @@ public final class EvaluatorBuilder {
     builder
         .addModuleKeyFactories(ModuleKeyFactories.fromServiceProviders())
         .addModuleKeyFactory(ModuleKeyFactories.file)
+        .addModuleKeyFactory(ModuleKeyFactories.http)
         .addModuleKeyFactory(ModuleKeyFactories.pkg)
         .addModuleKeyFactory(ModuleKeyFactories.projectpackage)
         .addModuleKeyFactory(ModuleKeyFactories.genericUrl)
@@ -136,6 +151,17 @@ public final class EvaluatorBuilder {
     return new EvaluatorBuilder();
   }
 
+  /** Sets the option to render errors in ANSI color. */
+  public EvaluatorBuilder setColor(boolean color) {
+    this.color = color;
+    return this;
+  }
+
+  /** Returns the current setting of the option to render errors in ANSI color. */
+  public boolean getColor() {
+    return color;
+  }
+
   /** Sets the given stack frame transformer, replacing any previously set transformer. */
   public EvaluatorBuilder setStackFrameTransformer(StackFrameTransformer stackFrameTransformer) {
     this.stackFrameTransformer = stackFrameTransformer;
@@ -226,6 +252,21 @@ public final class EvaluatorBuilder {
     return logger;
   }
 
+  /**
+   * Sets the HTTP client to be used.
+   *
+   * <p>Defaults to {@code HttpClient.builder().buildLazily()}.
+   */
+  public EvaluatorBuilder setHttpClient(HttpClient httpClient) {
+    this.httpClient = httpClient;
+    return this;
+  }
+
+  /** Returns the currently set HTTP client. */
+  public HttpClient getHttpClient() {
+    return httpClient;
+  }
+
   /**
    * Adds the given module key factory. Factories will be asked to resolve module keys in the order
    * they have been added to this builder.
@@ -343,7 +384,7 @@ public final class EvaluatorBuilder {
   /**
    * Sets an evaluation timeout to be enforced by the {@link Evaluator}'s {@code evaluate} methods.
    */
-  public EvaluatorBuilder setTimeout(java.time.@Nullable Duration timeout) {
+  public EvaluatorBuilder setTimeout(@Nullable java.time.Duration timeout) {
     this.timeout = timeout;
     return this;
   }
@@ -412,6 +453,21 @@ public final class EvaluatorBuilder {
     return this;
   }
 
+  public @Nullable DeclaredDependencies getProjectDependencies() {
+    return this.dependencies;
+  }
+
+  /** Sets whether calls to trace() produce indented, multi-line output. */
+  public EvaluatorBuilder setTraceMode(TraceMode traceMode) {
+    this.traceMode = traceMode;
+    return this;
+  }
+
+  /** Returns whether calls to trace() produce indented, multi-line output. */
+  public TraceMode getTraceMode() {
+    return this.traceMode;
+  }
+
   /**
    * Given a project, sets its dependencies, and also applies any evaluator settings if set.
    *
@@ -419,40 +475,82 @@ public final class EvaluatorBuilder {
    */
   public EvaluatorBuilder applyFromProject(Project project) {
     this.dependencies = project.getDependencies();
-    var settings = project.getSettings();
+    var settings = project.getEvaluatorSettings();
     if (securityManager != null) {
       throw new IllegalStateException(
           "Cannot call both `setSecurityManager` and `setProject`, because both define security manager settings. Call `setProjectOnly` if the security manager is desired.");
     }
-    if (settings.getAllowedModules() != null) {
-      setAllowedModules(settings.getAllowedModules());
+    if (settings.allowedModules() != null) {
+      setAllowedModules(settings.allowedModules());
     }
-    if (settings.getAllowedResources() != null) {
-      setAllowedResources(settings.getAllowedResources());
+    if (settings.allowedResources() != null) {
+      setAllowedResources(settings.allowedResources());
     }
-    if (settings.getExternalProperties() != null) {
-      setExternalProperties(settings.getExternalProperties());
+    if (settings.externalProperties() != null) {
+      setExternalProperties(settings.externalProperties());
     }
-    if (settings.getEnv() != null) {
-      setEnvironmentVariables(settings.getEnv());
+    if (settings.env() != null) {
+      setEnvironmentVariables(settings.env());
     }
-    if (settings.getTimeout() != null) {
-      setTimeout(settings.getTimeout().toJavaDuration());
+    if (settings.timeout() != null) {
+      setTimeout(settings.timeout().toJavaDuration());
     }
-    if (settings.getModulePath() != null) {
+    if (settings.modulePath() != null) {
       // indirectly closed by `ModuleKeyFactories.closeQuietly(builder.moduleKeyFactories)`
-      var modulePathResolver = new ModulePathResolver(settings.getModulePath());
+      var modulePathResolver = new ModulePathResolver(settings.modulePath());
       addResourceReader(ResourceReaders.modulePath(modulePathResolver));
       addModuleKeyFactory(ModuleKeyFactories.modulePath(modulePathResolver));
     }
-    if (settings.getRootDir() != null) {
-      setRootDir(settings.getRootDir());
+    if (settings.rootDir() != null) {
+      setRootDir(settings.rootDir());
     }
-    if (Boolean.TRUE.equals(settings.isNoCache())) {
+    if (settings.color() != null) {
+      setColor(settings.color().hasColor());
+    }
+    if (Boolean.TRUE.equals(settings.noCache())) {
       setModuleCacheDir(null);
-    } else if (settings.getModuleCacheDir() != null) {
-      setModuleCacheDir(settings.getModuleCacheDir());
+    } else if (settings.moduleCacheDir() != null) {
+      setModuleCacheDir(settings.moduleCacheDir());
     }
+
+    // this isn't ideal as project and non-project ExternalReaderProcess instances can be dupes
+    var procs = new HashMap<ExternalReader, ExternalReaderProcess>();
+    if (settings.externalModuleReaders() != null) {
+      for (var entry : settings.externalModuleReaders().entrySet()) {
+        addModuleKeyFactory(
+            ModuleKeyFactories.externalProcess(
+                entry.getKey(),
+                procs.computeIfAbsent(entry.getValue(), ExternalReaderProcess::of)));
+      }
+    }
+    if (settings.externalResourceReaders() != null) {
+      for (var entry : settings.externalResourceReaders().entrySet()) {
+        addResourceReader(
+            ResourceReaders.externalProcess(
+                entry.getKey(),
+                procs.computeIfAbsent(entry.getValue(), ExternalReaderProcess::of)));
+      }
+    }
+
+    if (settings.http() != null) {
+      var httpClientBuilder = HttpClient.builder();
+      if (settings.http().proxy() != null) {
+        var noProxy = settings.http().proxy().noProxy();
+        if (noProxy == null) {
+          noProxy = Collections.emptyList();
+        }
+        httpClientBuilder.setProxy(settings.http().proxy().address(), noProxy);
+      }
+      if (settings.http().rewrites() != null) {
+        httpClientBuilder.setRewrites(settings.http().rewrites());
+      }
+      setHttpClient(httpClientBuilder.buildLazily());
+    }
+
+    if (settings.traceMode() != null) {
+      setTraceMode(settings.traceMode());
+    }
+
     return this;
   }
 
@@ -467,7 +565,9 @@ public final class EvaluatorBuilder {
 
     return new EvaluatorImpl(
         stackFrameTransformer,
+        color,
         securityManager,
+        httpClient,
         new LoggerImpl(logger, stackFrameTransformer),
         // copy to shield against subsequent modification through builder
         new ArrayList<>(moduleKeyFactories),
@@ -477,6 +577,7 @@ public final class EvaluatorBuilder {
         timeout,
         moduleCacheDir,
         dependencies,
-        outputFormat);
+        outputFormat,
+        traceMode);
   }
 }
