@@ -1,5 +1,5 @@
-/**
- * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+/*
+ * Copyright © 2024-2025 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -28,8 +28,12 @@ import java.util.concurrent.TimeUnit;
 import java.util.function.Function;
 import java.util.function.Supplier;
 import org.graalvm.polyglot.Context;
+import org.msgpack.core.MessageBufferPacker;
+import org.msgpack.core.MessagePack;
 import org.pkl.core.ast.ConstantValueNode;
 import org.pkl.core.ast.internal.ToStringNodeGen;
+import org.pkl.core.evaluatorSettings.TraceMode;
+import org.pkl.core.http.HttpClient;
 import org.pkl.core.module.ModuleKeyFactory;
 import org.pkl.core.module.ProjectDependenciesManager;
 import org.pkl.core.packages.PackageResolver;
@@ -39,7 +43,6 @@ import org.pkl.core.runtime.BaseModule;
 import org.pkl.core.runtime.Identifier;
 import org.pkl.core.runtime.ModuleResolver;
 import org.pkl.core.runtime.ResourceManager;
-import org.pkl.core.runtime.TestResults;
 import org.pkl.core.runtime.TestRunner;
 import org.pkl.core.runtime.VmContext;
 import org.pkl.core.runtime.VmException;
@@ -47,6 +50,7 @@ import org.pkl.core.runtime.VmExceptionBuilder;
 import org.pkl.core.runtime.VmLanguage;
 import org.pkl.core.runtime.VmMapping;
 import org.pkl.core.runtime.VmNull;
+import org.pkl.core.runtime.VmPklBinaryEncoder;
 import org.pkl.core.runtime.VmStackOverflowException;
 import org.pkl.core.runtime.VmTyped;
 import org.pkl.core.runtime.VmUtils;
@@ -55,20 +59,24 @@ import org.pkl.core.runtime.VmValueRenderer;
 import org.pkl.core.util.ErrorMessages;
 import org.pkl.core.util.Nullable;
 
-public class EvaluatorImpl implements Evaluator {
-  protected final StackFrameTransformer frameTransformer;
-  protected final ModuleResolver moduleResolver;
-  protected final Context polyglotContext;
-  protected final @Nullable Duration timeout;
-  protected final @Nullable ScheduledExecutorService timeoutExecutor;
-  protected final SecurityManager securityManager;
-  protected final BufferedLogger logger;
-  protected final PackageResolver packageResolver;
+public final class EvaluatorImpl implements Evaluator {
+  private final StackFrameTransformer frameTransformer;
+  private final boolean color;
+  private final ModuleResolver moduleResolver;
+  private final Context polyglotContext;
+  private final @Nullable Duration timeout;
+  private final @Nullable ScheduledExecutorService timeoutExecutor;
+  private final SecurityManager securityManager;
+  private final BufferedLogger logger;
+  private final PackageResolver packageResolver;
   private final VmValueRenderer vmValueRenderer = VmValueRenderer.singleLine(1000);
+  private @Nullable MessageBufferPacker messagePacker;
 
   public EvaluatorImpl(
       StackFrameTransformer transformer,
+      boolean color,
       SecurityManager manager,
+      HttpClient httpClient,
       Logger logger,
       Collection<ModuleKeyFactory> factories,
       Collection<ResourceReader> readers,
@@ -77,13 +85,15 @@ public class EvaluatorImpl implements Evaluator {
       @Nullable Duration timeout,
       @Nullable Path moduleCacheDir,
       @Nullable DeclaredDependencies projectDependencies,
-      @Nullable String outputFormat) {
+      @Nullable String outputFormat,
+      TraceMode traceMode) {
 
     securityManager = manager;
     frameTransformer = transformer;
+    this.color = color;
     moduleResolver = new ModuleResolver(factories);
     this.logger = new BufferedLogger(logger);
-    packageResolver = PackageResolver.getInstance(securityManager, moduleCacheDir);
+    packageResolver = PackageResolver.getInstance(securityManager, httpClient, moduleCacheDir);
     polyglotContext =
         VmUtils.createContext(
             () -> {
@@ -92,6 +102,7 @@ public class EvaluatorImpl implements Evaluator {
                   new VmContext.Holder(
                       transformer,
                       manager,
+                      httpClient,
                       moduleResolver,
                       new ResourceManager(manager, readers),
                       this.logger,
@@ -102,7 +113,9 @@ public class EvaluatorImpl implements Evaluator {
                       packageResolver,
                       projectDependencies == null
                           ? null
-                          : new ProjectDependenciesManager(projectDependencies)));
+                          : new ProjectDependenciesManager(
+                              projectDependencies, moduleResolver, securityManager),
+                      traceMode));
             });
     this.timeout = timeout;
     // NOTE: would probably make sense to share executor between evaluators
@@ -133,20 +146,29 @@ public class EvaluatorImpl implements Evaluator {
     return doEvaluate(
         moduleSource,
         (module) -> {
-          var output = (VmTyped) VmUtils.readMember(module, Identifier.OUTPUT);
+          var output = readModuleOutput(module);
           return VmUtils.readTextProperty(output);
         });
   }
 
+  public byte[] evaluateOutputBytes(ModuleSource moduleSource) {
+    return doEvaluate(
+        moduleSource,
+        (module) -> {
+          var output = readModuleOutput(module);
+          var vmBytes = VmUtils.readBytesProperty(output);
+          return vmBytes.export();
+        });
+  }
+
   @Override
   public Object evaluateOutputValue(ModuleSource moduleSource) {
     return doEvaluate(
         moduleSource,
         (module) -> {
-          var output = (VmTyped) VmUtils.readMember(module, Identifier.OUTPUT);
+          var output = readModuleOutput(module);
           var value = VmUtils.readMember(output, Identifier.VALUE);
-          if (value instanceof VmValue) {
-            var vmValue = (VmValue) value;
+          if (value instanceof VmValue vmValue) {
             vmValue.force(false);
             return vmValue.export();
           }
@@ -159,7 +181,7 @@ public class EvaluatorImpl implements Evaluator {
     return doEvaluate(
         moduleSource,
         (module) -> {
-          var output = (VmTyped) VmUtils.readMember(module, Identifier.OUTPUT);
+          var output = readModuleOutput(module);
           var filesOrNull = VmUtils.readMember(output, Identifier.FILES);
           if (filesOrNull instanceof VmNull) {
             return Map.of();
@@ -178,33 +200,62 @@ public class EvaluatorImpl implements Evaluator {
 
   @Override
   public Object evaluateExpression(ModuleSource moduleSource, String expression) {
-    // optimization: if the expression is `output.text` or `output.value` (the common cases), read
-    // members directly instead of creating new truffle nodes.
-    if (expression.equals("output.text")) {
-      return evaluateOutputText(moduleSource);
-    }
-    if (expression.equals("output.value")) {
-      return evaluateOutputValue(moduleSource);
+    // optimization: if the expression is `output.text`, `output.value` or `output.bytes` (the
+    // common cases), read members directly instead of creating new truffle nodes.
+    return switch (expression) {
+      case "output.text" -> evaluateOutputText(moduleSource);
+      case "output.value" -> evaluateOutputValue(moduleSource);
+      case "output.bytes" -> evaluateOutputBytes(moduleSource);
+      default ->
+          doEvaluate(
+              moduleSource,
+              (module) -> {
+                var expressionResult =
+                    VmUtils.evaluateExpression(module, expression, securityManager, moduleResolver);
+                if (expressionResult instanceof VmValue value) {
+                  value.force(false);
+                  return value.export();
+                }
+                return expressionResult;
+              });
+    };
+  }
+
+  private MessageBufferPacker getMessagePacker() {
+    if (messagePacker == null) {
+      messagePacker = MessagePack.newDefaultBufferPacker();
     }
+    messagePacker.clear();
+    return messagePacker;
+  }
+
+  @Override
+  public byte[] evaluateExpressionPklBinary(ModuleSource moduleSource, String expression) {
     return doEvaluate(
         moduleSource,
         (module) -> {
           var expressionResult =
-              VmUtils.evaluateExpression(module, expression, securityManager, moduleResolver);
-          if (expressionResult instanceof VmValue) {
-            var value = (VmValue) expressionResult;
-            value.force(false);
-            return value.export();
-          }
-          return expressionResult;
+              switch (expression) {
+                case "module" -> module;
+                case "output.text" -> VmUtils.readTextProperty(readModuleOutput(module));
+                case "output.value" ->
+                    VmUtils.readMember(readModuleOutput(module), Identifier.VALUE);
+                case "output.bytes" -> VmUtils.readBytesProperty(readModuleOutput(module));
+                default ->
+                    VmUtils.evaluateExpression(module, expression, securityManager, moduleResolver);
+              };
+          VmValue.force(expressionResult, false);
+
+          var packer = getMessagePacker();
+          new VmPklBinaryEncoder(packer).renderDocument(expressionResult);
+          return packer.toByteArray();
         });
   }
 
   @Override
   public String evaluateExpressionString(ModuleSource moduleSource, String expression) {
     // optimization: if the expression is `output.text` (the common case), read members
-    // directly
-    // instead of creating new truffle nodes.
+    // directly instead of creating new truffle nodes.
     if (expression.equals("output.text")) {
       return evaluateOutputText(moduleSource);
     }
@@ -231,7 +282,7 @@ public class EvaluatorImpl implements Evaluator {
     return doEvaluate(
         moduleSource,
         (module) -> {
-          var testRunner = new TestRunner(logger, frameTransformer, overwrite);
+          var testRunner = new TestRunner(logger, frameTransformer, overwrite, color);
           return testRunner.run(module);
         });
   }
@@ -241,12 +292,11 @@ public class EvaluatorImpl implements Evaluator {
     return doEvaluate(
         moduleSource,
         (module) -> {
-          var output = (VmTyped) VmUtils.readMember(module, Identifier.OUTPUT);
+          var output = readModuleOutput(module);
           var value = VmUtils.readMember(output, Identifier.VALUE);
           var valueClassInfo = VmUtils.getClass(value).getPClassInfo();
           if (valueClassInfo.equals(classInfo)) {
-            if (value instanceof VmValue) {
-              var vmValue = (VmValue) value;
+            if (value instanceof VmValue vmValue) {
               vmValue.force(false);
               //noinspection unchecked
               return (T) vmValue.export();
@@ -277,6 +327,10 @@ public class EvaluatorImpl implements Evaluator {
     return doEvaluate(() -> VmUtils.readTextProperty(fileOutput));
   }
 
+  byte[] evaluateOutputBytes(VmTyped fileOutput) {
+    return doEvaluate(() -> VmUtils.readBytesProperty(fileOutput).export());
+  }
+
   private <T> T doEvaluate(Supplier<T> supplier) {
     @Nullable TimeoutTask timeoutTask = null;
     logger.clear();
@@ -304,24 +358,24 @@ public class EvaluatorImpl implements Evaluator {
             .bug("Stack overflow")
             .withCause(e.getCause())
             .build()
-            .toPklException(frameTransformer);
+            .toPklException(frameTransformer, color);
       }
       handleTimeout(timeoutTask);
-      throw e.toPklException(frameTransformer);
+      throw e.toPklException(frameTransformer, color);
     } catch (VmException e) {
       handleTimeout(timeoutTask);
-      throw e.toPklException(frameTransformer);
+      throw e.toPklException(frameTransformer, color);
     } catch (Exception e) {
       throw new PklBugException(e);
     } catch (ExceptionInInitializerError e) {
-      if (!(e.getCause() instanceof VmException)) {
+      if (!(e.getCause() instanceof VmException vmException)) {
         throw new PklBugException(e);
       }
-      var pklException = ((VmException) e.getCause()).toPklException(frameTransformer);
+      var pklException = vmException.toPklException(frameTransformer, color);
       var error = new ExceptionInInitializerError(pklException);
       error.setStackTrace(e.getStackTrace());
       throw new PklBugException(error);
-    } catch (ThreadDeath e) {
+    } catch (Throwable e) {
       if (e.getClass()
           .getName()
           .equals("com.oracle.truffle.polyglot.PolyglotEngineImpl$CancelExecution")) {
@@ -344,7 +398,7 @@ public class EvaluatorImpl implements Evaluator {
     return evalResult;
   }
 
-  protected <T> T doEvaluate(ModuleSource moduleSource, Function<VmTyped, T> doEvaluate) {
+  private <T> T doEvaluate(ModuleSource moduleSource, Function<VmTyped, T> doEvaluate) {
     return doEvaluate(
         () -> {
           var moduleKey = moduleResolver.resolve(moduleSource);
@@ -364,13 +418,40 @@ public class EvaluatorImpl implements Evaluator {
             "evaluationTimedOut", (timeout.getSeconds() + timeout.getNano() / 1_000_000_000d)));
   }
 
+  private VmTyped readModuleOutput(VmTyped module) {
+    var value = VmUtils.readMember(module, Identifier.OUTPUT);
+    if (value instanceof VmTyped typedOutput
+        && typedOutput.getVmClass().getPClassInfo() == PClassInfo.ModuleOutput) {
+      return typedOutput;
+    }
+
+    var moduleUri = module.getModuleInfo().getModuleKey().getUri();
+    var builder =
+        new VmExceptionBuilder()
+            .evalError(
+                "invalidModuleOutput",
+                "output",
+                PClassInfo.ModuleOutput.getDisplayName(),
+                VmUtils.getClass(value).getPClassInfo().getDisplayName(),
+                moduleUri);
+    var outputMember = module.getMember(Identifier.OUTPUT);
+    assert outputMember != null;
+    var uriOfValueMember = outputMember.getSourceSection().getSource().getURI();
+    // If `output` was explicitly re-assigned, show that in the stack trace.
+    if (!uriOfValueMember.equals(PClassInfo.pklBaseUri)) {
+      builder.withSourceSection(outputMember.getBodySection()).withMemberName("output");
+    }
+    throw builder.build();
+  }
+
   private VmException moduleOutputValueTypeMismatch(
       VmTyped module, PClassInfo<?> expectedClassInfo, Object value, VmTyped output) {
     var moduleUri = module.getModuleInfo().getModuleKey().getUri();
     var builder =
         new VmExceptionBuilder()
             .evalError(
-                "invalidModuleOutputValue",
+                "invalidModuleOutput",
+                "output.value",
                 expectedClassInfo.getDisplayName(),
                 VmUtils.getClass(value).getPClassInfo().getDisplayName(),
                 moduleUri);
