@@ -1,5 +1,5 @@
-/**
- * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+/*
+ * Copyright © 2024-2025 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -15,20 +15,20 @@
  */
 package org.pkl.core.project;
 
-import com.oracle.truffle.api.source.SourceSection;
-import java.io.FileOutputStream;
 import java.io.IOException;
 import java.io.OutputStream;
 import java.io.UncheckedIOException;
 import java.io.Writer;
+import java.net.URI;
+import java.net.URISyntaxException;
 import java.nio.file.Files;
 import java.nio.file.Path;
 import java.security.DigestOutputStream;
 import java.security.MessageDigest;
 import java.security.NoSuchAlgorithmException;
+import java.time.LocalDateTime;
+import java.time.Month;
 import java.util.ArrayList;
-import java.util.Calendar;
-import java.util.GregorianCalendar;
 import java.util.HashMap;
 import java.util.List;
 import java.util.Map;
@@ -43,6 +43,8 @@ import org.pkl.core.SecurityManager;
 import org.pkl.core.SecurityManagerException;
 import org.pkl.core.StackFrameTransformer;
 import org.pkl.core.ast.builder.ImportsAndReadsParser;
+import org.pkl.core.http.HttpClient;
+import org.pkl.core.module.ModuleKeyFactories;
 import org.pkl.core.module.ModuleKeys;
 import org.pkl.core.module.ProjectDependenciesManager;
 import org.pkl.core.module.ResolvedModuleKeys;
@@ -53,6 +55,7 @@ import org.pkl.core.packages.DependencyMetadata;
 import org.pkl.core.packages.PackageLoadError;
 import org.pkl.core.packages.PackageResolver;
 import org.pkl.core.packages.PackageUri;
+import org.pkl.core.runtime.ModuleResolver;
 import org.pkl.core.runtime.VmExceptionBuilder;
 import org.pkl.core.util.ByteArrayUtils;
 import org.pkl.core.util.ErrorMessages;
@@ -60,7 +63,6 @@ import org.pkl.core.util.GlobResolver;
 import org.pkl.core.util.GlobResolver.InvalidGlobPatternException;
 import org.pkl.core.util.IoUtils;
 import org.pkl.core.util.Nullable;
-import org.pkl.core.util.Pair;
 
 /**
  * Given a list of project directories, prepares artifacts to be published as a package.
@@ -78,16 +80,17 @@ import org.pkl.core.util.Pair;
  *   <li><em>thepackage@1.0.0.zip.sha256</em> - the SHA-256 checksum of the zip archive
  * </ul>
  */
-public class ProjectPackager {
+public final class ProjectPackager {
   /**
    * Modification time value for all zip entries in a package, to ensure that archives are
    * reproducible.
    *
-   * <p>Date is 1980 February 1st CET (value taken from {@code
-   * org.gradle.api.internal.file.archive.ZipCopyAction}).
+   * <p>Date is 1980 February 1st (value taken from {@code
+   * org.gradle.api.internal.file.archive.ZipCopyAction}). Note that this date does not contain time
+   * zone information.
    */
-  private static final long ZIP_ENTRY_MTIME =
-      new GregorianCalendar(1980, Calendar.FEBRUARY, 1, 0, 0, 0).getTimeInMillis();
+  private static final LocalDateTime ZIP_ENTRY_MTIME =
+      LocalDateTime.of(1980, Month.FEBRUARY, 1, 0, 0);
 
   private final EconomicMap<PackageUri, PackageResult> packageResults = EconomicMap.create();
 
@@ -95,6 +98,8 @@ public class ProjectPackager {
   private final Path workingDir;
   private final String outputPathPattern;
   private final StackFrameTransformer stackFrameTransformer;
+  private final boolean color;
+  private final SecurityManager securityManager;
   private final PackageResolver packageResolver;
   private final boolean skipPublishCheck;
   private final Writer outputWriter;
@@ -104,26 +109,35 @@ public class ProjectPackager {
       Path workingDir,
       String outputPathPattern,
       StackFrameTransformer stackFrameTransformer,
+      boolean color,
       SecurityManager securityManager,
+      HttpClient httpClient,
       boolean skipPublishCheck,
       Writer outputWriter) {
     this.projects = projects;
     this.workingDir = workingDir;
     this.outputPathPattern = outputPathPattern;
     this.stackFrameTransformer = stackFrameTransformer;
+    this.color = color;
+    this.securityManager = securityManager;
     // intentionally use InMemoryPackageResolver
-    this.packageResolver = PackageResolver.getInstance(securityManager, null);
+    this.packageResolver = PackageResolver.getInstance(securityManager, httpClient, null);
     this.skipPublishCheck = skipPublishCheck;
     this.outputWriter = outputWriter;
   }
 
+  private void writeLine(String line) throws IOException {
+    outputWriter.write(line);
+    outputWriter.write(IoUtils.getLineSeparator());
+  }
+
   public void createPackages() throws IOException {
     for (var project : projects) {
       var packageResult = doPackage(project);
-      outputWriter.write(workingDir.relativize(packageResult.getMetadataFile()) + "\n");
-      outputWriter.write(workingDir.relativize(packageResult.getMetadataChecksumFile()) + "\n");
-      outputWriter.write(workingDir.relativize(packageResult.getZipFile()) + "\n");
-      outputWriter.write(workingDir.relativize(packageResult.getZipChecksumFile()) + "\n");
+      writeLine(IoUtils.relativize(packageResult.metadataFile(), workingDir).toString());
+      writeLine(IoUtils.relativize(packageResult.metadataChecksumFile(), workingDir).toString());
+      writeLine(IoUtils.relativize(packageResult.zipFile(), workingDir).toString());
+      writeLine(IoUtils.relativize(packageResult.zipChecksumFile(), workingDir).toString());
       outputWriter.flush();
     }
   }
@@ -131,8 +145,8 @@ public class ProjectPackager {
   private Path resolveOutputDirectory(Package pkg) {
     var substituted =
         outputPathPattern
-            .replace("%{name}", pkg.getName())
-            .replace("%{version}", pkg.getVersion().toString());
+            .replace("%{name}", pkg.name())
+            .replace("%{version}", pkg.version().toString());
     return workingDir.resolve(substituted);
   }
 
@@ -142,13 +156,13 @@ public class ProjectPackager {
       throw new PklException(
           ErrorMessages.create("noPackageDefinedByProject", project.getProjectFileUri()));
     }
-    if (packageResults.containsKey(pkg.getUri())) {
-      return packageResults.get(pkg.getUri());
+    if (packageResults.containsKey(pkg.uri())) {
+      return packageResults.get(pkg.uri());
     }
     var files = collectPackageElements(project, pkg);
     validatePklImportsAndReads(project, files);
     var outputDir = resolveOutputDirectory(pkg);
-    var metadataFileName = IoUtils.takeLastSegment(pkg.getUri().getUri().getPath(), '/');
+    var metadataFileName = IoUtils.takeLastSegment(pkg.uri().getUri().getPath(), '/');
     var metadataFile = outputDir.resolve(metadataFileName);
     var metadataChecksumFile = outputDir.resolve(metadataFileName + ".sha256");
     var zipFile = outputDir.resolve(metadataFileName + ".zip");
@@ -164,27 +178,35 @@ public class ProjectPackager {
     var result =
         new PackageResult(
             metadataFile, metadataChecksumFile, zipFile, zipChecksumFile, metadataFileChecksum);
-    packageResults.put(pkg.getUri(), result);
+    packageResults.put(pkg.uri(), result);
     return result;
   }
 
   private void checkAlreadyPublishedPackage(Package pkg, String computedChecksum)
       throws IOException {
     try {
-      var metadataAndChecksum =
-          packageResolver.getDependencyMetadataAndComputeChecksum(pkg.getUri());
+      var metadataAndChecksum = packageResolver.getDependencyMetadataAndComputeChecksum(pkg.uri());
       var receivedChecksum = metadataAndChecksum.second.getSha256();
       if (!receivedChecksum.equals(computedChecksum)) {
         throw new PklException(
             ErrorMessages.create(
                 "packageAlreadyPublishedWithDifferentContents",
-                pkg.getUri(),
+                pkg.uri(),
                 computedChecksum,
                 receivedChecksum));
       }
     } catch (PackageLoadError e) {
-      if (e.getMessageName().equals("badHttpStatusCode") && (int) e.getArguments()[0] == 404) {
-        return;
+      if (e.getMessageName().equals("badHttpStatusCode")) {
+        if ((int) e.getArguments()[0] == 404) {
+          return;
+        } else {
+          throw new PklException(
+              ErrorMessages.create(
+                  "unableToAccessPublishedPackage",
+                  pkg.name(),
+                  pkg.packageZipUrl(),
+                  e.getArguments()[0]));
+        }
       }
       throw e;
     } catch (SecurityManagerException e) {
@@ -195,7 +217,7 @@ public class ProjectPackager {
   private String createDependencyMetadataAndComputeChecksum(
       Project project, Package pkg, Path metadataFile, String zipFileChecksum) throws IOException {
     var dependencyMetadata = createDependencyMetadata(project, pkg, zipFileChecksum);
-    try (var fos = newDigestOutputStream(new FileOutputStream(metadataFile.toFile()))) {
+    try (var fos = newDigestOutputStream(Files.newOutputStream((metadataFile)))) {
       dependencyMetadata.writeTo(fos);
       return ByteArrayUtils.toHex(fos.getMessageDigest().digest());
     }
@@ -206,10 +228,15 @@ public class ProjectPackager {
     try {
       var ret =
           new HashMap<String, RemoteDependency>(
-              project.getDependencies().getLocalDependencies().size()
-                  + project.getDependencies().getRemoteDependencies().size());
-      var projectDependenciesManager = new ProjectDependenciesManager(project.getDependencies());
-      for (var entry : project.getDependencies().getRemoteDependencies().entrySet()) {
+              project.getDependencies().localDependencies().size()
+                  + project.getDependencies().remoteDependencies().size());
+      // module resolver is only used for reading PklProject.deps.json, so provide one that reads
+      // files.
+      var moduleResolver = new ModuleResolver(List.of(ModuleKeyFactories.file));
+      var projectDependenciesManager =
+          new ProjectDependenciesManager(
+              project.getDependencies(), moduleResolver, this.securityManager);
+      for (var entry : project.getDependencies().remoteDependencies().entrySet()) {
         var resolved =
             (RemoteDependency)
                 projectDependenciesManager.getResolvedDependency(entry.getValue().getPackageUri());
@@ -221,7 +248,7 @@ public class ProjectPackager {
       for (var entry : project.getLocalProjectDependencies().entrySet()) {
         var localProject = entry.getValue();
         assert localProject.getPackage() != null;
-        var packageUri = localProject.getPackage().getUri();
+        var packageUri = localProject.getPackage().uri();
         var resolved = projectDependenciesManager.getResolvedDependency(packageUri);
         if (resolved instanceof LocalDependency) {
           var packageResult = doPackage(localProject);
@@ -229,7 +256,7 @@ public class ProjectPackager {
               entry.getKey(),
               new RemoteDependency(
                   packageUri.toExternalPackageUri(),
-                  new Checksums(packageResult.getMetadataChecksum())));
+                  new Checksums(packageResult.metadataChecksum())));
         } else {
           var remoteDep = (RemoteDependency) resolved;
           ret.put(
@@ -250,20 +277,21 @@ public class ProjectPackager {
   private DependencyMetadata createDependencyMetadata(
       Project project, Package pkg, String packageZipChecksum) throws IOException {
     return new DependencyMetadata(
-        pkg.getName(),
-        pkg.getUri(),
-        pkg.getVersion(),
-        pkg.getPackageZipUrl(),
+        pkg.name(),
+        pkg.uri(),
+        pkg.version(),
+        pkg.packageZipUrl(),
         new Checksums(packageZipChecksum),
         buildDependencies(project),
-        pkg.getSourceCodeUrlScheme(),
-        pkg.getSourceCode(),
-        pkg.getDocumentation(),
-        pkg.getLicense(),
-        pkg.getLicenseText(),
-        pkg.getAuthors(),
-        pkg.getIssueTracker(),
-        pkg.getDescription());
+        pkg.sourceCodeUrlScheme(),
+        pkg.sourceCode(),
+        pkg.documentation(),
+        pkg.license(),
+        pkg.licenseText(),
+        pkg.authors(),
+        pkg.issueTracker(),
+        pkg.description(),
+        project.getAnnotations());
   }
 
   private DigestOutputStream newDigestOutputStream(OutputStream outputStream) {
@@ -276,13 +304,6 @@ public class ProjectPackager {
     }
   }
 
-  private String ensureEndsWithSlash(Path file) {
-    if (file.endsWith("/")) {
-      return file.toString();
-    }
-    return file + "/";
-  }
-
   /**
    * Sets mtime to 0 so package creation is idempotent. Running the packager multiple times produces
    * the same output.
@@ -291,25 +312,18 @@ public class ProjectPackager {
       Project project, List<Path> files, Path outputZipFile) {
     DigestOutputStream digestOutputStream;
     try {
-      var outputFile = outputZipFile.toFile();
       Files.createDirectories(outputZipFile.getParent());
-      digestOutputStream = newDigestOutputStream(new FileOutputStream(outputFile));
+      digestOutputStream = newDigestOutputStream(Files.newOutputStream(outputZipFile));
     } catch (IOException e) {
       throw new UncheckedIOException(e);
     }
     try (var zos = new ZipOutputStream(digestOutputStream)) {
       for (var file : files) {
-        var relativePath = project.getProjectDir().relativize(file);
-        if (Files.isDirectory(file)) {
-          var zipEntry = new ZipEntry(ensureEndsWithSlash(relativePath));
-          zipEntry.setTime(ZIP_ENTRY_MTIME);
-          zos.putNextEntry(zipEntry);
-        } else {
-          var zipEntry = new ZipEntry(relativePath.toString());
-          zipEntry.setTime(ZIP_ENTRY_MTIME);
-          zos.putNextEntry(zipEntry);
-          Files.copy(file, zos);
-        }
+        var relativePath = IoUtils.relativize(file, project.getProjectDir());
+        var zipEntry = new ZipEntry(IoUtils.toNormalizedPathString(relativePath));
+        zipEntry.setTimeLocal(ZIP_ENTRY_MTIME);
+        zos.putNextEntry(zipEntry);
+        Files.copy(file, zos);
         zos.closeEntry();
       }
     } catch (IOException e) {
@@ -328,7 +342,7 @@ public class ProjectPackager {
 
   private List<Pattern> getExcludePatterns(Package pkg) {
     var excludePatterns = new ArrayList<Pattern>();
-    for (String s : pkg.getExclude()) {
+    for (String s : pkg.exclude()) {
       try {
         excludePatterns.add(GlobResolver.toRegexPattern(s));
       } catch (InvalidGlobPatternException e) {
@@ -342,10 +356,11 @@ public class ProjectPackager {
     var excludePatterns = getExcludePatterns(pkg);
     try (var stream = Files.walk(project.getProjectDir())) {
       return stream
+          .filter(Files::isRegularFile)
           .filter(
               (it) -> {
-                var fileNameRelativeToProjectRoot =
-                    project.getProjectDir().relativize(it).toString();
+                var relativePath = IoUtils.relativize(it, project.getProjectDir());
+                var fileNameRelativeToProjectRoot = IoUtils.toNormalizedPathString(relativePath);
                 for (var pattern : excludePatterns) {
                   if (pattern.matcher(it.getFileName().toString()).matches()) {
                     return false;
@@ -365,7 +380,7 @@ public class ProjectPackager {
   }
 
   private boolean isAbsoluteImport(String importStr) {
-    return importStr.matches("\\w:.*") || importStr.startsWith("@");
+    return importStr.matches("\\w+:.*") || importStr.startsWith("@");
   }
 
   /**
@@ -383,20 +398,30 @@ public class ProjectPackager {
       return;
     }
     for (var importContext : imports) {
-      var importStr = importContext.first;
-      var sourceSection = importContext.second;
+      var importStr = importContext.stringValue();
+      var sourceSection = importContext.sourceSection();
       if (isAbsoluteImport(importStr)) {
         continue;
       }
-      var importPath = Path.of(importStr);
-      if (importPath.isAbsolute() && !project.getProjectDir().toString().equals("/")) {
+      URI importUri;
+      try {
+        importUri = IoUtils.toUri(importStr);
+      } catch (URISyntaxException e) {
+        throw new VmExceptionBuilder()
+            .evalError("invalidModuleUri", importStr)
+            .withSourceSection(sourceSection)
+            .build()
+            .toPklException(stackFrameTransformer, color);
+      }
+      if (importStr.startsWith("/") && !project.getProjectDir().toString().equals("/")) {
         throw new VmExceptionBuilder()
             .evalError("invalidRelativeProjectImport", importStr)
             .withSourceSection(sourceSection)
             .build()
-            .toPklException(stackFrameTransformer);
+            .toPklException(stackFrameTransformer, color);
       }
       var currentPath = pklModulePath.getParent();
+      var importPath = Path.of(importUri.getPath());
       // It's not good enough to just check the normalized path to see whether it exists within the
       // root dir.
       // It's possible that the import path resolves to a path outside the project dir,
@@ -410,15 +435,15 @@ public class ProjectPackager {
               .evalError("invalidRelativeProjectImport", importStr)
               .withSourceSection(sourceSection)
               .build()
-              .toPklException(stackFrameTransformer);
+              .toPklException(stackFrameTransformer, color);
         }
       }
     }
   }
 
-  private @Nullable List<Pair<String, SourceSection>> getImportsAndReads(Path pklModulePath) {
+  private @Nullable List<ImportsAndReadsParser.Entry> getImportsAndReads(Path pklModulePath) {
     try {
-      var moduleKey = ModuleKeys.file(pklModulePath.toUri(), pklModulePath);
+      var moduleKey = ModuleKeys.file(pklModulePath);
       var resolvedModuleKey = ResolvedModuleKeys.file(moduleKey, moduleKey.getUri(), pklModulePath);
       return ImportsAndReadsParser.parse(moduleKey, resolvedModuleKey);
     } catch (IOException e) {
@@ -426,42 +451,48 @@ public class ProjectPackager {
     }
   }
 
-  public static class PackageResult {
-    private final Path zipFile;
-    private final Path zipChecksumFile;
-    private final Path metadataFile;
-    private final Path metadataChecksumFile;
-    private final String metadataChecksum;
-
-    public PackageResult(
-        Path zipFile,
-        Path zipChecksumFile,
-        Path metadataFile,
-        Path metadataChecksumFile,
-        String metadataChecksum) {
-      this.zipFile = zipFile;
-      this.zipChecksumFile = zipChecksumFile;
-      this.metadataFile = metadataFile;
-      this.metadataChecksumFile = metadataChecksumFile;
-      this.metadataChecksum = metadataChecksum;
-    }
-
+  public record PackageResult(
+      Path zipFile,
+      Path zipChecksumFile,
+      Path metadataFile,
+      Path metadataChecksumFile,
+      String metadataChecksum) {
+    /**
+     * @deprecated As of 0.28.0, replaced by {@link #zipFile()}.
+     */
+    @Deprecated(forRemoval = true)
     public Path getZipFile() {
       return zipFile;
     }
 
+    /**
+     * @deprecated As of 0.28.0, replaced by {@link #zipChecksumFile()}.
+     */
+    @Deprecated(forRemoval = true)
     public Path getZipChecksumFile() {
       return zipChecksumFile;
     }
 
+    /**
+     * @deprecated As of 0.28.0, replaced by {@link #metadataFile()}.
+     */
+    @Deprecated(forRemoval = true)
     public Path getMetadataFile() {
       return metadataFile;
     }
 
+    /**
+     * @deprecated As of 0.28.0, replaced by {@link #metadataChecksumFile()}.
+     */
+    @Deprecated(forRemoval = true)
     public Path getMetadataChecksumFile() {
       return metadataChecksumFile;
     }
 
+    /**
+     * @deprecated As of 0.28.0, replaced by {@link #metadataChecksum()}.
+     */
+    @Deprecated(forRemoval = true)
     public String getMetadataChecksum() {
       return metadataChecksum;
     }
