@@ -1,5 +1,5 @@
-/**
- * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+/*
+ * Copyright © 2024-2025 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -19,8 +19,9 @@ import java.io.FileNotFoundException;
 import java.io.IOException;
 import java.net.URI;
 import java.net.URISyntaxException;
+import java.net.http.HttpRequest;
+import java.net.http.HttpResponse.BodyHandlers;
 import java.nio.file.Files;
-import java.nio.file.Path;
 import java.util.ArrayList;
 import java.util.Collections;
 import java.util.List;
@@ -28,6 +29,10 @@ import java.util.Optional;
 import java.util.ServiceLoader;
 import org.pkl.core.SecurityManager;
 import org.pkl.core.SecurityManagerException;
+import org.pkl.core.externalreader.ExternalReaderProcess;
+import org.pkl.core.externalreader.ExternalReaderProcessException;
+import org.pkl.core.externalreader.ExternalResourceResolver;
+import org.pkl.core.externalreader.ResourceReaderSpec;
 import org.pkl.core.module.FileResolver;
 import org.pkl.core.module.ModulePathResolver;
 import org.pkl.core.module.PathElement;
@@ -37,7 +42,9 @@ import org.pkl.core.packages.Dependency.LocalDependency;
 import org.pkl.core.packages.PackageAssetUri;
 import org.pkl.core.packages.PackageResolver;
 import org.pkl.core.runtime.VmContext;
+import org.pkl.core.runtime.VmExceptionBuilder;
 import org.pkl.core.util.ErrorMessages;
+import org.pkl.core.util.HttpUtils;
 import org.pkl.core.util.IoUtils;
 import org.pkl.core.util.Nullable;
 
@@ -134,6 +141,33 @@ public final class ResourceReaders {
     return FromServiceProviders.INSTANCE;
   }
 
+  /**
+   * Returns a reader for external reader resources.
+   *
+   * <p>NOTE: {@code process} needs to be {@link ExternalReaderProcess#close closed} to avoid
+   * resource leaks.
+   */
+  public static ResourceReader externalProcess(String scheme, ExternalReaderProcess process) {
+    return new ExternalProcess(scheme, process, 0);
+  }
+
+  /**
+   * Returns a reader for external reader resources.
+   *
+   * <p>NOTE: {@code process} needs to be {@link ExternalReaderProcess#close closed} to avoid
+   * resource leaks.
+   */
+  public static ResourceReader externalProcess(
+      String scheme, ExternalReaderProcess process, long evaluatorId) {
+    return new ExternalProcess(scheme, process, evaluatorId);
+  }
+
+  /** Returns a reader for external and client reader resources. */
+  public static ResourceReader externalResolver(
+      ResourceReaderSpec spec, ExternalResourceResolver resolver) {
+    return new ExternalResolver(spec, resolver);
+  }
+
   private static final class EnvironmentVariable implements ResourceReader {
     static final ResourceReader INSTANCE = new EnvironmentVariable();
 
@@ -220,6 +254,12 @@ public final class ResourceReaders {
   private static final class FileResource extends UrlResource {
     static final ResourceReader INSTANCE = new FileResource();
 
+    @Override
+    public Optional<Object> read(URI uri) throws IOException, URISyntaxException {
+      IoUtils.validateFileUri(uri);
+      return super.read(uri);
+    }
+
     @Override
     public String getUriScheme() {
       return "file";
@@ -290,25 +330,23 @@ public final class ResourceReaders {
 
   private abstract static class UrlResource implements ResourceReader {
     @Override
-    public Optional<Object> read(URI uri) throws IOException {
+    public Optional<Object> read(URI uri) throws IOException, URISyntaxException {
+      if (HttpUtils.isHttpUrl(uri)) {
+        var httpClient = VmContext.get(null).getHttpClient();
+        var request = HttpRequest.newBuilder(uri).build();
+        var response = httpClient.send(request, BodyHandlers.ofByteArray());
+        if (response.statusCode() == 404) return Optional.empty();
+        HttpUtils.checkHasStatusCode200(response);
+        return Optional.of(new Resource(uri, response.body()));
+      }
+
       try {
-        var url = IoUtils.toUrl(uri);
-        var content = IoUtils.readBytes(url);
+        var content = IoUtils.readBytes(uri);
         return Optional.of(new Resource(uri, content));
       } catch (FileNotFoundException e) {
         return Optional.empty();
       }
     }
-
-    @Override
-    public boolean hasHierarchicalUris() {
-      return true;
-    }
-
-    @Override
-    public boolean isGlobbable() {
-      return false;
-    }
   }
 
   private static final class ClassPathResource implements ResourceReader {
@@ -485,11 +523,18 @@ public final class ResourceReaders {
         throws IOException, URISyntaxException, SecurityManagerException {
       var assetUri = new PackageAssetUri(uri);
       var dependency = getProjectDepsResolver().getResolvedDependency(assetUri.getPackageUri());
-      var path = getLocalPath(dependency, assetUri);
-      if (path != null) {
-        var url = path.toUri().toURL();
-        var bytes = IoUtils.readBytes(url);
-        return Optional.of(new Resource(uri, bytes));
+      var local = getLocalUri(dependency, assetUri);
+      if (local != null) {
+        var resourceManager = VmContext.get(null).getResourceManager();
+        var securityManager = VmContext.get(null).getSecurityManager();
+        securityManager.checkReadResource(local);
+        var reader = resourceManager.getResourceReader(local);
+        if (reader == null) {
+          throw new VmExceptionBuilder()
+              .evalError("noResourceReaderRegistered", uri.getScheme())
+              .build();
+        }
+        return resourceManager.doRead(reader, local, null);
       }
       var remoteDep = (Dependency.RemoteDependency) dependency;
       var bytes = getPackageResolver().getBytes(assetUri, true, remoteDep.getChecksums());
@@ -513,14 +558,20 @@ public final class ResourceReaders {
 
     @Override
     public List<PathElement> listElements(SecurityManager securityManager, URI baseUri)
-        throws IOException, SecurityManagerException {
+        throws IOException, SecurityManagerException, ExternalReaderProcessException {
       securityManager.checkResolveResource(baseUri);
       var packageAssetUri = PackageAssetUri.create(baseUri);
       var dependency =
           getProjectDepsResolver().getResolvedDependency(packageAssetUri.getPackageUri());
-      var path = getLocalPath(dependency, packageAssetUri);
-      if (path != null) {
-        return FileResolver.listElements(path);
+      var local = getLocalUri(dependency, packageAssetUri);
+      if (local != null) {
+        var reader = VmContext.get(null).getResourceManager().getResourceReader(local);
+        if (reader == null) {
+          throw new VmExceptionBuilder()
+              .evalError("noResourceReaderRegistered", local.getScheme())
+              .build();
+        }
+        return reader.listElements(securityManager, local);
       }
       var remoteDep = (Dependency.RemoteDependency) dependency;
       return getPackageResolver()
@@ -529,14 +580,20 @@ public final class ResourceReaders {
 
     @Override
     public boolean hasElement(SecurityManager securityManager, URI elementUri)
-        throws IOException, SecurityManagerException {
+        throws IOException, SecurityManagerException, ExternalReaderProcessException {
       securityManager.checkResolveResource(elementUri);
       var packageAssetUri = PackageAssetUri.create(elementUri);
       var dependency =
           getProjectDepsResolver().getResolvedDependency(packageAssetUri.getPackageUri());
-      var path = getLocalPath(dependency, packageAssetUri);
-      if (path != null) {
-        return FileResolver.hasElement(path);
+      var local = getLocalUri(dependency, packageAssetUri);
+      if (local != null) {
+        var reader = VmContext.get(null).getResourceManager().getResourceReader(local);
+        if (reader == null) {
+          throw new VmExceptionBuilder()
+              .evalError("noResourceReaderRegistered", local.getScheme())
+              .build();
+        }
+        return reader.hasElement(securityManager, local);
       }
       var remoteDep = (Dependency.RemoteDependency) dependency;
       return getPackageResolver()
@@ -555,16 +612,17 @@ public final class ResourceReaders {
       return projectDepsManager;
     }
 
-    private @Nullable Path getLocalPath(Dependency dependency, PackageAssetUri packageAssetUri) {
-      if (!(dependency instanceof LocalDependency)) {
+    private @Nullable URI getLocalUri(Dependency dependency, PackageAssetUri packageAssetUri) {
+      if (!(dependency instanceof LocalDependency localDependency)) {
         return null;
       }
-      return ((LocalDependency) dependency)
-          .resolveAssetPath(getProjectDepsResolver().getProjectDir(), packageAssetUri);
+      return localDependency.resolveAssetUri(
+          getProjectDepsResolver().getProjectBaseUri(), packageAssetUri);
     }
   }
 
   private static class FromServiceProviders {
+
     private static final List<ResourceReader> INSTANCE;
 
     static {
@@ -574,4 +632,111 @@ public final class ResourceReaders {
       INSTANCE = Collections.unmodifiableList(readers);
     }
   }
+
+  private static final class ExternalProcess implements ResourceReader {
+    private final String scheme;
+    private final ExternalReaderProcess process;
+    private final long evaluatorId;
+    private ExternalResolver underlying;
+
+    public ExternalProcess(String scheme, ExternalReaderProcess process, long evaluatorId) {
+      this.scheme = scheme;
+      this.process = process;
+      this.evaluatorId = evaluatorId;
+    }
+
+    private ExternalResolver getUnderlyingReader()
+        throws ExternalReaderProcessException, IOException {
+      if (underlying != null) {
+        return underlying;
+      }
+
+      var spec = process.getResourceReaderSpec(scheme);
+      if (spec == null) {
+        throw new ExternalReaderProcessException(
+            ErrorMessages.create("externalReaderDoesNotSupportScheme", "resource", scheme));
+      }
+      underlying = new ExternalResolver(spec, process.getResourceResolver(evaluatorId));
+      return underlying;
+    }
+
+    @Override
+    public String getUriScheme() {
+      return scheme;
+    }
+
+    @Override
+    public boolean hasHierarchicalUris() throws ExternalReaderProcessException, IOException {
+      return getUnderlyingReader().hasHierarchicalUris();
+    }
+
+    @Override
+    public boolean isGlobbable() throws ExternalReaderProcessException, IOException {
+      return getUnderlyingReader().isGlobbable();
+    }
+
+    @Override
+    public Optional<Object> read(URI uri) throws IOException, ExternalReaderProcessException {
+      return getUnderlyingReader().read(uri);
+    }
+
+    @Override
+    public boolean hasElement(SecurityManager securityManager, URI elementUri)
+        throws IOException, SecurityManagerException, ExternalReaderProcessException {
+      return getUnderlyingReader().hasElement(securityManager, elementUri);
+    }
+
+    @Override
+    public List<PathElement> listElements(SecurityManager securityManager, URI baseUri)
+        throws IOException, SecurityManagerException, ExternalReaderProcessException {
+      return getUnderlyingReader().listElements(securityManager, baseUri);
+    }
+
+    @Override
+    public void close() {
+      process.close();
+    }
+  }
+
+  private static final class ExternalResolver implements ResourceReader {
+    private final ResourceReaderSpec readerSpec;
+    private final ExternalResourceResolver resolver;
+
+    public ExternalResolver(ResourceReaderSpec readerSpec, ExternalResourceResolver resolver) {
+      this.readerSpec = readerSpec;
+      this.resolver = resolver;
+    }
+
+    @Override
+    public boolean hasHierarchicalUris() {
+      return readerSpec.hasHierarchicalUris();
+    }
+
+    @Override
+    public boolean isGlobbable() {
+      return readerSpec.isGlobbable();
+    }
+
+    @Override
+    public String getUriScheme() {
+      return readerSpec.scheme();
+    }
+
+    @Override
+    public Optional<Object> read(URI uri) throws IOException {
+      return resolver.read(uri);
+    }
+
+    @Override
+    public boolean hasElement(SecurityManager securityManager, URI elementUri)
+        throws SecurityManagerException {
+      return resolver.hasElement(securityManager, elementUri);
+    }
+
+    @Override
+    public List<PathElement> listElements(SecurityManager securityManager, URI baseUri)
+        throws IOException, SecurityManagerException {
+      return resolver.listElements(securityManager, baseUri);
+    }
+  }
 }
