@@ -1,4 +1,4 @@
-/**
+/*
  * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
@@ -15,7 +15,6 @@
  */
 package org.pkl.core.runtime;
 
-import com.oracle.truffle.api.CompilerDirectives;
 import java.net.URI;
 
 public final class MathModule extends StdLibModule {
@@ -30,10 +29,4 @@ public final class MathModule extends StdLibModule {
   public static VmTyped getModule() {
     return instance;
   }
-
-  @CompilerDirectives.TruffleBoundary
-  private static VmClass loadClass(String className) {
-    var theModule = getModule();
-    return (VmClass) VmUtils.readMember(theModule, Identifier.get(className));
-  }
 }
