@@ -1,4 +1,4 @@
-/**
+/*
  * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
@@ -15,13 +15,13 @@
  */
 package org.pkl.core.runtime;
 
-import java.net.URI;
+import static org.pkl.core.PClassInfo.pklSettingsUri;
 
 public final class SettingsModule extends StdLibModule {
   private static final VmTyped instance = VmUtils.createEmptyModule();
 
   static {
-    loadModule(URI.create("pkl:settings"), instance);
+    loadModule(pklSettingsUri, instance);
   }
 
   private SettingsModule() {}
