@@ -1,4 +1,4 @@
-/**
+/*
  * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
@@ -19,7 +19,7 @@ import java.nio.file.Path;
 import java.nio.file.spi.FileTypeDetector;
 import org.pkl.core.util.Nullable;
 
-public class VmFileDetector extends FileTypeDetector {
+public final class VmFileDetector extends FileTypeDetector {
   @Override
   public @Nullable String probeContentType(Path path) {
     var fileName = path.getFileName();
