@@ -1,5 +1,5 @@
-/**
- * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+/*
+ * Copyright © 2024-2025 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -15,6 +15,7 @@
  */
 package org.pkl.core.runtime;
 
+import com.oracle.truffle.api.CallTarget;
 import com.oracle.truffle.api.CompilerDirectives;
 import com.oracle.truffle.api.CompilerDirectives.TruffleBoundary;
 import com.oracle.truffle.api.Truffle;
@@ -43,10 +44,10 @@ import org.pkl.core.StackFrame;
 import org.pkl.core.StackFrameTransformer;
 import org.pkl.core.ast.ConstantNode;
 import org.pkl.core.ast.ExpressionNode;
-import org.pkl.core.ast.MemberNode;
 import org.pkl.core.ast.SimpleRootNode;
 import org.pkl.core.ast.VmModifier;
 import org.pkl.core.ast.builder.AstBuilder;
+import org.pkl.core.ast.builder.SymbolTable.CustomThisScope;
 import org.pkl.core.ast.expression.primary.CustomThisNode;
 import org.pkl.core.ast.expression.primary.ThisNode;
 import org.pkl.core.ast.member.*;
@@ -55,15 +56,14 @@ import org.pkl.core.ast.type.UnresolvedTypeNode;
 import org.pkl.core.module.ModuleKey;
 import org.pkl.core.module.ModuleKeys;
 import org.pkl.core.module.ResolvedModuleKey;
-import org.pkl.core.parser.LexParseException;
-import org.pkl.core.parser.Parser;
-import org.pkl.core.parser.antlr.PklParser.ExprContext;
 import org.pkl.core.util.EconomicMaps;
 import org.pkl.core.util.Nullable;
+import org.pkl.parser.Parser;
+import org.pkl.parser.ParserError;
+import org.pkl.parser.syntax.Expr;
 
 public final class VmUtils {
-  /** See {@link MemberNode#shouldRunTypecheck(VirtualFrame)}. */
-  @SuppressWarnings("JavadocReference")
+  /** See {@link VmUtils#shouldRunTypeCheck(VirtualFrame)}. */
   public static final Object SKIP_TYPECHECK_MARKER = new Object();
 
   public static final String REPL_TEXT = "repl:text";
@@ -175,6 +175,10 @@ public final class VmUtils {
     return (String) VmUtils.readMember((VmObjectLike) receiver, Identifier.TEXT);
   }
 
+  public static VmBytes readBytesProperty(VmObjectLike receiver) {
+    return (VmBytes) VmUtils.readMember(receiver, Identifier.BYTES);
+  }
+
   @TruffleBoundary
   public static Object readMember(VmObjectLike receiver, Object memberKey) {
     var result = readMemberOrNull(receiver, memberKey);
@@ -225,7 +229,7 @@ public final class VmUtils {
   @TruffleBoundary
   public static @Nullable Object readMemberOrNull(
       VmObjectLike receiver, Object memberKey, boolean checkType, IndirectCallNode callNode) {
-    assert (!(memberKey instanceof Identifier) || !((Identifier) memberKey).isLocalProp())
+    assert (!(memberKey instanceof Identifier identifier) || !identifier.isLocalProp())
         : "Must use ReadLocalPropertyNode for local properties.";
 
     final var cachedValue = receiver.getCachedValue(memberKey);
@@ -254,52 +258,138 @@ public final class VmUtils {
       IndirectCallNode callNode) {
 
     final var constantValue = member.getConstantValue();
+
+    // const members only need to be executed once on the prototype, and its cached value
+    // can be re-used for all children in the amends chain.
+    if (member.isConst() && owner != receiver) {
+      assert member.isProp();
+      assert owner.isPrototype();
+      var result = readMemberOrNull(owner, memberKey, checkType, callNode);
+      assert result != null;
+      receiver.setCachedValue(memberKey, result);
+      return result;
+    }
+
     if (constantValue != null) {
-      // for a property, do a type check
+      var result = constantValue;
+      // for a property, Listing element, or Mapping value, do a type check
       if (member.isProp()) {
         var property = receiver.getVmClass().getProperty(member.getName());
         if (property != null && property.getTypeNode() != null) {
           var callTarget = property.getTypeNode().getCallTarget();
           try {
             if (checkType) {
-              callNode.call(callTarget, receiver, property.getOwner(), constantValue);
+              result = callNode.call(callTarget, receiver, property.getOwner(), constantValue);
             } else {
-              callNode.call(
-                  callTarget, receiver, property.getOwner(), constantValue, SKIP_TYPECHECK_MARKER);
+              result =
+                  callNode.call(
+                      callTarget,
+                      receiver,
+                      property.getOwner(),
+                      constantValue,
+                      VmUtils.SKIP_TYPECHECK_MARKER);
             }
           } catch (VmException e) {
             CompilerDirectives.transferToInterpreter();
-            // A failed property type check looks as follows in the stack trace:
-            // x: Int(isPositive)
-            // at ...
-            // x = -10
-            // at ...
-            // However, if the value of `x` is a parse-time constant (as in `x = -10`),
-            // there's no root node for it and hence no stack trace element.
-            // In this case, insert a stack trace element to make the stack trace look the same
-            // as in the non-constant case. (Parse-time constants are an internal optimization.)
-            e.getInsertedStackFrames()
-                .put(
-                    callTarget,
-                    createStackFrame(member.getBodySection(), member.getQualifiedName()));
+            insertStackFrame(member, callTarget, e);
             throw e;
           }
         }
+      } else if (receiver instanceof VmListingOrMapping listingOrMapping
+          && owner instanceof VmListingOrMapping) {
+        // `owner instanceof VmListingOrMapping` guards against
+        // PropertiesRenderer amending VmDynamic with VmListing (hack?)
+        result = listingOrMapping.executeTypeCasts(constantValue, owner, callNode, member, null);
       }
 
-      receiver.setCachedValue(memberKey, constantValue);
-      return constantValue;
+      receiver.setCachedValue(memberKey, result);
+      return result;
     }
 
     var callTarget = member.getCallTarget();
-    Object computedValue;
+    Object result;
     if (checkType) {
-      computedValue = callNode.call(callTarget, receiver, owner, memberKey);
+      result = callNode.call(callTarget, receiver, owner, memberKey);
     } else {
-      computedValue = callNode.call(callTarget, receiver, owner, memberKey, SKIP_TYPECHECK_MARKER);
+      result = callNode.call(callTarget, receiver, owner, memberKey, VmUtils.SKIP_TYPECHECK_MARKER);
+    }
+    receiver.setCachedValue(memberKey, result);
+    return result;
+  }
+
+  /**
+   * Copies {@code numberOfLocalsToCopy} locals from {@code sourceFrame}, starting at {@code
+   * firstSourceSlot}, to {@code firstSourceSlot}, starting at {@code firstTargetSlot}.
+   */
+  public static void copyLocals(
+      VirtualFrame sourceFrame,
+      int firstSourceSlot,
+      VirtualFrame targetFrame,
+      int firstTargetSlot,
+      int numberOfLocalsToCopy) {
+    var sourceDescriptor = sourceFrame.getFrameDescriptor();
+    var targetDescriptor = targetFrame.getFrameDescriptor();
+    // Alternatively, locals could be copied with `numberOfLocalsToCopy`
+    // `ReadFrameSlotNode/WriteFrameSlotNode`'s.
+    for (int i = 0; i < numberOfLocalsToCopy; i++) {
+      var sourceSlot = firstSourceSlot + i;
+      var targetSlot = firstTargetSlot + i;
+      // If, for a particular call site of this method,
+      // slot kinds of `sourceDescriptor` will reach a steady state,
+      // then slot kinds of `targetDescriptor` will too.
+      var slotKind = sourceDescriptor.getSlotKind(sourceSlot);
+      switch (slotKind) {
+        case Boolean -> {
+          targetDescriptor.setSlotKind(targetSlot, FrameSlotKind.Boolean);
+          targetFrame.setBoolean(targetSlot, sourceFrame.getBoolean(sourceSlot));
+        }
+        case Long -> {
+          targetDescriptor.setSlotKind(targetSlot, FrameSlotKind.Long);
+          targetFrame.setLong(targetSlot, sourceFrame.getLong(sourceSlot));
+        }
+        case Double -> {
+          targetDescriptor.setSlotKind(targetSlot, FrameSlotKind.Double);
+          targetFrame.setDouble(targetSlot, sourceFrame.getDouble(sourceSlot));
+        }
+        case Object -> {
+          targetDescriptor.setSlotKind(targetSlot, FrameSlotKind.Object);
+          targetFrame.setObject(
+              targetSlot,
+              sourceFrame instanceof MaterializedFrame
+                  // Even though sourceDescriptor.getSlotKind is now Object,
+                  // it may have been a primitive kind when `sourceFrame`'s local was written.
+                  // Hence we need to read the local with getValue() instead of getObject().
+                  ? sourceFrame.getValue(sourceSlot)
+                  : sourceFrame.getObject(sourceSlot));
+        }
+        default -> {
+          CompilerDirectives.transferToInterpreter();
+          throw new VmExceptionBuilder().bug("Unexpected FrameSlotKind: " + slotKind).build();
+        }
+      }
+    }
+  }
+
+  // A failed property type check looks as follows in the stack trace:
+  // x: Int(isPositive)
+  // at ...
+  // x = -10
+  // at ...
+  // However, if the value of `x` is a parse-time constant (as in `x = -10`),
+  // there's no root node for it and hence no stack trace element.
+  // In this case, insert a stack trace element to make the stack trace look the same
+  // as in the non-constant case. (Parse-time constants are an internal optimization.)
+  public static void insertStackFrame(
+      ObjectMember member, CallTarget location, VmException exception) {
+    var sourceSection = member.getBodySection();
+    if (!sourceSection.isAvailable()) {
+      sourceSection = member.getSourceSection();
+    }
+    if (sourceSection.isAvailable()) {
+      exception
+          .getInsertedStackFrames()
+          .put(location, createStackFrame(sourceSection, member.getQualifiedName()));
     }
-    receiver.setCachedValue(memberKey, computedValue);
-    return computedValue;
   }
 
   public static @Nullable ObjectMember findMember(VmObjectLike receiver, Object memberKey) {
@@ -316,17 +406,9 @@ public final class VmUtils {
     return isCustomThisScope ? new CustomThisNode(sourceSection) : new ThisNode(sourceSection);
   }
 
-  public static boolean isRenderDirective(VmValue value) {
-    return value.getVmClass() == BaseModule.getRenderDirectiveClass();
-  }
-
-  public static boolean isRenderDirective(Object value) {
-    return value instanceof VmTyped && isRenderDirective((VmTyped) value);
-  }
-
   public static boolean isPcfRenderDirective(Object value) {
-    return value instanceof VmTyped
-        && ((VmTyped) value).getVmClass().getPClassInfo() == PClassInfo.PcfRenderDirective;
+    return value instanceof VmTyped typed
+        && typed.getVmClass().getPClassInfo() == PClassInfo.PcfRenderDirective;
   }
 
   @TruffleBoundary
@@ -340,8 +422,8 @@ public final class VmUtils {
 
   // implements same behavior as AnyNodes#getClass
   public static VmClass getClass(Object value) {
-    if (value instanceof VmValue) {
-      return ((VmValue) value).getVmClass();
+    if (value instanceof VmValue vmValue) {
+      return vmValue.getVmClass();
     }
     if (value instanceof String) {
       return BaseModule.getStringClass();
@@ -420,7 +502,7 @@ public final class VmUtils {
   }
 
   public static VmException toVmException(
-      LexParseException e, String text, URI moduleUri, String moduleName) {
+      ParserError e, String text, URI moduleUri, String moduleName) {
     var source =
         Source.newBuilder("pkl", text, moduleName)
             .mimeType(VmLanguage.MIME_TYPE)
@@ -430,47 +512,36 @@ public final class VmUtils {
     return toVmException(e, source, moduleName);
   }
 
-  // wanted to keep Parser/LexParseException API free from
+  // wanted to keep Parser/ParserError API free from
   // Truffle classes (Source), hence put this method here
-  public static VmException toVmException(LexParseException e, Source source, String moduleName) {
-    int lineStartOffset;
-    try {
-      lineStartOffset = source.getLineStartOffset(e.getLine());
-    } catch (IllegalArgumentException iae) {
-      // work around the fact that antlr and truffle disagree on how many lines a file that is
-      // ending in a newline has
-      lineStartOffset = source.getLineStartOffset(e.getLine() - 1);
-    }
-
+  public static VmException toVmException(ParserError e, Source source, String moduleName) {
     return new VmExceptionBuilder()
         .adhocEvalError(e.getMessage())
-        .withSourceSection(
-            source.createSection(
-                // compute char offset manually to work around
-                // https://github.com/graalvm/truffle/issues/184
-                lineStartOffset + e.getColumn() - 1, e.getLength()))
+        .withSourceSection(source.createSection(e.span().charIndex(), e.span().length()))
         .withMemberName(moduleName)
         .build();
   }
 
-  public static @Nullable String exportDocComment(@Nullable SourceSection docComment) {
+  public static @Nullable String exportDocComment(SourceSection @Nullable [] docComment) {
     if (docComment == null) return null;
 
     var builder = new StringBuilder();
-    var matcher = DOC_COMMENT_LINE_START.matcher(docComment.getCharacters());
-    var firstMatch = true;
-    while (matcher.find()) {
-      if (firstMatch) {
-        matcher.appendReplacement(builder, "");
-        firstMatch = false;
-      } else {
-        matcher.appendReplacement(builder, "\n");
+    for (var i = 0; i < docComment.length; i++) {
+      if (i > 0) {
+        builder.append("\n");
       }
+      var matcher = DOC_COMMENT_LINE_START.matcher(docComment[i].getCharacters());
+      var firstMatch = true;
+      while (matcher.find()) {
+        if (firstMatch) {
+          matcher.appendReplacement(builder, "");
+          firstMatch = false;
+        } else {
+          matcher.appendReplacement(builder, "\n");
+        }
+      }
+      matcher.appendTail(builder);
     }
-    matcher.appendTail(builder);
-    var newLength = builder.length() - 1;
-    assert builder.charAt(newLength) == '\n';
-    builder.setLength(newLength);
     return builder.toString();
   }
 
@@ -619,8 +690,8 @@ public final class VmUtils {
     // because constant type check wouldn't find the property (type)
     var isLocalTyped = property.isLocal() && typeNode != null;
 
-    if (bodyNode instanceof ConstantNode && !isLocalTyped) {
-      property.initConstantValue((ConstantNode) bodyNode);
+    if (bodyNode instanceof ConstantNode constantNode && !isLocalTyped) {
+      property.initConstantValue(constantNode);
       return property;
     }
 
@@ -653,8 +724,8 @@ public final class VmUtils {
     // because constant type check wouldn't find the property (type)
     var isLocalTyped = property.isLocal() && typeNode != null;
 
-    if (bodyNode instanceof ConstantNode && !isLocalTyped) {
-      property.initConstantValue((ConstantNode) bodyNode);
+    if (bodyNode instanceof ConstantNode constantNode && !isLocalTyped) {
+      property.initConstantValue(constantNode);
       return property;
     }
 
@@ -667,9 +738,7 @@ public final class VmUtils {
   }
 
   public static TypeNode[] resolveParameterTypes(
-      VirtualFrame frame,
-      FrameDescriptor descriptor,
-      @Nullable UnresolvedTypeNode[] parameterTypeNodes) {
+      VirtualFrame frame, FrameDescriptor descriptor, UnresolvedTypeNode[] parameterTypeNodes) {
 
     var resolvedNodes = new TypeNode[parameterTypeNodes.length];
 
@@ -773,10 +842,10 @@ public final class VmUtils {
         section.getEndColumn());
   }
 
-  private static ExprContext parseExpressionContext(String expression, Source source) {
+  private static Expr parseExpressionNode(String expression, Source source) {
     try {
-      return new Parser().parseExpressionInput(expression).expr();
-    } catch (LexParseException e) {
+      return new Parser().parseExpressionInput(expression);
+    } catch (ParserError e) {
       throw VmUtils.toVmException(e, source, REPL_TEXT);
     }
   }
@@ -807,8 +876,8 @@ public final class VmUtils {
             false);
     var language = VmLanguage.get(null);
     var builder = new AstBuilder(source, language, moduleInfo, moduleResolver);
-    var exprContext = parseExpressionContext(expression, source);
-    var exprNode = (ExpressionNode) exprContext.accept(builder);
+    var parsedExpression = parseExpressionNode(expression, source);
+    var exprNode = builder.visitExpr(parsedExpression);
     var rootNode =
         new SimpleRootNode(
             language, new FrameDescriptor(), exprNode.getSourceSection(), "", exprNode);
@@ -816,24 +885,28 @@ public final class VmUtils {
     return callNode.call(rootNode.getCallTarget(), module, module);
   }
 
-  public static int findSlot(VirtualFrame frame, Object identifier) {
-    var descriptor = frame.getFrameDescriptor();
-    for (int i = 0; i < descriptor.getNumberOfSlots(); i++) {
-      if (descriptor.getSlotName(i) == identifier
-          && frame.getTag(i) != FrameSlotKind.Illegal.tag
-          // Truffle initializes all frame tags as `FrameSlotKind.Object`, instead of
-          // `FrameSlotKind.Illegal`. Unevaluated (in a `when` with `false` condition) `for`
-          // generators, therefore, leave their bound variables tagged as `Object`, but `null`. If
-          // another `for` generator in the same scope binds variables with the same names, they
-          // resolve the wrong slot number.
-          && (frame.getTag(i) != FrameSlotKind.Object.tag || frame.getObject(i) != null)) {
-        return i;
-      }
-    }
-    return -1;
+  public static int findCustomThisSlot(VirtualFrame frame) {
+    return frame
+        .getFrameDescriptor()
+        .getAuxiliarySlots()
+        .getOrDefault(CustomThisScope.FRAME_SLOT_ID, -1);
   }
 
-  public static int findAuxiliarySlot(VirtualFrame frame, Object identifier) {
-    return frame.getFrameDescriptor().getAuxiliarySlots().getOrDefault(identifier, -1);
+  @TruffleBoundary
+  public static <K, V> V getMapValue(Map<K, V> map, K key) {
+    return map.get(key);
+  }
+
+  /**
+   * If true, the value computed by this node is not the final value exposed to user code but will
+   * still be amended.
+   *
+   * <p>Used to disable type check for to-be-amended properties. See {@link
+   * org.pkl.core.runtime.VmUtils#SKIP_TYPECHECK_MARKER}. IDEA: might be more appropriate to only
+   * skip constraints check
+   */
+  public static boolean shouldRunTypeCheck(VirtualFrame frame) {
+    return frame.getArguments().length != 4
+        || frame.getArguments()[3] != VmUtils.SKIP_TYPECHECK_MARKER;
   }
 }
