@@ -1,5 +1,5 @@
-/**
- * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+/*
+ * Copyright © 2024-2025 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -18,20 +18,22 @@ package org.pkl.core.stdlib.base;
 import com.oracle.truffle.api.CompilerDirectives;
 import com.oracle.truffle.api.CompilerDirectives.TruffleBoundary;
 import com.oracle.truffle.api.dsl.Specialization;
+import com.oracle.truffle.api.frame.VirtualFrame;
 import com.oracle.truffle.api.nodes.LoopNode;
-import org.graalvm.collections.EconomicMap;
 import org.pkl.core.ast.expression.binary.*;
 import org.pkl.core.ast.internal.IsInstanceOfNode;
 import org.pkl.core.ast.internal.IsInstanceOfNodeGen;
 import org.pkl.core.ast.lambda.*;
-import org.pkl.core.ast.member.ObjectMember;
+import org.pkl.core.ast.type.TypeNode;
+import org.pkl.core.ast.type.TypeNode.UInt8TypeAliasTypeNode;
+import org.pkl.core.ast.type.VmTypeMismatchException;
 import org.pkl.core.runtime.*;
 import org.pkl.core.stdlib.*;
 import org.pkl.core.stdlib.base.CollectionNodes.CompareByNode;
 import org.pkl.core.stdlib.base.CollectionNodes.CompareNode;
 import org.pkl.core.stdlib.base.CollectionNodes.CompareWithNode;
-import org.pkl.core.util.EconomicMaps;
 import org.pkl.core.util.EconomicSets;
+import org.pkl.core.util.LateInit;
 
 // duplication between ListNodes and SetNodes is "intentional"
 // (sharing nodes between VmCollection subtypes results in
@@ -557,6 +559,22 @@ public final class ListNodes {
     }
   }
 
+  public abstract static class mapIndexed extends ExternalMethod1Node {
+    @Child private ApplyVmFunction2Node applyLambdaNode = ApplyVmFunction2NodeGen.create();
+
+    @Specialization
+    protected VmList eval(VmList self, VmFunction function) {
+      var builder = self.builder();
+      long index = 0;
+
+      for (var elem : self) {
+        builder.add(applyLambdaNode.execute(function, index++, elem));
+      }
+      LoopNode.reportLoopCount(this, self.getLength());
+      return builder.build();
+    }
+  }
+
   public abstract static class mapNonNull extends ExternalMethod1Node {
     @Child private ApplyVmFunction1Node applyLambdaNode = ApplyVmFunction1Node.create();
 
@@ -573,7 +591,7 @@ public final class ListNodes {
     }
   }
 
-  public abstract static class mapIndexed extends ExternalMethod1Node {
+  public abstract static class mapNonNullIndexed extends ExternalMethod1Node {
     @Child private ApplyVmFunction2Node applyLambdaNode = ApplyVmFunction2NodeGen.create();
 
     @Specialization
@@ -582,7 +600,9 @@ public final class ListNodes {
       long index = 0;
 
       for (var elem : self) {
-        builder.add(applyLambdaNode.execute(function, index++, elem));
+        var newValue = applyLambdaNode.execute(function, index++, elem);
+        if (newValue instanceof VmNull) continue;
+        builder.add(newValue);
       }
       LoopNode.reportLoopCount(this, self.getLength());
       return builder.build();
@@ -1295,11 +1315,7 @@ public final class ListNodes {
     @Specialization
     @TruffleBoundary
     protected VmListing eval(VmList self) {
-      return new VmListing(
-          VmUtils.createEmptyMaterializedFrame(),
-          BaseModule.getListingClass().getPrototype(),
-          toObjectMembers(self),
-          self.getLength());
+      return self.toListing();
     }
   }
 
@@ -1307,22 +1323,37 @@ public final class ListNodes {
     @Specialization
     @TruffleBoundary
     protected VmDynamic eval(VmList self) {
-      return new VmDynamic(
-          VmUtils.createEmptyMaterializedFrame(),
-          BaseModule.getDynamicClass().getPrototype(),
-          toObjectMembers(self),
-          self.getLength());
+      return self.toDynamic();
     }
   }
 
-  private static EconomicMap<Object, ObjectMember> toObjectMembers(VmList self) {
-    var result = EconomicMaps.<Object, ObjectMember>create(self.getLength());
+  public abstract static class toBytes extends ExternalMethod0Node {
 
-    for (long idx = 0; idx < self.getLength(); idx++) {
-      EconomicMaps.put(
-          result, idx, VmUtils.createSyntheticObjectElement(String.valueOf(idx), self.get(idx)));
+    @Child @LateInit private TypeNode typeNode;
+
+    private TypeNode getTypeNode() {
+      if (typeNode == null) {
+        CompilerDirectives.transferToInterpreterAndInvalidate();
+        typeNode = new UInt8TypeAliasTypeNode();
+      }
+      return typeNode;
     }
 
-    return result;
+    @Specialization
+    protected VmBytes eval(VirtualFrame frame, VmList self) {
+      var typeNode = getTypeNode();
+      var bytes = new byte[self.getLength()];
+      try {
+        for (var i = 0; i < self.getLength(); i++) {
+          var elem = self.get(i);
+          var num = (Long) typeNode.executeEagerly(frame, elem);
+          bytes[i] = num.byteValue();
+        }
+      } catch (VmTypeMismatchException e) {
+        CompilerDirectives.transferToInterpreter();
+        throw e.toVmException();
+      }
+      return new VmBytes(self, bytes);
+    }
   }
 }
