@@ -1,5 +1,5 @@
-/**
- * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+/*
+ * Copyright © 2024-2025 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -19,10 +19,13 @@ import com.oracle.truffle.api.CompilerDirectives;
 import com.oracle.truffle.api.CompilerDirectives.TruffleBoundary;
 import com.oracle.truffle.api.dsl.Specialization;
 import com.oracle.truffle.api.nodes.LoopNode;
+import java.io.UnsupportedEncodingException;
 import java.nio.charset.StandardCharsets;
 import java.util.*;
 import java.util.regex.*;
+import org.pkl.core.PklBugException;
 import org.pkl.core.ast.lambda.ApplyVmFunction1Node;
+import org.pkl.core.ast.lambda.ApplyVmFunction1NodeGen;
 import org.pkl.core.runtime.*;
 import org.pkl.core.stdlib.*;
 import org.pkl.core.util.ByteArrayUtils;
@@ -148,6 +151,19 @@ public final class StringNodes {
     }
   }
 
+  public abstract static class isBase64 extends ExternalPropertyNode {
+    @Specialization
+    @TruffleBoundary
+    protected boolean eval(String self) {
+      try {
+        Base64.getDecoder().decode(self);
+        return true;
+      } catch (IllegalArgumentException e) {
+        return false;
+      }
+    }
+  }
+
   public abstract static class chars extends ExternalPropertyNode {
     @Specialization
     @TruffleBoundary
@@ -573,6 +589,8 @@ public final class StringNodes {
   }
 
   public abstract static class replaceFirstMapped extends ExternalMethod2Node {
+    @Child private ApplyVmFunction1Node applyLambdaNode = ApplyVmFunction1NodeGen.create();
+
     @Specialization
     @TruffleBoundary
     protected String eval(String self, String pattern, VmFunction mapper) {
@@ -588,12 +606,14 @@ public final class StringNodes {
     private String doEval(String self, Matcher matcher, VmFunction mapper) {
       if (!matcher.find()) return self;
       return self.substring(0, matcher.start())
-          + applyMapper(matcher, mapper)
+          + applyMapper(applyLambdaNode, matcher, mapper)
           + self.substring(matcher.end());
     }
   }
 
   public abstract static class replaceLastMapped extends ExternalMethod2Node {
+    @Child private ApplyVmFunction1Node applyLambdaNode = ApplyVmFunction1NodeGen.create();
+
     @Specialization
     @TruffleBoundary
     protected String eval(String self, String pattern, VmFunction mapper) {
@@ -610,12 +630,14 @@ public final class StringNodes {
       return !findLast(matcher)
           ? self
           : self.substring(0, matcher.start())
-              + applyMapper(matcher, mapper)
+              + applyMapper(applyLambdaNode, matcher, mapper)
               + self.substring(matcher.end());
     }
   }
 
   public abstract static class replaceAllMapped extends ExternalMethod2Node {
+    @Child private ApplyVmFunction1Node applyLambdaNode = ApplyVmFunction1NodeGen.create();
+
     @Specialization
     @TruffleBoundary
     protected String eval(String self, String pattern, VmFunction mapper) {
@@ -634,7 +656,7 @@ public final class StringNodes {
       var buffer = new StringBuilder();
 
       do {
-        matcher.appendReplacement(buffer, applyMapper(matcher, mapper));
+        matcher.appendReplacement(buffer, applyMapper(applyLambdaNode, matcher, mapper));
       } while (matcher.find());
 
       matcher.appendTail(buffer);
@@ -739,6 +761,22 @@ public final class StringNodes {
     }
   }
 
+  public abstract static class splitLimit extends ExternalMethod2Node {
+    @TruffleBoundary
+    @Specialization
+    protected VmList eval(String self, String separator, long limit) {
+      var parts = self.split(Pattern.quote(separator), (int) limit);
+      return VmList.create(parts);
+    }
+
+    @TruffleBoundary
+    @Specialization
+    protected VmList eval(String self, VmRegex separator, long limit) {
+      var parts = separator.getPattern().split(self, (int) limit);
+      return VmList.create(parts);
+    }
+  }
+
   public abstract static class capitalize extends ExternalMethod0Node {
     @TruffleBoundary
     @Specialization
@@ -768,7 +806,7 @@ public final class StringNodes {
     @Specialization
     protected long eval(String self) {
       try {
-        return Long.parseLong(self);
+        return Long.parseLong(removeUnderlinesFromNumber(self));
       } catch (NumberFormatException e) {
         throw exceptionBuilder()
             .evalError("cannotParseStringAs", "Int")
@@ -783,7 +821,7 @@ public final class StringNodes {
     @Specialization
     protected Object eval(String self) {
       try {
-        return Long.parseLong(self);
+        return Long.parseLong(removeUnderlinesFromNumber(self));
       } catch (NumberFormatException e) {
         return VmNull.withoutDefault();
       }
@@ -795,7 +833,7 @@ public final class StringNodes {
     @Specialization
     protected double eval(String self) {
       try {
-        return Double.parseDouble(self);
+        return Double.parseDouble(removeUnderlinesFromNumber(self));
       } catch (NumberFormatException e) {
         throw exceptionBuilder()
             .evalError("cannotParseStringAs", "Float")
@@ -810,7 +848,7 @@ public final class StringNodes {
     @Specialization
     protected Object eval(String self) {
       try {
-        return Double.parseDouble(self);
+        return Double.parseDouble(removeUnderlinesFromNumber(self));
       } catch (NumberFormatException e) {
         return VmNull.withoutDefault();
       }
@@ -877,7 +915,7 @@ public final class StringNodes {
     @TruffleBoundary
     @Specialization
     protected String eval(String self) {
-      return Base64.getEncoder().encodeToString(self.getBytes(StandardCharsets.UTF_8));
+      return ByteArrayUtils.base64(self.getBytes(StandardCharsets.UTF_8));
     }
   }
 
@@ -897,6 +935,35 @@ public final class StringNodes {
     }
   }
 
+  public abstract static class base64DecodedBytes extends ExternalPropertyNode {
+    @TruffleBoundary
+    @Specialization
+    protected VmBytes eval(String self) {
+      try {
+        return new VmBytes(Base64.getDecoder().decode(self));
+      } catch (IllegalArgumentException e) {
+        throw exceptionBuilder()
+            .adhocEvalError(e.getMessage())
+            .withProgramValue("String", self)
+            .withCause(e)
+            .build();
+      }
+    }
+  }
+
+  public abstract static class encodeToBytes extends ExternalMethod1Node {
+    @TruffleBoundary
+    @Specialization
+    protected VmBytes eval(String self, String charsetName) {
+      try {
+        var bytes = self.getBytes(charsetName);
+        return new VmBytes(bytes);
+      } catch (UnsupportedEncodingException e) {
+        throw PklBugException.unreachableCode();
+      }
+    }
+  }
+
   @TruffleBoundary
   private static String substringFrom(String string, int start) {
     return string.substring(start);
@@ -928,10 +995,30 @@ public final class StringNodes {
    * (i.e. the previous `find` returned `true`). The string returned is safe as a drop-in
    * replacement.
    */
-  private static String applyMapper(Matcher matcher, VmFunction mapper) {
+  private static String applyMapper(
+      ApplyVmFunction1Node applyNode, Matcher matcher, VmFunction mapper) {
     // -1 indicates regex match instead of group match (see comments in RegexMatchNodes)
     var regexMatch = RegexMatchFactory.create(Pair.of(matcher.toMatchResult(), -1));
-    var replacement = mapper.applyString(regexMatch);
+    var replacement = applyNode.executeString(mapper, regexMatch);
     return Matcher.quoteReplacement(replacement);
   }
+
+  /**
+   * Removes `_` from numbers to be parsed to be compatible with how Pkl parses numbers. Will return
+   * the string unmodified if it's invalid.
+   */
+  private static String removeUnderlinesFromNumber(String number) {
+    var builder = new StringBuilder();
+    var numberStart = true;
+    for (var i = 0; i < number.length(); i++) {
+      var c = number.charAt(i);
+      if (c != '_') {
+        builder.append(c);
+      } else if (numberStart) return number;
+
+      numberStart = c == '.' || c == 'e' || c == 'E';
+    }
+
+    return builder.toString();
+  }
 }
