@@ -1,5 +1,5 @@
-/**
- * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+/*
+ * Copyright © 2024-2025 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -39,6 +39,8 @@ import org.snakeyaml.engine.v2.constructor.StandardConstructor;
 import org.snakeyaml.engine.v2.exceptions.Mark;
 import org.snakeyaml.engine.v2.exceptions.YamlEngineException;
 import org.snakeyaml.engine.v2.nodes.*;
+import org.snakeyaml.engine.v2.resolver.ScalarResolver;
+import org.snakeyaml.engine.v2.schema.Schema;
 
 public final class ParserNodes {
   private static final Pattern WHITESPACE = Pattern.compile("\\s");
@@ -133,12 +135,24 @@ public final class ParserNodes {
 
   private static Load createLoad(VmTyped self, String text, String uri, PklConverter converter) {
     var mode = (String) VmUtils.readMember(self, Identifier.MODE);
-    var resolver = YamlUtils.getParserResolver(mode);
     var useMapping = (boolean) VmUtils.readMember(self, Identifier.USE_MAPPING);
     var settings =
         LoadSettings.builder()
             .setMaxAliasesForCollections(getMaxCollectionAliases(self))
-            .setScalarResolver(resolver)
+            .setSchema(
+                new Schema() {
+                  @Override
+                  public ScalarResolver getScalarResolver() {
+                    return YamlUtils.getParserResolver(mode);
+                  }
+
+                  @Override
+                  public Map<Tag, ConstructNode> getSchemaTagConstructors() {
+                    // ParserNodes.Constructor ignores this method,
+                    // but it is called from its superclass.
+                    return Map.of();
+                  }
+                })
             .setLabel(uri)
             .build();
     var source =
@@ -194,22 +208,10 @@ public final class ParserNodes {
       @Override
       public Object construct(Node node) {
         var value = ((ScalarNode) node).getValue();
-        switch (value) {
-          case "true":
-          case "True":
-          case "TRUE":
-          case "on":
-          case "On":
-          case "ON":
-          case "y":
-          case "Y":
-          case "yes":
-          case "Yes":
-          case "YES":
-            return true;
-          default:
-            return false;
-        }
+        return switch (value) {
+          case "true", "True", "TRUE", "on", "On", "ON", "y", "Y", "yes", "Yes", "YES" -> true;
+          default -> false;
+        };
       }
     }
 
@@ -250,23 +252,24 @@ public final class ParserNodes {
 
         if (value.charAt(offset) == '0' && value.length() > offset + 1) {
           switch (value.charAt(offset + 1)) {
-            case 'b':
+            case 'b' -> {
               radix = 2;
               offset += 2;
-              break;
-            case 'o':
+            }
+            case 'o' -> {
               radix = 8;
               offset += 2;
-              break;
-            case 'x':
+            }
+            case 'x' -> {
               radix = 16;
               offset += 2;
-              break;
-            default:
+            }
+            default -> {
               if (enable11Octals) {
                 radix = 8;
                 offset += 1;
               }
+            }
           }
         }
 
@@ -297,25 +300,12 @@ public final class ParserNodes {
 
         if (value.contains(":")) return parseBase60(value);
 
-        switch (value) {
-          case ".nan":
-          case ".NaN":
-          case ".NAN":
-            return Double.NaN;
-          case ".inf":
-          case ".Inf":
-          case ".INF":
-          case "+.inf":
-          case "+.Inf":
-          case "+.INF":
-            return Double.POSITIVE_INFINITY;
-          case "-.inf":
-          case "-.Inf":
-          case "-.INF":
-            return Double.NEGATIVE_INFINITY;
-          default:
-            return Double.valueOf(value);
-        }
+        return switch (value) {
+          case ".nan", ".NaN", ".NAN" -> Double.NaN;
+          case ".inf", ".Inf", ".INF", "+.inf", "+.Inf", "+.INF" -> Double.POSITIVE_INFINITY;
+          case "-.inf", "-.Inf", "-.INF" -> Double.NEGATIVE_INFINITY;
+          default -> Double.valueOf(value);
+        };
       }
 
       private static double parseBase60(String value) {
@@ -332,12 +322,13 @@ public final class ParserNodes {
       }
     }
 
-    // Pkl doesn't have a binary type, so parse as base64 string with whitespace removed
     private static class ConstructBinary implements ConstructNode {
       @Override
-      public String construct(Node node) {
+      public VmBytes construct(Node node) {
         var value = ((ScalarNode) node).getValue();
-        return WHITESPACE.matcher(value).replaceAll("");
+        var encoded = WHITESPACE.matcher(value).replaceAll("");
+        var decoded = Base64.getDecoder().decode(encoded);
+        return decoded.length == 0 ? VmBytes.EMPTY : new VmBytes(decoded);
       }
     }
 
@@ -345,14 +336,15 @@ public final class ParserNodes {
       @Override
       public VmListing construct(Node node) {
         var sequenceNode = (SequenceNode) node;
-        var members = EconomicMaps.<Object, ObjectMember>create(sequenceNode.getValue().size());
+        var size = sequenceNode.getValue().size();
+        var members = EconomicMaps.<Object, ObjectMember>create(size);
 
         var result =
             new VmListing(
                 VmUtils.createEmptyMaterializedFrame(),
                 BaseModule.getListingClass().getPrototype(),
                 members,
-                EconomicMaps.size(members));
+                size);
 
         if (!node.isRecursive()) {
           addMembers(sequenceNode, result);
@@ -390,14 +382,15 @@ public final class ParserNodes {
       @Override
       public VmListing construct(Node node) {
         var mappingNode = (MappingNode) node;
-        var members = EconomicMaps.<Object, ObjectMember>create(mappingNode.getValue().size());
+        var size = mappingNode.getValue().size();
+        var members = EconomicMaps.<Object, ObjectMember>create(size);
 
         var result =
             new VmListing(
                 VmUtils.createEmptyMaterializedFrame(),
                 BaseModule.getListingClass().getPrototype(),
                 members,
-                EconomicMaps.size(members));
+                size);
 
         if (!node.isRecursive()) {
           addMembers(mappingNode, result);
@@ -437,7 +430,8 @@ public final class ParserNodes {
       @Override
       public VmObject construct(Node node) {
         var mappingNode = (MappingNode) node;
-        var members = EconomicMaps.<Object, ObjectMember>create(mappingNode.getValue().size());
+        var size = mappingNode.getValue().size();
+        var members = EconomicMaps.<Object, ObjectMember>create(size);
 
         VmObject result;
         if (useMapping) {
@@ -488,9 +482,7 @@ public final class ParserNodes {
           currPath = valuePath;
 
           var memberName =
-              convertedKey instanceof String && !useMapping
-                  ? Identifier.get((String) convertedKey)
-                  : null;
+              convertedKey instanceof String string && !useMapping ? Identifier.get(string) : null;
 
           var member =
               new ObjectMember(
@@ -501,8 +493,8 @@ public final class ParserNodes {
                   "generated");
 
           currPath.push(
-              key instanceof String
-                  ? Identifier.get((String) key)
+              key instanceof String string
+                  ? Identifier.get(string)
                   : VmValueConverter.WILDCARD_PROPERTY);
           var value = constructObject(valueNode);
           var convertedValue = converter.convert(value, currPath);
