@@ -1,4 +1,4 @@
-/**
+/*
  * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
@@ -30,4 +30,9 @@ public final class EconomicSets {
   public static <E> boolean add(EconomicSet<E> self, E element) {
     return self.add(element);
   }
+
+  @TruffleBoundary
+  public static <E> boolean contains(EconomicSet<E> self, E element) {
+    return self.contains(element);
+  }
 }
