@@ -1,5 +1,5 @@
-/**
- * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+/*
+ * Copyright © 2024-2025 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -17,7 +17,6 @@ package org.pkl.core.util;
 
 import com.oracle.truffle.api.TruffleOptions;
 import java.io.FileNotFoundException;
-import java.io.FileOutputStream;
 import java.io.IOException;
 import java.io.InputStream;
 import java.io.Reader;
@@ -36,15 +35,22 @@ import java.util.regex.Pattern;
 import java.util.zip.ZipEntry;
 import java.util.zip.ZipOutputStream;
 import org.pkl.core.PklBugException;
+import org.pkl.core.Platform;
 import org.pkl.core.SecurityManager;
 import org.pkl.core.SecurityManagerException;
+import org.pkl.core.externalreader.ExternalReaderProcessException;
 import org.pkl.core.module.ModuleKey;
+import org.pkl.core.packages.PackageLoadError;
 import org.pkl.core.runtime.ReaderBase;
+import org.pkl.core.runtime.VmContext;
 import org.pkl.core.runtime.VmExceptionBuilder;
 
 public final class IoUtils {
 
-  private static final Pattern uriLike = Pattern.compile("\\w+:.*");
+  // Don't match paths like `C:\`, which are drive letters on Windows.
+  private static final Pattern uriLike = Pattern.compile("\\w+:[^\\\\].*");
+
+  private static final Pattern windowsPathLike = Pattern.compile("\\w:\\\\.*");
 
   private IoUtils() {}
 
@@ -67,12 +73,34 @@ public final class IoUtils {
     return uriLike.matcher(str).matches();
   }
 
+  public static boolean isWindowsAbsolutePath(String str) {
+    if (!isWindows()) return false;
+    return windowsPathLike.matcher(str).matches();
+  }
+
   /**
    * Converts the given string to a {@link URI}. This method MUST be used for constructing module
    * and resource URIs. Unlike {@code new URI(str)}, it correctly escapes paths of relative URIs.
    */
   public static URI toUri(String str) throws URISyntaxException {
-    return isUriLike(str) ? new URI(str) : new URI(null, null, str, null);
+    if (isUriLike(str)) {
+      return new URI(str);
+    }
+    return new URI(null, null, str, null);
+  }
+
+  /** Converts a URI to a Path, normalizing any non-ASCII characters. */
+  public static Path pathOf(URI uri) {
+    // Path.of(URI) throws on non-ASCII characters so the module URI here must be normalized to
+    // ASCII
+    // Unfortunately there's no way to go from URI -> ASCII URI directly
+    // so this must transform URI -> ASCII String -> ASCII URI
+    try {
+      return Path.of(new URI(uri.toASCIIString()));
+    } catch (URISyntaxException e) {
+      // impossible to get here; we started from a valid URI to begin with
+      throw PklBugException.unreachableCode();
+    }
   }
 
   /** Like {@link #toUri(String)}, except without checked exceptions. */
@@ -100,6 +128,9 @@ public final class IoUtils {
   }
 
   public static String readString(URL url) throws IOException {
+    if (HttpUtils.isHttpUrl(url)) {
+      throw new IllegalArgumentException("Should use HTTP client to GET " + url);
+    }
     try (var stream = url.openStream()) {
       return readString(stream);
     }
@@ -109,8 +140,11 @@ public final class IoUtils {
     return new String(inputStream.readAllBytes(), StandardCharsets.UTF_8);
   }
 
-  public static byte[] readBytes(URL url) throws IOException {
-    try (var stream = url.openStream()) {
+  public static byte[] readBytes(URI uri) throws IOException {
+    if (HttpUtils.isHttpUrl(uri)) {
+      throw new IllegalArgumentException("Should use HTTP client to GET " + uri);
+    }
+    try (var stream = IoUtils.toUrl(uri).openStream()) {
       return stream.readAllBytes();
     }
   }
@@ -139,13 +173,14 @@ public final class IoUtils {
   }
 
   public static void zipDirectory(Path sourceDir, Path targetFile) throws IOException {
-    try (var zipStream = new ZipOutputStream(new FileOutputStream(targetFile.toFile()))) {
+    try (var zipStream = new ZipOutputStream(Files.newOutputStream(targetFile))) {
       Files.walkFileTree(
           sourceDir,
           new SimpleFileVisitor<>() {
             public FileVisitResult visitFile(Path file, BasicFileAttributes attrs)
                 throws IOException {
-              zipStream.putNextEntry(new ZipEntry(sourceDir.relativize(file).toString()));
+              var relativePath = relativize(file, sourceDir);
+              zipStream.putNextEntry(new ZipEntry(toNormalizedPathString(relativePath)));
               Files.copy(file, zipStream);
               zipStream.closeEntry();
               return FileVisitResult.CONTINUE;
@@ -170,10 +205,15 @@ public final class IoUtils {
   }
 
   // not stored to avoid build-time initialization by native-image
+  @SuppressWarnings("SystemGetProperty")
   public static String getLineSeparator() {
     return System.getProperty("line.separator");
   }
 
+  public static Boolean isWindows() {
+    return Platform.current().operatingSystem().name().equals("Windows");
+  }
+
   public static String getName(String path) {
     var lastSep = Math.max(path.lastIndexOf('/'), path.lastIndexOf('\\'));
     return path.substring(lastSep + 1);
@@ -276,29 +316,15 @@ public final class IoUtils {
     return resolve(baseUri, importUri);
   }
 
-  /**
-   * Resolves {@code importUri} against the module key.
-   *
-   * <p>When {@code importUri} contains a triple-dot, it is resolved if the module key returns true
-   * for both {@link ModuleKey#isLocal()} and {@link ModuleKey#hasHierarchicalUris()}. Otherwise, an
-   * error is thrown.
-   */
-  public static URI resolve(SecurityManager securityManager, ModuleKey moduleKey, URI importUri)
-      throws URISyntaxException, IOException, SecurityManagerException {
-    if (importUri.isAbsolute()) {
-      return moduleKey.resolveUri(importUri);
-    }
+  private static URI resolveTripleDotImport(
+      SecurityManager securityManager, ModuleKey moduleKey, String tripleDotPath)
+      throws IOException, SecurityManagerException, ExternalReaderProcessException {
     var moduleKeyUri = moduleKey.getUri();
-    var tripleDotPath = parseTripleDotPath(importUri);
-    if (tripleDotPath == null) {
-      return moduleKey.resolveUri(importUri);
-    }
     if (!moduleKey.isLocal() || !moduleKey.hasHierarchicalUris()) {
       throw new VmExceptionBuilder()
           .evalError("cannotResolveTripleDotImports", moduleKeyUri)
           .build();
     }
-
     var currentPath =
         moduleKey.hasFragmentPaths() ? moduleKeyUri.getFragment() : moduleKeyUri.getPath();
     var effectiveImportPath =
@@ -328,6 +354,73 @@ public final class IoUtils {
     throw new FileNotFoundException();
   }
 
+  public static Pair<String, String> parseDependencyNotation(String importPath) {
+    var idx = importPath.indexOf('/');
+    if (idx == -1) {
+      // treat named dependency without a subpath as the root path.
+      // i.e. resolve to `@foo` to `package://example.com/foo@1.0.0#/`
+      return Pair.of(importPath.substring(1), "/");
+    }
+    return Pair.of(importPath.substring(1, idx), importPath.substring(idx));
+  }
+
+  private static URI resolveProjectDependency(ModuleKey moduleKey, String notation)
+      throws IOException, ExternalReaderProcessException {
+    var parsed = parseDependencyNotation(notation);
+    var name = parsed.getFirst();
+    var path = parsed.getSecond();
+    var projectDependenciesManager = VmContext.get(null).getProjectDependenciesManager();
+    if (!moduleKey.hasHierarchicalUris() && projectDependenciesManager != null) {
+      throw new PackageLoadError(
+          "cannotResolveDependencyWithoutHierarchicalUris",
+          projectDependenciesManager.getProjectFileUri());
+    }
+    if (projectDependenciesManager == null
+        || !projectDependenciesManager.hasUri(moduleKey.getUri())) {
+      throw new PackageLoadError("cannotResolveDependencyNoProject");
+    }
+    var dependency = projectDependenciesManager.getDependencies().get(name);
+    if (dependency != null) {
+      return dependency.getPackageUri().toPackageAssetUri(path).getUri();
+    }
+    throw new PackageLoadError("cannotFindDependencyInProject", name);
+  }
+
+  /**
+   * Resolves {@code importUri} against the module key.
+   *
+   * <p>When {@code importUri} contains a triple-dot, it is resolved if the module key returns true
+   * for both {@link ModuleKey#isLocal()} and {@link ModuleKey#hasHierarchicalUris()}. Otherwise, an
+   * error is thrown.
+   *
+   * <p>When {@code importUri} starts with a {@code @}, it is resolved if the module key supports
+   * dependency notation ()
+   */
+  public static URI resolve(SecurityManager securityManager, ModuleKey moduleKey, URI importUri)
+      throws URISyntaxException,
+          IOException,
+          SecurityManagerException,
+          ExternalReaderProcessException {
+    if (importUri.isAbsolute()) {
+      return moduleKey.resolveUri(importUri);
+    }
+    var tripleDotPath = parseTripleDotPath(importUri);
+    if (tripleDotPath != null) {
+      return resolveTripleDotImport(securityManager, moduleKey, tripleDotPath);
+    }
+    var moduleScheme = moduleKey.getUri().getScheme();
+    var isPackage =
+        moduleScheme.equalsIgnoreCase("package") || moduleScheme.equalsIgnoreCase("projectpackage");
+    var relativePart = importUri.getSchemeSpecificPart();
+    // Special-case handling of project dependencies.
+    // We'll allow the Package and ProjectPackage module keys to resolve dependency notation on
+    // their own.
+    if (relativePart.startsWith("@") && !isPackage) {
+      return resolveProjectDependency(moduleKey, relativePart);
+    }
+    return moduleKey.resolveUri(importUri);
+  }
+
   public static URI resolve(URI baseUri, URI newUri) {
     if (newUri.isAbsolute()) return newUri;
 
@@ -356,7 +449,9 @@ public final class IoUtils {
     }
   }
 
-  // URI.relativize() won't construct relative paths containing ".."
+  // URI.relativize won't construct relative paths containing `..`.
+  // Can't use Path.relativize because certain URI characters will throw InvalidPathException
+  // on Windows.
   public static URI relativize(URI uri, URI base) {
     if (uri.isOpaque()
         || base.isOpaque()
@@ -364,19 +459,60 @@ public final class IoUtils {
         || !Objects.equals(uri.getAuthority(), base.getAuthority())) {
       return uri;
     }
-
-    var basePath = Path.of(base.getRawPath());
-    if (!base.getRawPath().endsWith("/")) basePath = basePath.getParent();
-    var resultPath = basePath.relativize(Path.of(uri.getRawPath()));
-
+    var uriPath = uri.normalize().getPath();
+    var basePath = base.normalize().getPath();
     try {
-      return new URI(
-          null, null, null, -1, resultPath.toString(), uri.getRawQuery(), uri.getRawFragment());
+      if (basePath.isEmpty()) {
+        return uri;
+      }
+      var uriParts = Arrays.asList(uriPath.split("/"));
+      var baseParts = Arrays.asList(basePath.split("/"));
+      if (!basePath.endsWith("/")) {
+        // strip the last path segment of the base uri, unless it ends in a slash. `/foo/bar.pkl` ->
+        // `/foo`
+        baseParts = baseParts.subList(0, baseParts.size() - 1);
+      }
+      if (uriParts.equals(baseParts)) {
+        return new URI(null, null, null, -1, "", uri.getQuery(), uri.getFragment());
+      }
+      var start = 0;
+      while (start < Math.min(uriParts.size(), baseParts.size())) {
+        if (!uriParts.get(start).equals(baseParts.get(start))) {
+          break;
+        }
+        start++;
+      }
+      var uriPartsRemaining = uriParts.subList(start, uriParts.size());
+      var basePartsRemaining = baseParts.subList(start, baseParts.size());
+      if (basePartsRemaining.isEmpty()) {
+        return new URI(
+            null,
+            null,
+            null,
+            -1,
+            String.join("/", uriPartsRemaining),
+            uri.getQuery(),
+            uri.getFragment());
+      }
+      var resultingPath =
+          "../".repeat(basePartsRemaining.size()) + String.join("/", uriPartsRemaining);
+      return new URI(null, null, null, -1, resultingPath, uri.getQuery(), uri.getFragment());
     } catch (URISyntaxException e) {
-      throw new IllegalArgumentException(e);
+      // Impossible; started from a valid URI to begin with.
+      throw PklBugException.unreachableCode();
     }
   }
 
+  // On Windows, `Path.relativize` will fail if the two paths have different roots.
+  public static Path relativize(Path path, Path base) {
+    if (isWindows()) {
+      if (path.isAbsolute() && base.isAbsolute() && !path.getRoot().equals(base.getRoot())) {
+        return path;
+      }
+    }
+    return base.relativize(path);
+  }
+
   public static boolean isWhitespace(String str) {
     return str.codePoints().allMatch(Character::isWhitespace);
   }
@@ -414,7 +550,7 @@ public final class IoUtils {
     return ServiceLoader.load(serviceClass, IoUtils.class.getClassLoader());
   }
 
-  // not a static property to avoid compile-time evaluation by native-image
+  // not a static field to avoid compile-time evaluation by native-image
   public static boolean isTestMode() {
     return Boolean.getBoolean("org.pkl.testMode");
   }
@@ -423,6 +559,18 @@ public final class IoUtils {
     System.setProperty("org.pkl.testMode", "true");
   }
 
+  public static void setSystemProxy(URI proxyAddress) {
+    // Set HTTP proxy settings to configure the certificate revocation checker, because
+    // there is no other way to configure it. (see https://bugs.openjdk.org/browse/JDK-8256409)
+    //
+    // This only influences the behavior of the revocation checker.
+    // Otherwise, proxying is handled by [ProxySelector].
+    System.setProperty("http.proxyHost", proxyAddress.getHost());
+    System.setProperty(
+        "http.proxyPort",
+        proxyAddress.getPort() == -1 ? "80" : String.valueOf(proxyAddress.getPort()));
+  }
+
   public static @Nullable String parseTripleDotPath(URI importUri) throws URISyntaxException {
     var importScheme = importUri.getScheme();
     if (importScheme != null) return null;
@@ -442,191 +590,125 @@ public final class IoUtils {
 
   public static String toUnicodeEscape(int ch) {
     var hex = Integer.toHexString(ch);
-    switch (hex.length()) {
-      case 1:
-        return "\\u000" + hex;
-      case 2:
-        return "\\u00" + hex;
-      case 3:
-        return "\\u0" + hex;
-      case 4:
-        return "\\u" + hex;
-      default:
-        throw new IllegalArgumentException(String.valueOf(ch));
-    }
+    return switch (hex.length()) {
+      case 1 -> "\\u000" + hex;
+      case 2 -> "\\u00" + hex;
+      case 3 -> "\\u0" + hex;
+      case 4 -> "\\u" + hex;
+      default -> throw new IllegalArgumentException(String.valueOf(ch));
+    };
   }
 
   public static String toHexEscape(int ch) {
     var hex = Integer.toHexString(ch);
-    switch (hex.length()) {
-      case 1:
-        return "\\x0" + hex;
-      case 2:
-        return "\\x" + hex;
-      default:
-        throw new IllegalArgumentException(String.valueOf(ch));
-    }
+    return switch (hex.length()) {
+      case 1 -> "\\x0" + hex;
+      case 2 -> "\\x" + hex;
+      default -> throw new IllegalArgumentException(String.valueOf(ch));
+    };
   }
 
   public static boolean isHexDigit(char ch) {
-    switch (ch) {
-      case '0':
-      case '1':
-      case '2':
-      case '3':
-      case '4':
-      case '5':
-      case '6':
-      case '7':
-      case '8':
-      case '9':
-      case 'A':
-      case 'B':
-      case 'C':
-      case 'D':
-      case 'E':
-      case 'F':
-      case 'a':
-      case 'b':
-      case 'c':
-      case 'd':
-      case 'e':
-      case 'f':
-        return true;
-      default:
-        return false;
-    }
+    return switch (ch) {
+      case '0',
+          '1',
+          '2',
+          '3',
+          '4',
+          '5',
+          '6',
+          '7',
+          '8',
+          '9',
+          'A',
+          'B',
+          'C',
+          'D',
+          'E',
+          'F',
+          'a',
+          'b',
+          'c',
+          'd',
+          'e',
+          'f' ->
+          true;
+      default -> false;
+    };
   }
 
   public static boolean isHexDigitOrUnderscore(char ch) {
-    switch (ch) {
-      case '0':
-      case '1':
-      case '2':
-      case '3':
-      case '4':
-      case '5':
-      case '6':
-      case '7':
-      case '8':
-      case '9':
-      case 'A':
-      case 'B':
-      case 'C':
-      case 'D':
-      case 'E':
-      case 'F':
-      case 'a':
-      case 'b':
-      case 'c':
-      case 'd':
-      case 'e':
-      case 'f':
-      case '_':
-        return true;
-      default:
-        return false;
-    }
+    return switch (ch) {
+      case '0',
+          '1',
+          '2',
+          '3',
+          '4',
+          '5',
+          '6',
+          '7',
+          '8',
+          '9',
+          'A',
+          'B',
+          'C',
+          'D',
+          'E',
+          'F',
+          'a',
+          'b',
+          'c',
+          'd',
+          'e',
+          'f',
+          '_' ->
+          true;
+      default -> false;
+    };
   }
 
   @SuppressWarnings("BooleanMethodIsAlwaysInverted")
   public static boolean isDecimalDigit(char ch) {
-    switch (ch) {
-      case '0':
-      case '1':
-      case '2':
-      case '3':
-      case '4':
-      case '5':
-      case '6':
-      case '7':
-      case '8':
-      case '9':
-        return true;
-      default:
-        return false;
-    }
+    return switch (ch) {
+      case '0', '1', '2', '3', '4', '5', '6', '7', '8', '9' -> true;
+      default -> false;
+    };
   }
 
   @SuppressWarnings("BooleanMethodIsAlwaysInverted")
   public static boolean isDecimalDigitOrUnderscore(char ch) {
-    switch (ch) {
-      case '0':
-      case '1':
-      case '2':
-      case '3':
-      case '4':
-      case '5':
-      case '6':
-      case '7':
-      case '8':
-      case '9':
-      case '_':
-        return true;
-      default:
-        return false;
-    }
+    return switch (ch) {
+      case '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '_' -> true;
+      default -> false;
+    };
   }
 
   public static boolean isNonZeroDecimalDigit(char ch) {
-    switch (ch) {
-      case '1':
-      case '2':
-      case '3':
-      case '4':
-      case '5':
-      case '6':
-      case '7':
-      case '8':
-      case '9':
-        return true;
-      default:
-        return false;
-    }
+    return switch (ch) {
+      case '1', '2', '3', '4', '5', '6', '7', '8', '9' -> true;
+      default -> false;
+    };
   }
 
   public static boolean isOctalDigit(char ch) {
-    switch (ch) {
-      case '0':
-      case '1':
-      case '2':
-      case '3':
-      case '4':
-      case '5':
-      case '6':
-      case '7':
-        return true;
-      default:
-        return false;
-    }
+    return switch (ch) {
+      case '0', '1', '2', '3', '4', '5', '6', '7' -> true;
+      default -> false;
+    };
   }
 
   public static boolean isOctalDigitOrUnderscore(char ch) {
-    switch (ch) {
-      case '0':
-      case '1':
-      case '2':
-      case '3':
-      case '4':
-      case '5':
-      case '6':
-      case '7':
-      case '_':
-        return true;
-      default:
-        return false;
-    }
+    return switch (ch) {
+      case '0', '1', '2', '3', '4', '5', '6', '7', '_' -> true;
+      default -> false;
+    };
   }
 
   public static boolean isBinaryDigitOrUnderscore(char ch) {
-    switch (ch) {
-      case '0':
-      case '1':
-      case '_':
-        return true;
-      default:
-        return false;
-    }
+    return switch (ch) {
+      case '0', '1', '_' -> true;
+      default -> false;
+    };
   }
 
   /**
@@ -634,7 +716,7 @@ public final class IoUtils {
    * URI#resolve(URI)}
    */
   public static URI fixTripleSlashUri(URI baseUri, URI newUri) {
-    // `getHost()` is erroroneously `null` when parsing triple-slash URIs.
+    // `getHost()` is erroneously `null` when parsing triple-slash URIs.
     // Ensure that they are preserved during resolution.
     if (baseUri.getScheme() != null
         && baseUri.getScheme().equalsIgnoreCase(newUri.getScheme())
@@ -657,6 +739,95 @@ public final class IoUtils {
     return newUri;
   }
 
+  public static boolean isReservedFilenameChar(char character) {
+    if (isWindows()) {
+      return isReservedWindowsFilenameChar(character);
+    }
+    // posix; only NULL and `/` are reserved.
+    return character == 0 || character == '/';
+  }
+
+  /** Tells if this character cannot be used for filenames on Windows. */
+  public static boolean isReservedWindowsFilenameChar(char character) {
+    return switch (character) {
+      case 0,
+          1,
+          2,
+          3,
+          4,
+          5,
+          6,
+          7,
+          8,
+          9,
+          10,
+          11,
+          12,
+          13,
+          14,
+          15,
+          16,
+          17,
+          18,
+          19,
+          20,
+          21,
+          22,
+          23,
+          24,
+          25,
+          26,
+          27,
+          28,
+          29,
+          30,
+          31,
+          '<',
+          '>',
+          ':',
+          '"',
+          '\\',
+          '/',
+          '|',
+          '?',
+          '*' ->
+          true;
+      default -> false;
+    };
+  }
+
+  /**
+   * Windows reserves characters {@code <>:"\|?*} in filenames.
+   *
+   * <p>For any such characters, enclose their decimal character code with parentheses. Verbatim
+   * {@code (} is encoded as {@code ((}.
+   */
+  public static String encodePath(String path) {
+    if (path.isEmpty()) return path;
+    var sb = new StringBuilder();
+    for (var i = 0; i < path.length(); i++) {
+      var character = path.charAt(i);
+      if (isReservedWindowsFilenameChar(character) && character != '/') {
+        sb.append('(');
+        sb.append(ByteArrayUtils.toHex(new byte[] {(byte) character}));
+        sb.append(")");
+      } else if (character == '(') {
+        sb.append("((");
+      } else {
+        sb.append(character);
+      }
+    }
+    return sb.toString();
+  }
+
+  /** Returns a path string that uses unix-like path separators. */
+  public static String toNormalizedPathString(Path path) {
+    if (isWindows()) {
+      return path.toString().replace("\\", "/");
+    }
+    return path.toString();
+  }
+
   private static int getExclamationMarkIndex(String jarUri) {
     var index = jarUri.indexOf('!');
     if (index == -1) {
@@ -664,4 +835,23 @@ public final class IoUtils {
     }
     return index;
   }
+
+  public static void validateFileUri(URI uri) throws URISyntaxException {
+    if (!uri.getSchemeSpecificPart().startsWith("/")) {
+      throw new URISyntaxException(uri.toString(), ErrorMessages.create("invalidOpaqueFileUri"));
+    }
+  }
+
+  public static void validateRewriteRule(URI rewrite) {
+    if (!Objects.equals(rewrite.getScheme(), "http")
+        && !Objects.equals(rewrite.getScheme(), "https")) {
+      throw new IllegalArgumentException(
+          "Rewrite rule must start with 'http://' or 'https://', but was '%s'".formatted(rewrite));
+    }
+
+    if (!rewrite.toString().endsWith("/")) {
+      throw new IllegalArgumentException(
+          "Rewrite rule must end with '/', but was '%s'".formatted(rewrite));
+    }
+  }
 }
