@@ -1,3 +1,18 @@
+/*
+ * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+ *
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
+ *
+ *     https://www.apache.org/licenses/LICENSE-2.0
+ *
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
+ */
 package org.pkl.core
 
 import org.junit.jupiter.api.Test
@@ -5,19 +20,21 @@ import org.junit.jupiter.api.assertDoesNotThrow
 import org.junit.jupiter.api.assertThrows
 
 class DynamicTest {
-  private fun brokenInput(value: String): String = """
+  private fun brokenInput(value: String): String =
+    """
       class Person { name: String; age: Int }
 
       person: Person = new { name = 42; age = "Pigeon" } // oops
 
       output { value = $value }
-    """.trimIndent()
+    """
+      .trimIndent()
 
   @Test
   fun `property access respects type`() {
     // Does not involve Dynamic, but is a baseline for the other cases.
     val evaluator = Evaluator.preconfigured()
-    assertThrows<PklException> { 
+    assertThrows<PklException> {
       evaluator.evaluateOutputText(ModuleSource.text(brokenInput("person.name")))
     }
   }
