@@ -1,3 +1,18 @@
+/*
+ * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+ *
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
+ *
+ *     https://www.apache.org/licenses/LICENSE-2.0
+ *
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
+ */
 package org.pkl.core
 
 import java.io.StringWriter
@@ -12,9 +27,7 @@ class JsonRendererTest {
   @Test
   fun `render document`() {
     val evaluator = Evaluator.preconfigured()
-    val module = evaluator.evaluate(
-      ModuleSource.modulePath("org/pkl/core/rendererTest.pkl")
-    )
+    val module = evaluator.evaluate(ModuleSource.modulePath("org/pkl/core/rendererTest.pkl"))
     val writer = StringWriter()
     val renderer = ValueRenderers.json(writer, "  ", true)
 
@@ -23,14 +36,13 @@ class JsonRendererTest {
     val expected = IoUtils.readClassPathResourceAsString(javaClass, "rendererTest.json")
 
     assertThat(output).isEqualTo(expected)
-    assertThatCode { JsonParser(object : JsonHandler<Any, Any>() {}).parse(output) }.doesNotThrowAnyException()
+    assertThatCode { JsonParser(object : JsonHandler<Any, Any>() {}).parse(output) }
+      .doesNotThrowAnyException()
   }
 
   @Test
   fun `rendered document ends in newline`() {
-    val module: PModule = Evaluator
-      .preconfigured()
-      .evaluate(ModuleSource.text("foo { bar = 0 }"))
+    val module: PModule = Evaluator.preconfigured().evaluate(ModuleSource.text("foo { bar = 0 }"))
 
     for (omitNullProperties in listOf(false, true)) {
       val writer = StringWriter()
