@@ -1,5 +1,5 @@
-/**
- * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+/*
+ * Copyright © 2024-2025 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -15,5 +15,8 @@
  */
 package org.pkl.doc
 
-class DocGeneratorException(message: String, cause: Throwable? = null) :
+open class DocGeneratorException(message: String, cause: Throwable? = null) :
   RuntimeException(message, cause)
+
+class DocGeneratorBugException(message: String, cause: Throwable? = null) :
+  DocGeneratorException(message, cause)
