@@ -1,5 +1,5 @@
-/**
- * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+/*
+ * Copyright © 2024-2025 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -35,7 +35,7 @@ internal sealed class DocScope {
 
   abstract val parent: DocScope?
 
-  private val siteScope: SiteScope? by lazy {
+  val siteScope: SiteScope? by lazy {
     var scope = this
     while (scope !is SiteScope) {
       scope = scope.parent ?: return@lazy null
@@ -71,10 +71,7 @@ internal sealed class DocScope {
   fun resolveModuleNameToRelativeDocUrl(name: String): URI? =
     resolveModuleNameToDocUrl(name)?.let { IoUtils.relativize(it, url) }
 
-  abstract fun resolveModuleNameToSourceUrl(
-    name: String,
-    sourceLocation: Member.SourceLocation
-  ): URI?
+  abstract fun resolveModuleNameToSourceUrl(name: String, sourceLocation: SourceLocation): URI?
 
   /** Resolves the given method name relative to this scope. */
   abstract fun resolveMethod(name: String): MethodScope?
@@ -120,33 +117,41 @@ internal sealed class DocScope {
 /** A scope that corresponds to an entire Pkldoc page. */
 internal abstract class PageScope : DocScope() {
   /** The location of the runtime data file for this page. */
-  abstract val dataUrl: URI
+  abstract val perPackageDataUrl: URI
+
+  /** The location of the runtime data JSON file that is per-version for this page. */
+  abstract val perPackageVersionDataUrl: URI
 }
 
 // equality is identity
 internal class SiteScope(
-  docPackages: List<DocPackage>,
+  val docPackages: List<DocPackage>,
   private val overviewImports: Map<String, URI>,
   private val importResolver: (URI) -> ModuleSchema,
-  outputDir: Path
+  val outputDir: Path,
 ) : PageScope() {
   private val pklVersion = Release.current().version().withBuild(null).toString()
 
   private val pklBaseModule: ModuleSchema by lazy { importResolver("pkl:base".toUri()) }
 
-  val packageScopes: Map<String, PackageScope> by lazy {
+  val packageScopes: Map<DocPackageInfo, PackageScope> by lazy {
     docPackages.associate { docPackage ->
-      docPackage.name to
+      docPackage.docPackageInfo to
         PackageScope(
           docPackage.docPackageInfo,
           docPackage.docModules.map { it.schema },
           pklBaseModule,
           docPackage.docPackageInfo.overviewImports,
-          this
+          this,
         )
     }
   }
 
+  val stdlibPackageScope: PackageScope? by lazy {
+    val pklPackage = docPackages.find { it.name == "pkl" }
+    packageScopes[pklPackage?.docPackageInfo]
+  }
+
   private val pklBaseScope: ModuleScope by lazy {
     ModuleScope(pklBaseModule, resolveModuleNameToDocUrl("pkl.base")!!, null)
   }
@@ -158,14 +163,17 @@ internal class SiteScope(
     IoUtils.ensurePathEndsWithSlash(outputDir.toUri()).resolve("index.html")
   }
 
-  override val dataUrl: URI
+  override val perPackageDataUrl: URI
+    get() = throw UnsupportedOperationException("perVersionDataUrl")
+
+  override val perPackageVersionDataUrl: URI
     get() = throw UnsupportedOperationException("perVersionDataUrl")
 
   fun createEmptyPackageScope(
     name: String,
     version: String,
     sourceCodeUrlScheme: String?,
-    sourceCode: URI?
+    sourceCode: URI?,
   ): PackageScope =
     PackageScope(
       DocPackageInfo(
@@ -178,19 +186,19 @@ internal class SiteScope(
         importUri = "",
         issueTracker = null,
         overview = null,
-        uri = null
+        uri = null,
       ),
       emptyList(),
       pklBaseModule,
       emptyMap(),
-      this
+      this,
     )
 
   override fun getMethod(name: String): MethodScope? = null
 
   override fun getProperty(name: String): DocScope? = null
 
-  fun getPackage(name: String): PackageScope = packageScopes.getValue(name)
+  fun getPackage(packageInfo: DocPackageInfo): PackageScope = packageScopes.getValue(packageInfo)
 
   fun resolveImport(uri: URI): ModuleSchema = importResolver(uri)
 
@@ -198,8 +206,8 @@ internal class SiteScope(
     when {
       name.startsWith("pkl.") -> {
         val packagePage =
-          packageScopes["pkl"]?.url // link to locally generated stdlib docs if available
-           ?: PklInfo.current().packageIndex.getPackagePage("pkl", pklVersion).toUri()
+          stdlibPackageScope?.url // link to locally generated stdlib docs if available
+          ?: PklInfo.current().packageIndex.getPackagePage("pkl", pklVersion).toUri()
         packagePage.resolve(name.substring(4) + "/")
       }
       // doesn't make much sense to search in [packageScopes]
@@ -207,16 +215,13 @@ internal class SiteScope(
       else -> null
     }
 
-  override fun resolveModuleNameToSourceUrl(
-    name: String,
-    sourceLocation: Member.SourceLocation
-  ): URI? =
+  override fun resolveModuleNameToSourceUrl(name: String, sourceLocation: SourceLocation): URI? =
     when {
       name.startsWith("pkl.") -> {
         val path = "/stdlib/${name.substring(4)}.pkl"
         Release.current()
           .sourceCode()
-          .sourceCodeUrlScheme
+          .sourceCodeUrlScheme()
           .replaceSourceCodePlaceholders(path, sourceLocation)
           .toUri()
       }
@@ -233,8 +238,7 @@ internal class SiteScope(
     overviewImports[name]?.let { uri ->
       val mod = resolveImport(uri)
       resolveModuleNameToDocUrl(mod.moduleName)?.let { url -> ModuleScope(mod, url, null) }
-    }
-      ?: pklBaseScope.getProperty(name)
+    } ?: pklBaseScope.getProperty(name)
 }
 
 internal class PackageScope(
@@ -242,7 +246,7 @@ internal class PackageScope(
   modules: List<ModuleSchema>,
   pklBaseModule: ModuleSchema,
   private val overviewImports: Map<String, URI>,
-  override val parent: SiteScope
+  override val parent: SiteScope,
 ) : PageScope() {
   val name = docPackageInfo.name
 
@@ -254,7 +258,7 @@ internal class PackageScope(
     modules.associate { module ->
       val docUrl =
         url.resolve(
-          "${module.moduleName.substring(modulePrefix.length).replace('.', '/')}/index.html"
+          getModulePath(module.moduleName, modulePrefix).pathEncoded.uriEncoded + "/index.html"
         )
       module.moduleName to ModuleScope(module, docUrl, this)
     }
@@ -264,9 +268,20 @@ internal class PackageScope(
     ModuleScope(pklBaseModule, resolveModuleNameToDocUrl("pkl.base")!!, null)
   }
 
-  override val url: URI by lazy { parent.url.resolve("./$name/$version/index.html") }
+  override val url: URI by lazy { parent.url.resolve("./${name.pathEncoded}/$version/index.html") }
 
-  override val dataUrl: URI by lazy { parent.url.resolve("./data/$name/$version/index.js") }
+  fun urlForVersionRelativeTo(scope: DocScope, version: String): URI {
+    val myVersion = parent.url.resolve("./${name.pathEncoded}/$version/index.html")
+    return IoUtils.relativize(myVersion, scope.url)
+  }
+
+  override val perPackageDataUrl: URI by lazy {
+    parent.url.resolve("./data/${name.pathEncoded}/_/index.json")
+  }
+
+  override val perPackageVersionDataUrl: URI by lazy {
+    parent.url.resolve("./data/${name.pathEncoded}/$version/index.json")
+  }
 
   fun getModule(name: String): ModuleScope = moduleScopes.getValue(name)
 
@@ -309,8 +324,7 @@ internal class PackageScope(
     overviewImports[name]?.let { uri ->
       val mod = resolveImport(uri)
       resolveModuleNameToDocUrl(mod.moduleName)?.let { url -> ModuleScope(mod, url, null) }
-    }
-      ?: getPklBaseProperty(name)
+    } ?: getPklBaseProperty(name)
 
   override fun equals(other: Any?): Boolean =
     other is PackageScope && docPackageInfo.name == other.docPackageInfo.name
@@ -322,16 +336,22 @@ internal class PackageScope(
 internal class ModuleScope(
   val module: ModuleSchema,
   override val url: URI,
-  override val parent: PackageScope?
+  override val parent: PackageScope?,
 ) : PageScope() {
   val name: String
     get() = module.moduleName
 
   val path: String by lazy {
-    name.substring(parent!!.docPackageInfo.moduleNamePrefix.length).replace('.', '/')
+    getModulePath(module.moduleName, parent!!.docPackageInfo.moduleNamePrefix).uriEncoded
   }
 
-  override val dataUrl: URI by lazy { parent!!.dataUrl.resolve("./$path/index.js") }
+  override val perPackageDataUrl: URI by lazy {
+    parent!!.perPackageDataUrl.resolve("./$path/index.json")
+  }
+
+  override val perPackageVersionDataUrl: URI by lazy {
+    parent!!.perPackageVersionDataUrl.resolve("./$path/index.json")
+  }
 
   override fun getMethod(name: String): MethodScope? =
     module.moduleClass.allMethods[name]?.let { MethodScope(it, this) }
@@ -339,7 +359,7 @@ internal class ModuleScope(
   override fun getProperty(name: String): DocScope? =
     module.moduleClass.allProperties[name]?.let { PropertyScope(it, this) }
       ?: module.allClasses[name]?.let { ClassScope(it, url, this) }
-        ?: module.allTypeAliases[name]?.let { TypeAliasScope(it, url, this) }
+      ?: module.allTypeAliases[name]?.let { TypeAliasScope(it, url, this) }
 
   private fun resolveImport(uri: URI): ModuleSchema = parent!!.resolveImport(uri)
 
@@ -362,7 +382,8 @@ internal class ModuleScope(
 
   override fun resolveMethod(name: String): MethodScope? =
     module.moduleClass.methods[name]?.let { MethodScope(it, this) }
-      ?: parent!!.getPklBaseMethod(name) ?: getMethod(name)
+      ?: parent!!.getPklBaseMethod(name)
+      ?: getMethod(name)
 
   override fun resolveVariable(name: String): DocScope? =
     name.takeIf { it == "module" }?.let { this }
@@ -370,11 +391,12 @@ internal class ModuleScope(
         val mod = resolveImport(uri)
         resolveModuleNameToDocUrl(mod.moduleName)?.let { url -> ModuleScope(mod, url, null) }
       }
-        ?: module.moduleClass.properties[name]?.let { PropertyScope(it, this) }
+      ?: module.moduleClass.properties[name]?.let { PropertyScope(it, this) }
       // inherited classes/type aliases are in scope when resolving types -> search `all`
       ?: module.allClasses[name]?.let { ClassScope(it, url, this) }
-        ?: module.allTypeAliases[name]?.let { TypeAliasScope(it, url, this) }
-        ?: parent!!.getPklBaseProperty(name) ?: getProperty(name)
+      ?: module.allTypeAliases[name]?.let { TypeAliasScope(it, url, this) }
+      ?: parent!!.getPklBaseProperty(name)
+      ?: getProperty(name)
 
   override fun equals(other: Any?): Boolean = other is ModuleScope && module == other.module
 
@@ -384,14 +406,23 @@ internal class ModuleScope(
 internal class ClassScope(
   val clazz: PClass,
   private val parentUrl: URI,
-  override val parent: ModuleScope?
+  override val parent: ModuleScope?,
 ) : PageScope() {
   override val url: URI by lazy {
     // `isModuleClass` distinction is relevant when this scope is a link target
-    if (clazz.isModuleClass) parentUrl else parentUrl.resolve("${clazz.simpleName}.html")
+    if (clazz.isModuleClass) parentUrl
+    else parentUrl.resolve("${clazz.simpleName.pathEncoded.uriEncodedComponent}.html")
   }
 
-  override val dataUrl: URI by lazy { parent!!.dataUrl.resolve("${clazz.simpleName}.js") }
+  override val perPackageDataUrl: URI by lazy {
+    parent!!.perPackageDataUrl.resolve("${clazz.simpleName.pathEncoded.uriEncodedComponent}.json")
+  }
+
+  override val perPackageVersionDataUrl: URI by lazy {
+    parent!!
+      .perPackageVersionDataUrl
+      .resolve("${clazz.simpleName.pathEncoded.uriEncodedComponent}.json")
+  }
 
   override fun getMethod(name: String): MethodScope? =
     clazz.allMethods[name]?.let { MethodScope(it, this) }
@@ -402,19 +433,19 @@ internal class ClassScope(
   override fun resolveModuleNameToDocUrl(name: String): URI? =
     parent!!.resolveModuleNameToDocUrl(name)
 
-  override fun resolveModuleNameToSourceUrl(
-    name: String,
-    sourceLocation: Member.SourceLocation
-  ): URI? = parent!!.resolveModuleNameToSourceUrl(name, sourceLocation)
+  override fun resolveModuleNameToSourceUrl(name: String, sourceLocation: SourceLocation): URI? =
+    parent!!.resolveModuleNameToSourceUrl(name, sourceLocation)
 
   override fun resolveMethod(name: String): MethodScope? =
     clazz.methods[name]?.let { MethodScope(it, this) }
-      ?: parent!!.resolveMethod(name) ?: getMethod(name)
+      ?: parent!!.resolveMethod(name)
+      ?: getMethod(name)
 
   override fun resolveVariable(name: String): DocScope? =
     clazz.typeParameters.find { it.name == name }?.let { ParameterScope(name, this) }
-      ?: clazz.properties[name]?.let { PropertyScope(it, this) } ?: parent!!.resolveVariable(name)
-        ?: clazz.allProperties[name]?.let { PropertyScope(it, this) }
+      ?: clazz.properties[name]?.let { PropertyScope(it, this) }
+      ?: parent!!.resolveVariable(name)
+      ?: clazz.allProperties[name]?.let { PropertyScope(it, this) }
 
   override fun equals(other: Any?): Boolean = other is ClassScope && clazz == other.clazz
 
@@ -424,7 +455,7 @@ internal class ClassScope(
 internal class TypeAliasScope(
   val typeAlias: TypeAlias,
   private val parentDocUrl: URI,
-  override val parent: ModuleScope?
+  override val parent: ModuleScope?,
 ) : DocScope() {
   override val url: URI
     get() = parentDocUrl.resolve("#${typeAlias.simpleName}")
@@ -437,7 +468,7 @@ internal class TypeAliasScope(
     // only used for page scopes
     throw UnsupportedOperationException("resolveModuleNameToDocUrl")
 
-  override fun resolveModuleNameToSourceUrl(name: String, sourceLocation: Member.SourceLocation) =
+  override fun resolveModuleNameToSourceUrl(name: String, sourceLocation: SourceLocation) =
     // only used for page scopes
     throw UnsupportedOperationException("resolveModuleNameToSourceUrl")
 
@@ -463,7 +494,7 @@ internal class MethodScope(val method: PClass.Method, override val parent: DocSc
     // only used for page scopes
     throw UnsupportedOperationException("resolveModuleNameToDocUrl")
 
-  override fun resolveModuleNameToSourceUrl(name: String, sourceLocation: Member.SourceLocation) =
+  override fun resolveModuleNameToSourceUrl(name: String, sourceLocation: SourceLocation) =
     // only used for page scopes
     throw UnsupportedOperationException("resolveModuleNameToSourceUrl")
 
@@ -471,7 +502,8 @@ internal class MethodScope(val method: PClass.Method, override val parent: DocSc
 
   override fun resolveVariable(name: String): DocScope? =
     method.typeParameters.find { it.name == name }?.let { ParameterScope(name, this) }
-      ?: method.parameters[name]?.let { ParameterScope(name, this) } ?: parent.resolveVariable(name)
+      ?: method.parameters[name]?.let { ParameterScope(name, this) }
+      ?: parent.resolveVariable(name)
 
   override fun equals(other: Any?): Boolean = other is MethodScope && method == other.method
 
@@ -480,7 +512,7 @@ internal class MethodScope(val method: PClass.Method, override val parent: DocSc
 
 internal class PropertyScope(
   val property: PClass.Property,
-  override val parent: DocScope // ModuleScope|ClassScope
+  override val parent: DocScope, // ModuleScope|ClassScope
 ) : DocScope() {
   override val url: URI
     get() = parent.url.resolve("#${property.simpleName}")
@@ -493,7 +525,7 @@ internal class PropertyScope(
     // only used for page scopes
     throw UnsupportedOperationException("resolveModuleNameToDocUrl")
 
-  override fun resolveModuleNameToSourceUrl(name: String, sourceLocation: Member.SourceLocation) =
+  override fun resolveModuleNameToSourceUrl(name: String, sourceLocation: SourceLocation) =
     // only used for page scopes
     throw UnsupportedOperationException("resolveModuleNameToSourceUrl")
 
@@ -524,7 +556,7 @@ internal class ParameterScope(val name: String, override val parent: DocScope) :
     // only used for page scopes
     throw UnsupportedOperationException("resolveModuleNameToDocUrl")
 
-  override fun resolveModuleNameToSourceUrl(name: String, sourceLocation: Member.SourceLocation) =
+  override fun resolveModuleNameToSourceUrl(name: String, sourceLocation: SourceLocation) =
     // only used for page scopes
     throw UnsupportedOperationException("resolveModuleNameToSourceUrl")
 
