@@ -1,5 +1,5 @@
-/**
- * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+/*
+ * Copyright © 2024-2025 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -15,21 +15,29 @@
  */
 package org.pkl.doc
 
+import java.io.OutputStream
 import kotlin.io.path.bufferedWriter
+import kotlin.io.path.createParentDirectories
+import kotlin.io.path.exists
 import kotlinx.html.*
 import kotlinx.html.stream.appendHTML
+import kotlinx.serialization.json.Json
 import org.commonmark.ext.gfm.tables.TablesExtension
 import org.commonmark.parser.Parser
 import org.commonmark.renderer.html.HtmlRenderer
-import org.pkl.commons.createParentDirectories
 import org.pkl.commons.toPath
 import org.pkl.core.*
 import org.pkl.core.util.IoUtils
 
 internal abstract class PageGenerator<out S>(
   protected val docsiteInfo: DocsiteInfo,
-  protected val pageScope: S
-) where S : PageScope {
+  protected val pageScope: S,
+  consoleOut: OutputStream,
+) : AbstractGenerator(consoleOut) where S : PageScope {
+  companion object {
+    private val json = Json {}
+  }
+
   private val markdownInlineParserFactory = MarkdownParserFactory(pageScope)
 
   private val markdownParser =
@@ -47,33 +55,62 @@ internal abstract class PageGenerator<out S>(
   fun run() {
     val path = pageScope.url.toPath()
     path.createParentDirectories()
-    path.bufferedWriter().use {
-      it.appendLine("<!DOCTYPE html>")
-      it.appendHTML().html(null, html)
+    path.bufferedWriter().use { writer ->
+      writer.appendLine("<!DOCTYPE html>")
+      writer.appendHTML().html(null, html)
     }
+    writeOutput("Wrote file ${pageScope.url}\r")
   }
 
   protected abstract val html: HTML.() -> Unit
 
   protected abstract fun HTMLTag.renderPageTitle()
 
+  private fun DocScope?.getVersion(): String? {
+    return when (this) {
+      null -> null
+      is SiteScope -> null
+      is PackageScope -> version
+      else -> parent?.getVersion()
+    }
+  }
+
   protected fun HTML.renderHtmlHead() {
+    lang = "en-US"
+
     head {
+      meta { charset = "UTF-8" }
       title { renderPageTitle() }
       script {
         src = pageScope.relativeSiteUrl.resolve("scripts/pkldoc.js").toString()
         defer = true
       }
+      pageScope.getVersion()?.let { version ->
+        script {
+          type = "module"
+          unsafe {
+            val packageVersionUrl =
+              IoUtils.relativize(pageScope.perPackageDataUrl, pageScope.url).toString()
+            val perPackageVersionUrl =
+              IoUtils.relativize(pageScope.perPackageVersionDataUrl, pageScope.url).toString()
+
+            raw(
+              """
+      import perPackageData from "$packageVersionUrl" with { type: "json" }
+      import perPackageVersionData from "$perPackageVersionUrl" with { type: "json" }
+
+      runtimeData.knownVersions(perPackageData.knownVersions, ${json.encodeToString(version)});
+      runtimeData.knownUsagesOrSubtypes("known-subtypes", perPackageVersionData.knownSubtypes);
+      runtimeData.knownUsagesOrSubtypes("known-usages", perPackageVersionData.knownUsages);
+            """
+            )
+          }
+        }
+      }
       script {
         src = pageScope.relativeSiteUrl.resolve("scripts/scroll-into-view.min.js").toString()
         defer = true
       }
-      if (pageScope !is SiteScope) {
-        script {
-          src = IoUtils.relativize(pageScope.dataUrl, pageScope.url).toString()
-          defer = true
-        }
-      }
       link {
         href = pageScope.relativeSiteUrl.resolve("styles/pkldoc.css").toString()
         media = "screen"
@@ -102,7 +139,6 @@ internal abstract class PageGenerator<out S>(
         sizes = "16x16"
         href = pageScope.relativeSiteUrl.resolve("images/favicon-16x16.png").toString()
       }
-      meta { charset = "UTF-8" }
     }
   }
 
@@ -110,7 +146,7 @@ internal abstract class PageGenerator<out S>(
     packageName: String?,
     packageVersion: String?,
     moduleName: String?,
-    className: String?
+    className: String?,
   ) {
     header {
       if (docsiteInfo.title != null) {
@@ -127,10 +163,13 @@ internal abstract class PageGenerator<out S>(
       div {
         id = "search"
 
-        i {
-          id = "search-icon"
-          classes = setOf("material-icons")
-          +"search"
+        label {
+          htmlFor = "search-input"
+          i {
+            id = "search-icon"
+            classes = setOf("material-icons")
+            +"search"
+          }
         }
 
         input {
@@ -227,11 +266,11 @@ internal abstract class PageGenerator<out S>(
   ) {
     ul {
       classes = setOf("member-group-links")
-      for ((name, _href, show) in groups) {
+      for ((name, href, show) in groups) {
         if (show) {
           li {
             a {
-              href = _href
+              this.href = href
               +name
             }
           }
@@ -268,19 +307,19 @@ internal abstract class PageGenerator<out S>(
       a {
         href = targetScope.urlRelativeTo(pageScope).toString()
         classes = setOf(cssClass)
-        +clazz.simpleDisplayName
+        +clazz.simpleDisplayName.asIdentifier
       }
     } else {
       span {
         classes = setOf(cssClass)
-        +clazz.simpleDisplayName
+        +clazz.simpleDisplayName.asIdentifier
       }
     }
   }
 
   protected fun HtmlBlockTag.renderTypeAliasName(
     typeAlias: TypeAlias,
-    cssClass: String = "name-ref"
+    cssClass: String = "name-ref",
   ) {
     val moduleDocUrl = pageScope.resolveModuleNameToDocUrl(typeAlias.moduleName)
 
@@ -289,12 +328,12 @@ internal abstract class PageGenerator<out S>(
       a {
         href = targetScope.urlRelativeTo(pageScope).toString()
         classes = setOf(cssClass)
-        +typeAlias.simpleName
+        +typeAlias.simpleName.asIdentifier
       }
     } else {
       span {
         classes = setOf(cssClass)
-        +typeAlias.simpleName
+        +typeAlias.simpleName.asIdentifier
       }
     }
   }
@@ -302,7 +341,7 @@ internal abstract class PageGenerator<out S>(
   protected fun HtmlBlockTag.renderType(
     type: PType,
     currScope: DocScope,
-    isNested: Boolean = false
+    isNested: Boolean = false,
   ) {
     when (type) {
       PType.UNKNOWN -> {
@@ -390,7 +429,7 @@ internal abstract class PageGenerator<out S>(
 
   private fun HtmlBlockTag.renderTypeVariable(
     typeVariable: PType.TypeVariable,
-    currentScope: DocScope
+    currentScope: DocScope,
   ) {
     val parameterScope = currentScope.resolveVariable(typeVariable.name) as? ParameterScope
 
@@ -426,7 +465,7 @@ internal abstract class PageGenerator<out S>(
   // anchors, and requires no JS
   protected fun HtmlBlockTag.renderAnchor(anchorId: String, cssClass: String = "anchor") {
     div {
-      id = anchorId
+      id = anchorId.uriEncodedComponent
       classes = setOf(cssClass)
       +" " // needs some content to be considered a valid anchor by browsers
     }
@@ -457,7 +496,7 @@ internal abstract class PageGenerator<out S>(
   protected fun HtmlBlockTag.renderSelfLink(memberName: String) {
     a {
       classes = setOf("member-selflink", "material-icons")
-      href = "#$memberName"
+      href = "#${memberName.uriEncodedComponent}"
       +"link"
     }
   }
@@ -518,20 +557,23 @@ internal abstract class PageGenerator<out S>(
       }
     }
 
-    if (docPackage.docPackageInfo.dependencies.isNotEmpty()) {
+    // Every package implicitly depends on `pkl`; omit to reduce noise.
+    val dependencies = docPackage.docPackageInfo.dependencies.filter { it.name != "pkl" }
+    if (dependencies.isNotEmpty()) {
       result[MemberInfoKey("Dependencies")] = {
         var first = true
-        for (dep in docPackage.docPackageInfo.dependencies) {
+        for (dep in dependencies) {
           if (first) first = false else +", "
-          a {
-            href =
-              dep.documentation?.toString()
-                ?: pageScope.relativeSiteUrl
-                  .resolve("${dep.name}/${dep.version}/index.html")
-                  .toString()
-            +dep.name
-            +":"
-            +dep.version
+          val link = getDependencyLink(dep)
+          if (link != null) {
+            a {
+              href = link
+              +dep.name
+              +":"
+              +dep.version
+            }
+          } else {
+            span { +"${dep.name}:${dep.version}" }
           }
         }
       }
@@ -554,6 +596,32 @@ internal abstract class PageGenerator<out S>(
     return result
   }
 
+  private fun getDependencyLink(dep: DocPackageInfo.PackageDependency): String? {
+    val siteScope = pageScope.siteScope ?: return null
+    return siteScope.packageScopes.values
+      .find { it.name == dep.qualifiedName }
+      ?.urlForVersionRelativeTo(pageScope, dep.version)
+      ?.toString() ?: findAlreadyExistingPackageScope(siteScope, dep.qualifiedName, dep.version)
+  }
+
+  private fun findAlreadyExistingPackageScope(
+    siteScope: SiteScope,
+    name: String,
+    version: String,
+  ): String? {
+    // hack: determine if we know about this dependency just by checking to see if the directory
+    // exists.
+    // we don't need to know if the specific _version_ exists (we assume that the version must be
+    // published by the docsite and tolerate possibly broken links).
+    val pkgPath = siteScope.outputDir.resolve(name.pathEncoded)
+    if (pkgPath.exists()) {
+      val targetPath = pkgPath.resolve("$version/index.html")
+      val myUrl = pageScope.url
+      return IoUtils.relativize(targetPath.toUri(), myUrl).toString()
+    }
+    return null
+  }
+
   protected class MemberInfoKey(val name: String, val classes: Set<String> = setOf())
 
   protected fun collectMemberInfo(
@@ -600,19 +668,20 @@ internal abstract class PageGenerator<out S>(
         for (example in examples) {
           if (first) first = false else +", "
           a {
-            href = docModule.parent.docPackageInfo.getModuleSourceCode(example.moduleName)!!
+            href =
+              docModule.parent.docPackageInfo.getModuleSourceCode(example.moduleName)!!.toString()
             +example.shortModuleName
           }
         }
       }
     }
 
-    result[MemberInfoKey("Known subtypes", runtimeDataClasses)] = {
+    result[MemberInfoKey("Known subtypes in package", runtimeDataClasses)] = {
       id = HtmlConstants.KNOWN_SUBTYPES
       classes = runtimeDataClasses
     }
 
-    result[MemberInfoKey("Known usages", runtimeDataClasses)] = {
+    result[MemberInfoKey("Known usages in package", runtimeDataClasses)] = {
       id = HtmlConstants.KNOWN_USAGES
       classes = runtimeDataClasses
     }
@@ -631,7 +700,7 @@ internal abstract class PageGenerator<out S>(
     annotations: List<PObject>,
     /** Whether these member docs are for the main declaration at the top of a page. */
     private val isDeclaration: Boolean = false,
-    private val extraMemberInfo: Map<MemberInfoKey, HtmlBlockTag.() -> Unit> = mapOf()
+    private val extraMemberInfo: Map<MemberInfoKey, HtmlBlockTag.() -> Unit> = mapOf(),
   ) {
     init {
       markdownInlineParserFactory.docScope = docScope
