@@ -1,5 +1,5 @@
-/**
- * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+/*
+ * Copyright © 2024-2025 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -16,7 +16,8 @@
 package org.pkl.gradle;
 
 import java.io.File;
-import java.lang.reflect.InvocationTargetException;
+import java.nio.file.Files;
+import java.util.Collections;
 import java.util.LinkedHashSet;
 import java.util.List;
 import java.util.Optional;
@@ -26,8 +27,8 @@ import org.gradle.api.GradleException;
 import org.gradle.api.NamedDomainObjectContainer;
 import org.gradle.api.Plugin;
 import org.gradle.api.Project;
+import org.gradle.api.Transformer;
 import org.gradle.api.file.SourceDirectorySet;
-import org.gradle.api.plugins.Convention;
 import org.gradle.api.provider.Provider;
 import org.gradle.api.tasks.SourceSet;
 import org.gradle.api.tasks.SourceSetContainer;
@@ -36,8 +37,12 @@ import org.gradle.language.base.plugins.LifecycleBasePlugin;
 import org.gradle.plugins.ide.idea.model.IdeaModel;
 import org.gradle.util.GradleVersion;
 import org.pkl.cli.CliEvaluatorOptions;
+import org.pkl.core.ImportGraph;
+import org.pkl.core.OutputFormat;
 import org.pkl.core.util.IoUtils;
 import org.pkl.core.util.LateInit;
+import org.pkl.core.util.Nullable;
+import org.pkl.gradle.spec.AnalyzeImportsSpec;
 import org.pkl.gradle.spec.BasePklSpec;
 import org.pkl.gradle.spec.CodeGenSpec;
 import org.pkl.gradle.spec.EvalSpec;
@@ -48,6 +53,7 @@ import org.pkl.gradle.spec.PkldocSpec;
 import org.pkl.gradle.spec.ProjectPackageSpec;
 import org.pkl.gradle.spec.ProjectResolveSpec;
 import org.pkl.gradle.spec.TestSpec;
+import org.pkl.gradle.task.AnalyzeImportsTask;
 import org.pkl.gradle.task.BasePklTask;
 import org.pkl.gradle.task.CodeGenTask;
 import org.pkl.gradle.task.EvalTask;
@@ -58,11 +64,12 @@ import org.pkl.gradle.task.PkldocTask;
 import org.pkl.gradle.task.ProjectPackageTask;
 import org.pkl.gradle.task.ProjectResolveTask;
 import org.pkl.gradle.task.TestTask;
+import org.pkl.gradle.utils.PluginUtils;
 
 @SuppressWarnings("unused")
 public class PklPlugin implements Plugin<Project> {
 
-  private static final String MIN_GRADLE_VERSION = "7.2";
+  private static final String MIN_GRADLE_VERSION = "8.2";
 
   @LateInit private Project project;
 
@@ -87,6 +94,7 @@ public class PklPlugin implements Plugin<Project> {
     configureTestTasks(extension.getTests());
     configureProjectPackageTasks(extension.getProject().getPackagers());
     configureProjectResolveTasks(extension.getProject().getResolvers());
+    configureAnalyzeImportsTasks(extension.getAnalyzers().getImports());
   }
 
   private void configureProjectPackageTasks(NamedDomainObjectContainer<ProjectPackageSpec> specs) {
@@ -128,6 +136,21 @@ public class PklPlugin implements Plugin<Project> {
         });
   }
 
+  private void configureAnalyzeImportsTasks(NamedDomainObjectContainer<AnalyzeImportsSpec> specs) {
+    specs.all(
+        spec -> {
+          configureBaseSpec(spec);
+          spec.getOutputFormat().convention(OutputFormat.PCF.toString());
+          var analyzeImportsTask = createTask(AnalyzeImportsTask.class, spec);
+          analyzeImportsTask.configure(
+              task -> {
+                task.getOutputFormat().set(spec.getOutputFormat());
+                task.getOutputFile().set(spec.getOutputFile());
+                configureModulesTask(task, spec, null);
+              });
+        });
+  }
+
   private void configureEvalTasks(NamedDomainObjectContainer<EvalSpec> specs) {
     specs.all(
         spec -> {
@@ -141,7 +164,7 @@ public class PklPlugin implements Plugin<Project> {
                       // and the working directory is set to the project directory,
                       // so this path works correctly.
                       .file("%{moduleDir}/%{moduleName}.%{outputFormat}"));
-          spec.getOutputFormat().convention("pcf");
+          spec.getOutputFormat().convention(OutputFormat.PCF.toString());
           spec.getModuleOutputSeparator()
               .convention(CliEvaluatorOptions.Companion.getDefaults().getModuleOutputSeparator());
           spec.getExpression()
@@ -167,6 +190,15 @@ public class PklPlugin implements Plugin<Project> {
 
           spec.getGenerateGetters().convention(false);
           spec.getGenerateJavadoc().convention(false);
+          // Not using `convention()` so that users can disable generation of
+          // constructor parameters annotations by setting this property to `null`.
+          spec.getParamsAnnotation()
+              .set(
+                  project.provider(
+                      () ->
+                          spec.getGenerateSpringBootConfig().get()
+                              ? null
+                              : "org.pkl.config.java.mapper.Named"));
 
           createModulesTask(JavaCodeGenTask.class, spec)
               .configure(
@@ -227,8 +259,14 @@ public class PklPlugin implements Plugin<Project> {
                       .getBuildDirectory()
                       .map(it -> it.dir("pkldoc").dir(spec.getName())));
 
+          spec.getNoSymlinks().convention(false);
+
           createModulesTask(PkldocTask.class, spec)
-              .configure(task -> task.getOutputDir().set(spec.getOutputDir()));
+              .configure(
+                  task -> {
+                    task.getOutputDir().set(spec.getOutputDir());
+                    task.getNoSymlinks().set(spec.getNoSymlinks());
+                  });
         });
   }
 
@@ -279,6 +317,10 @@ public class PklPlugin implements Plugin<Project> {
     spec.getModuleCacheDir().set(IoUtils.getDefaultModuleCacheDir().toFile());
 
     spec.getNoCache().convention(false);
+
+    spec.getTestPort().convention(-1);
+
+    spec.getHttpNoProxy().convention(List.of());
   }
 
   private void configureCodeGenSpec(CodeGenSpec spec) {
@@ -308,6 +350,8 @@ public class PklPlugin implements Plugin<Project> {
 
     spec.getImplementSerializable().convention(false);
 
+    spec.getAddGeneratedAnnotation().convention(false);
+
     configureCodeGenSpecModulePath(spec);
   }
 
@@ -396,7 +440,7 @@ public class PklPlugin implements Plugin<Project> {
               var outputDir = spec.getOutputDir().get().getAsFile();
               module.getGeneratedSourceDirs().add(outputDir);
               if (spec.getSourceSet().get().getName().toLowerCase().contains("test")) {
-                module.setTestSourceDirs(append(module.getTestSourceDirs(), outputDir));
+                module.getTestSources().from(append(module.getTestSources().getFiles(), outputDir));
               } else {
                 module.setSourceDirs(append(module.getSourceDirs(), outputDir));
               }
@@ -408,6 +452,8 @@ public class PklPlugin implements Plugin<Project> {
     task.getOutputDir().set(spec.getOutputDir());
     task.getGenerateSpringBootConfig().set(spec.getGenerateSpringBootConfig());
     task.getImplementSerializable().set(spec.getImplementSerializable());
+    task.getAddGeneratedAnnotation().set(spec.getAddGeneratedAnnotation());
+    task.getRenames().set(spec.getRenames());
   }
 
   private <T extends BasePklTask, S extends BasePklSpec> void configureBaseTask(T task, S spec) {
@@ -418,25 +464,115 @@ public class PklPlugin implements Plugin<Project> {
     task.getModulePath().from(spec.getModulePath());
     task.getSettingsModule().set(spec.getSettingsModule());
     task.getEvalRootDir().set(spec.getEvalRootDir());
+    task.getColor().set(spec.getColor());
     task.getNoCache().set(spec.getNoCache());
     task.getModuleCacheDir().set(spec.getModuleCacheDir());
     task.getEvalTimeout().set(spec.getEvalTimeout());
+    task.getTestPort().set(spec.getTestPort());
+    task.getHttpProxy().set(spec.getHttpProxy());
+    task.getHttpNoProxy().set(spec.getHttpNoProxy());
+    task.getHttpRewrites().set(spec.getHttpRewrites());
   }
 
-  private <T extends ModulesTask, S extends ModulesSpec> void configureModulesTask(T task, S spec) {
+  private List<File> getTransitiveModules(AnalyzeImportsTask analyzeTask) {
+    var outputFile = analyzeTask.getOutputFile().get().getAsFile().toPath();
+    if (!analyzeTask.getOnlyIf().isSatisfiedBy(analyzeTask)) {
+      return Collections.emptyList();
+    }
+    try {
+      var contents = Files.readString(outputFile);
+      ImportGraph importGraph = ImportGraph.parseFromJson(contents);
+      var imports = importGraph.resolvedImports().values();
+      return imports.stream()
+          .filter((it) -> it.getScheme().equalsIgnoreCase("file"))
+          .map(File::new)
+          .toList();
+    } catch (Exception e) {
+      throw new RuntimeException(e);
+    }
+  }
+
+  private <T extends ModulesTask, S extends ModulesSpec> void configureModulesTask(
+      T task,
+      S spec,
+      @Nullable TaskProvider<AnalyzeImportsTask> analyzeImportsTask,
+      @Nullable Transformer<List<?>, List<?>> mapSourceModules) {
     configureBaseTask(task, spec);
-    task.getSourceModules().set(spec.getSourceModules());
-    task.getTransitiveModules().from(spec.getTransitiveModules());
+    if (mapSourceModules != null) {
+      task.getSourceModules().set(spec.getSourceModules().map(mapSourceModules));
+    } else {
+      task.getSourceModules().set(spec.getSourceModules());
+    }
     task.getNoProject().set(spec.getNoProject());
     task.getProjectDir().set(spec.getProjectDir());
     task.getOmitProjectSettings().set(spec.getOmitProjectSettings());
+    if (!spec.getTransitiveModules().isEmpty()) {
+      task.getTransitiveModules().set(spec.getTransitiveModules());
+    } else if (analyzeImportsTask != null) {
+      task.dependsOn(analyzeImportsTask);
+      task.getTransitiveModules().set(analyzeImportsTask.map(this::getTransitiveModules));
+    }
   }
 
-  private <T extends ModulesTask> TaskProvider<T> createModulesTask(
-      Class<T> taskClass, ModulesSpec spec) {
+  private <T extends ModulesTask, S extends ModulesSpec> void configureModulesTask(
+      T task, S spec, @Nullable TaskProvider<AnalyzeImportsTask> analyzeImportsTask) {
+    configureModulesTask(task, spec, analyzeImportsTask, null);
+  }
+
+  private TaskProvider<AnalyzeImportsTask> createAnalyzeImportsTask(ModulesSpec spec) {
+    var outputFile =
+        project
+            .getLayout()
+            .getBuildDirectory()
+            .file("pkl-gradle/imports/" + spec.getName() + ".json");
     return project
         .getTasks()
-        .register(spec.getName(), taskClass, task -> configureModulesTask(task, spec));
+        .register(
+            spec.getName() + "GatherImports",
+            AnalyzeImportsTask.class,
+            task -> {
+              configureModulesTask(
+                  task,
+                  spec,
+                  null,
+                  (modules) ->
+                      // only need to analyze imports of file-based modules; it's unlikely that a
+                      // non-file-based module will import a file-based module due to security
+                      // manager trust levels (see
+                      // org.pkl.core.SecurityManagers.getDefaultTrustLevel).
+                      modules.stream()
+                          .map(PluginUtils::parseModuleNotationToUri)
+                          .filter(
+                              (it) ->
+                                  it.getScheme() == null || it.getScheme().equalsIgnoreCase("file"))
+                          .toList());
+              task.setDescription("Compute the set of imports declared by input modules");
+              task.setGroup("build");
+              task.getOutputFormat().set(OutputFormat.JSON.toString());
+              task.getOutputFile().set(outputFile);
+              task.onlyIf(ignored -> !task.getSourceModules().get().isEmpty());
+            });
+  }
+
+  /**
+   * Implicitly also create a task of type {@link AnalyzeImportsTask}, postfixing the spec name with
+   * {@code "GatherImports"}.
+   *
+   * <p>The resulting task depends on the analyze task, and configures its own input files based on
+   * the result of analysis.
+   *
+   * <p>The end result is that the task automatically has correct up-to-date checks without users
+   * needing to manually provide transitive modules.
+   */
+  private <T extends ModulesTask> TaskProvider<T> createModulesTask(
+      Class<T> taskClass, ModulesSpec spec) {
+    var analyzeImportsTask = createAnalyzeImportsTask(spec);
+    return project
+        .getTasks()
+        .register(
+            spec.getName(),
+            taskClass,
+            task -> configureModulesTask(task, spec, analyzeImportsTask));
   }
 
   private <T extends BasePklTask> TaskProvider<T> createTask(Class<T> taskClass, BasePklSpec spec) {
@@ -453,67 +589,10 @@ public class PklPlugin implements Plugin<Project> {
   }
 
   private Optional<SourceDirectorySet> getKotlinSourceDirectorySet(SourceSet sourceSet) {
-    // First, try loading it as an extension - 1.8+ version of Kotlin plugin does this.
     var kotlinExtension = sourceSet.getExtensions().findByName("kotlin");
-    if (kotlinExtension instanceof SourceDirectorySet) {
-      return Optional.of((SourceDirectorySet) kotlinExtension);
+    if (kotlinExtension instanceof SourceDirectorySet sourceDirSet) {
+      return Optional.of(sourceDirSet);
     }
-
-    // Otherwise, try to load it as a convention. First, we attempt to get the convention
-    // object of the source set via the HasConvention.getConvention() method.
-    // We don't use the HasConvention interface directly as it is deprecated.
-    // Then, we extract the `kotlin` plugin from the convention, which "provides"
-    // the additional properties for the source set. This plugin has a method named
-    // `getKotlin` whose return type is a source directory set, so we use reflection
-    // to call it too.
-    // Basically, this is equivalent to calling `sourceSet.kotlin`, where `kotlin` is a property
-    // contributed by a plugin also named `kotlin`.
-    // This part of logic can be removed once we stop supporting Kotlin plugin with version
-    // less than 1.8.x.
-    try {
-      var getConventionMethod = sourceSet.getClass().getMethod("getConvention");
-      var convention = getConventionMethod.invoke(sourceSet);
-      if (convention instanceof Convention) {
-        var kotlinSourceSet = ((Convention) convention).getPlugins().get("kotlin");
-        if (kotlinSourceSet == null) {
-          project
-              .getLogger()
-              .debug(
-                  "Cannot obtain Kotlin source directory set of source set [{}], "
-                      + "it does not have the `kotlin` convention plugin",
-                  sourceSet.getName());
-          return Optional.empty();
-        }
-
-        var getKotlinMethod = kotlinSourceSet.getClass().getMethod("getKotlin");
-        var kotlinSourceDirectorySet = getKotlinMethod.invoke(kotlinSourceSet);
-        if (kotlinSourceDirectorySet instanceof SourceDirectorySet) {
-          return Optional.of((SourceDirectorySet) kotlinSourceDirectorySet);
-        }
-
-        project
-            .getLogger()
-            .debug(
-                "Cannot obtain Kotlin source directory set, sourceSets.{}.kotlin is of wrong type",
-                sourceSet.getName());
-      } else {
-        project
-            .getLogger()
-            .debug(
-                "Cannot obtain Kotlin source directory set, sourceSets.{}.convention "
-                    + "returned unexpected type",
-                sourceSet.getName());
-      }
-
-    } catch (NoSuchMethodException | IllegalAccessException | InvocationTargetException e) {
-      project
-          .getLogger()
-          .debug(
-              "Cannot obtain Kotlin source directory set of source set [{}] via a convention",
-              sourceSet.getName(),
-              e);
-    }
-
     return Optional.empty();
   }
 }
