@@ -1,5 +1,5 @@
-/**
- * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+/*
+ * Copyright © 2024-2025 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -16,6 +16,7 @@
 package org.pkl.gradle.spec;
 
 import org.gradle.api.file.DirectoryProperty;
+import org.gradle.api.provider.MapProperty;
 import org.gradle.api.provider.Property;
 import org.gradle.api.tasks.SourceSet;
 
@@ -23,6 +24,8 @@ import org.gradle.api.tasks.SourceSet;
 public interface CodeGenSpec extends ModulesSpec {
   DirectoryProperty getOutputDir();
 
+  Property<Boolean> getAddGeneratedAnnotation();
+
   Property<SourceSet> getSourceSet();
 
   Property<String> getIndent();
@@ -30,4 +33,6 @@ public interface CodeGenSpec extends ModulesSpec {
   Property<Boolean> getGenerateSpringBootConfig();
 
   Property<Boolean> getImplementSerializable();
+
+  MapProperty<String, String> getRenames();
 }
