@@ -1,5 +1,5 @@
-/**
- * Copyright © 2024 Apple Inc. and the Pkl project authors. All rights reserved.
+/*
+ * Copyright © 2024-2025 Apple Inc. and the Pkl project authors. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -16,14 +16,20 @@
 package org.pkl.gradle.task;
 
 import org.gradle.api.file.DirectoryProperty;
+import org.gradle.api.provider.MapProperty;
 import org.gradle.api.provider.Property;
 import org.gradle.api.tasks.Input;
+import org.gradle.api.tasks.Optional;
 import org.gradle.api.tasks.OutputDirectory;
 
 public abstract class CodeGenTask extends ModulesTask {
   @OutputDirectory
   public abstract DirectoryProperty getOutputDir();
 
+  @Input
+  @Optional
+  public abstract Property<Boolean> getAddGeneratedAnnotation();
+
   @Input
   public abstract Property<String> getIndent();
 
@@ -32,4 +38,7 @@ public abstract class CodeGenTask extends ModulesTask {
 
   @Input
   public abstract Property<Boolean> getImplementSerializable();
+
+  @Input
+  public abstract MapProperty<String, String> getRenames();
 }
